(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Array: ARRAY_EXTRA =
   struct
      structure A = Sequence(type 'a sequence = 'a array
			     type 'a elt = 'a
			     val fromArray = fn a => a
			     open Primitive.Array)
      open A
      open Primitive.Int

      type 'a array = 'a array
      type 'a vector = 'a vector

      val array = new

      (* can't use o because of value restriction *)
      val extract = fn arg => Primitive.Vector.fromArray(extract arg)

      fun modifyi f (slice as (a, _, _)) =
	 appi (fn (i, x) => update(a, i, f(i, x))) slice

      fun modify f a = modifyi (f o #2) (wholeSlice a)

      local
	 fun make (checkSlice, sub) {src, si, len, dst, di} =
	    let val sm = checkSlice(src, si, len)
	       val diff = si - di
	    in modifyi (fn (i, _) => sub(src, i + diff)) (dst, di, SOME(sm - si))
	    end
      in
	 fun copy(arg as {src, si, len, dst, di}) =
	    if src = dst andalso si < di
	       then (* Must go right-to-left *)
		  let val sm = checkSlice(src, si, len)
		     fun loop i =
			if i < si then ()
			else (update(dst, di + i, sub(src, i))
			      ; loop(i - 1))
		  in loop(sm - 1)
		  end
	    else make(checkSlice, sub) arg
	       
	 fun copyVec arg = make(Vector.checkSlice, Primitive.Vector.unsafeSub) arg
      end

      local open Primitive.Array
      in val unsafeSub = unsafeSub
	 val unsafeUpdate = unsafeUpdate
      end
   end

structure ArrayGlobal: ARRAY_GLOBAL = Array
open ArrayGlobal
