(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Pack32(val isBigEndian: bool
	       val shift3: Word.word
	       val shift2: Word.word
	       val shift1: Word.word
	       val shift0: Word.word): PACK_WORD =
   struct
      structure LW = LargeWord
      structure W8 = Word8

      val bytesPerElem = 4
      val isBigEndian = isBigEndian

      local
	 fun sub s (av, i) =
	    let
	       val i = bytesPerElem * i
	       fun sub(i, shift: Word.word): LW.word =
		  LW.<<(W8.toLargeWord(s(av, i)), shift)
	    in LW.orb(sub(i, shift0),
		      LW.orb(sub(i + 1, shift1),
			     LW.orb(sub(i + 2, shift2),
				    sub(i + 3, shift3))))
	    end
      in
	 val subArr = sub Word8Array.sub
	 val subArrX = subArr
	 val subVec = sub Word8Vector.sub
	 val subVecX = subVec
      end

      fun update(a, i, w) =
	 let
	    fun update(i, shift) =
	       Word8Array.update(a, i, W8.fromLargeWord(LW.>>(w, shift)))
	 in update(i, shift0);
	    update(i + 1, shift1);
	    update(i + 2, shift2);
	    update(i + 3, shift3)
	 end
   end

structure Pack32Big =
   Pack32(val isBigEndian = true
	  val shift3 = 0w0
	  val shift2 = 0w8
	  val shift1 = 0w16
	  val shift0 = 0w24)

structure Pack32Little =
   Pack32(val isBigEndian = false
	  val shift0 = 0w0
	  val shift1 = 0w8
	  val shift2 = 0w16
	  val shift3 = 0w24)

