(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* The declaration of "safe" must be first, because it is modified by mlton in
 * order to implement the -u option.
 *)
val safe = true

type 'a array = 'a array
datatype bool = datatype bool
type char = char
type exn = exn
type int = int
type int31 = int31
type intInf = intInf
datatype list = datatype list
type nullString = string
type pointer = pointer
type real = real
datatype ref = datatype ref
type stack = stack
type string = string
type word = word
type word8 = word8
type word31 = word31
type 'a vector = 'a vector

(*  Some primitive names are special and are used in atoms/prim.fun.
 *)
   
structure Primitive =
   struct
      structure C =
	 struct
	    (* char* *)
	    structure CS =
	       struct
		  type cs = pointer

		  val null = _prim "(uint)NULL": cs;		     
		  val sub = _prim "MLTON_cs_sub": cs * int -> char;
		  val update = _prim "MLTON_cs_update": cs * int * char -> unit;
		  val isNull = _prim "MLTON_isNull": cs -> bool;
	       end
	    
	    (* char** *)
	    structure CSS =
	       struct
		  type css = pointer
		  val sub = _prim "MLTON_css_sub": css * int -> CS.cs;
	       end
	 end

      structure Array =
	 struct
	    val unsafeSub = _prim "MLTON_Array_sub": 'a array * int -> 'a;
	    val unsafeUpdate =
	       _prim "MLTON_Array_update": 'a array * int * 'a -> unit;
	    val array = _prim "MLTON_Array_array": int -> 'a array;
	    val length = _prim "MLTON_Array_length": 'a array -> int;
	 end

      structure Byte =
	 struct
	    val byteToChar = _prim "MLTON_Byte_byteToChar": word8 -> char;
	    val charToByte = _prim "MLTON_Byte_charToByte": char -> word8;
	 end

      structure Char =
	 struct
	    val ord = _prim "MLTON_Char_ord": char -> int;
	    val unsafeChr = _prim "MLTON_Char_chr": int -> char;
	    val > = _prim "MLTON_Char_gt": char * char -> bool;
	    val >= = _prim "MLTON_Char_ge": char * char -> bool;
	    val < = _prim "MLTON_Char_lt": char * char -> bool;
	    val <= = _prim "MLTON_Char_le": char * char -> bool;
	 end

      structure General =
	 struct
	    val exnName = _prim "MLTON_General_exnName": exn -> string;
	    val ! = _prim "!": 'a ref -> 'a;
	    val := = _prim ":=": 'a ref * 'a -> unit;
	 end

      structure IEEEReal =
	 struct
	    val setRoundingMode =
	       _prim "MLTON_IEEEReal_setRoundingMode": int -> unit;
	    val getRoundingMode =
	       _prim "MLTON_IEEEReal_getRoundingMode": unit -> int;
	 end

      structure Int =
	 struct
	    type int = int
	    val + = _prim "MLTON_Int_add": int * int -> int;
	    val - = _prim "MLTON_Int_sub": int * int -> int;
	    val * = _prim "MLTON_Int_mul": int * int -> int;
	    val ~ = _prim "MLTON_Int_neg": int -> int;
	    val > = _prim "MLTON_Int_gt": int * int -> bool;
	    val >= = _prim "MLTON_Int_ge": int * int -> bool;
	    val < = _prim "MLTON_Int_lt": int * int -> bool;
	    val <= = _prim "MLTON_Int_le": int * int -> bool;
	    val geu = _prim "MLTON_Int_geu": int * int -> bool;
	    val gtu = _prim "MLTON_Int_gtu": int * int -> bool;
	    val quot = _prim "MLTON_Int_quot": int * int -> int;
	    val rem = _prim "MLTON_Int_rem": int * int -> int;
	 end

      structure IntInf =
	 struct
	    type int = intInf

	    val toWord = _prim "MLTON_IntInf_toWord": int -> word;
	    val fromWord = _prim "MLTON_IntInf_fromWord": word -> int;
	    val toVector = _prim "MLTON_id": int -> word vector;
	    val fromArray = _prim "MLTON_id": word array -> int;
	    val isSmall = _prim "MLTON_IntInf_isSmall": int -> bool;
	    val areSmall = _prim "MLTON_IntInf_areSmall": int * int -> bool;

	    val quot = _prim "MLTON_IntInf_quot": int * int * word array -> int;
	    val rem = _prim "MLTON_IntInf_rem": int * int * word array -> int;
	    val + = _prim "MLTON_IntInf_add": int * int * word array -> int;
	    val - = _prim "MLTON_IntInf_sub": int * int * word array -> int;
	    val smallMul =
	       _prim "MLTON_smallMul": word * word * word ref -> word;
	    val * = _prim "MLTON_IntInf_mul": int * int * word array -> int;
	    val compare = _prim "MLTON_IntInf_compare": int * int -> Int.int;
	    val toString =
	       _prim "MLTON_IntInf_toString": int * Int.int * char array -> unit;
	    val ~ = _prim "MLTON_IntInf_neg": int * word array -> int;
	 end

      structure Real =
	 struct
	    structure Math =
	       struct
		  type real = real
		     
		  val pi = _prim "MLTON_Real_Math_pi": real;
		  val e = _prim "MLTON_Real_Math_e": real;
		  val cos = _prim "MLTON_Real_Math_cos": real -> real;
		  val sin = _prim "MLTON_Real_Math_sin": real -> real;
		  val tan = _prim "MLTON_Real_Math_tan": real -> real;
		  val cosh = _prim "MLTON_Real_Math_cosh": real -> real;
		  val sinh = _prim "MLTON_Real_Math_sinh": real -> real;
		  val tanh = _prim "MLTON_Real_Math_tanh": real -> real;
		  val acos = _prim "MLTON_Real_Math_acos": real -> real;
		  val asin = _prim "MLTON_Real_Math_asin": real -> real;
		  val atan = _prim "MLTON_Real_Math_atan": real -> real;
		  val atan2 = _prim "MLTON_Real_Math_atan2": real * real -> real;
		  val ln = _prim "MLTON_Real_Math_ln": real -> real;
		  val log10 = _prim "MLTON_Real_Math_log10": real -> real;
		  val pow = _prim "MLTON_Real_Math_pow": real * real -> real;
		  val sqrt = _prim "MLTON_Real_Math_sqrt": real -> real;
		  val exp = _prim "MLTON_Real_Math_exp": real -> real;
	       end

	    val maxFinite = _prim "MLTON_Real_maxFinite": real;
	    val minPos = _prim "MLTON_Real_minPos": real;
	    (* val minNormalPos = _prim "MLTON_Real_minNormalPos": real; *)
	    val posInf = _prim "MLTON_Real_posInf": real;
	    val + = _prim "MLTON_Real_add": real * real -> real ;
	    val - = _prim "MLTON_Real_sub": real * real -> real ;
	    val * = _prim "MLTON_Real_mul": real * real -> real ;
	    val / = _prim "MLTON_Real_div": real * real -> real ;
	    val *+ = _prim "MLTON_Real_muladd": real * real * real -> real ;
	    val *- = _prim "MLTON_Real_mulsub": real * real * real -> real ;
	    val ~ = _prim "MLTON_Real_neg": real -> real ;
	    val abs = _prim "MLTON_Real_abs": real -> real;
	    val signBit = _prim "MLTON_Real_signBit": real -> bool;
	    val copySign = _prim "MLTON_Real_copySign": real -> real;
	    val > = _prim "MLTON_Real_gt": real * real -> bool ;
	    val >= = _prim "MLTON_Real_ge": real * real -> bool ;
	    val < = _prim "MLTON_Real_lt": real * real -> bool ;
	    val <= = _prim "MLTON_Real_le": real * real -> bool ;
	    val == = _prim "MLTON_Real_equal": real * real -> bool ;
	    val != = _prim "MLTON_Real_nequal": real * real -> bool ;
	    val ?= = _prim "MLTON_Real_qequal": real * real -> bool ;
	    val isFinite = _prim "MLTON_Real_isFinite": real -> bool;
	    val isNan = _prim "MLTON_Real_isNan": real -> bool;
	    val isNormal = _prim "MLTON_Real_isNormal": real -> bool;
	    val frexp = _prim "MLTON_Real_frexp": real * int ref -> real;
	    val ldexp = _prim "MLTON_Real_ldexp": real * int -> real;
	    val modf = _prim "MLTON_Real_modf": real * real ref -> real;
	    val class = _prim "MLTON_Real_class": real -> int;
	    val fromInt = _prim "MLTON_Real_fromInt": int -> real;
	    val toInt = _prim "MLTON_Real_toInt": real -> int;
	    val round = _prim "MLTON_Real_round": real -> real;

	    (* side effects the first string *)
	    val sprintf =
	       _prim "MLTON_sprintf": char array * nullString * real -> int;
	 end

      structure MLton =
	 struct
	    val random = _prim "MLTON_MLton_random": unit -> word;
	 end
      
      structure Stdio =
	 struct
	    val print = _prim "print": string -> unit;
	 end

      structure String =
	 struct
	    val unsafeSub = _prim "MLTON_String_sub": string * int -> char;
	    val size = _prim "MLTON_String_size": string -> int;

	    val toWord8Vector = _prim "MLTON_id": string -> word8 vector;
	    val fromWord8Vector = _prim "MLTON_id": word8 vector -> string;
	    val fromCharVector = _prim "MLTON_id": char vector -> string;
	 end

      structure Vector =
	 struct
	    val unsafeSub = _prim "MLTON_Vector_sub": 'a vector * int -> 'a;
	    val length = _prim "MLTON_Vector_length": 'a vector -> int;

	    (* Don't mutate the array after you apply fromArray, because vectors
	     * are supposed to be immutable and the optimizer depends on this.
	     *)
	    val fromArray = _prim "MLTON_Vector_fromArray": 'a array -> 'a vector;
	 end

      structure Word8 =
	 struct
	    val toInt = _prim "MLTON_Word8_toInt": word8 -> int ;
	    val toIntX = _prim "MLTON_Word8_toIntX": word8 -> int ;
	    val fromInt = _prim "MLTON_Word8_fromInt": int -> word8;
	    val toLargeWord = _prim "MLTON_Word8_toLargeWord": word8 -> word;
	    val toLargeWordX = _prim "MLTON_Word8_toLargeWordX": word8 -> word;
	    val fromLargeWord = _prim "MLTON_Word8_fromLargeWord": word -> word8;
	    val orb = _prim "MLTON_Word8_orb": word8 * word8 -> word8 ;
	    val xorb = _prim "MLTON_Word8_xorb": word8 * word8 -> word8 ;
	    val andb = _prim "MLTON_Word8_andb": word8 * word8 -> word8 ;
	    val notb = _prim "MLTON_Word8_notb": word8 -> word8 ;
	    val << = _prim "MLTON_Word8_lshift": word8 * word -> word8 ;
	    val >> = _prim "MLTON_Word8_rshift": word8 * word -> word8 ;
	    val ~>> =
	       _prim "MLTON_Word8_arshift": word8 * word -> word8 ;
	    val + = _prim "MLTON_Word8_add": word8 * word8 -> word8;
	    val - = _prim "MLTON_Word8_sub": word8 * word8 -> word8;
	    val * = _prim "MLTON_Word8_mul": word8 * word8 -> word8;
	    val div = _prim "MLTON_Word8_div": word8 * word8 -> word8;
	    val mod = _prim "MLTON_Word8_mod": word8 * word8 -> word8;
	    val > = _prim "MLTON_Word8_gt" : word8 * word8 -> bool;
	    val >= = _prim "MLTON_Word8_ge": word8 * word8 -> bool;
	    val < = _prim "MLTON_Word8_lt" : word8 * word8 -> bool;
	    val <= = _prim "MLTON_Word8_le": word8 * word8 -> bool;
	 end

      structure Word32 =
	 struct
	    val toIntX = _prim "MLTON_Word32_toIntX": word -> int ;
	    val fromInt = _prim "MLTON_Word32_fromInt": int -> word;
	    val orb = _prim "MLTON_Word32_orb": word * word -> word ;
	    val xorb = _prim "MLTON_Word32_xorb": word * word -> word ;
	    val andb = _prim "MLTON_Word32_andb": word * word -> word ;
	    val notb = _prim "MLTON_Word32_notb": word -> word ;
	    val << = _prim "MLTON_Word32_lshift": word * word -> word ;
	    val >> = _prim "MLTON_Word32_rshift": word * word -> word ;
	    val ~>> =
	       _prim "MLTON_Word32_arshift": word * word -> word;
	    val + = _prim "MLTON_Word32_add": word * word -> word;
	    val - = _prim "MLTON_Word32_sub": word * word -> word;
	    val * = _prim "MLTON_Word32_mul": word * word -> word;
	    val div = _prim "MLTON_Word32_div": word * word -> word;
	    val mod = _prim "MLTON_Word32_mod": word * word -> word;
	    val > = _prim "MLTON_Word32_gt" : word * word -> bool;
	    val >= = _prim "MLTON_Word32_ge": word * word -> bool;
	    val < = _prim "MLTON_Word32_lt" : word * word -> bool;
	    val <= = _prim "MLTON_Word32_le": word * word -> bool;
	 end
   end
