(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Signals: SIGNALS =
   struct
      structure Prim = PosixPrimitive.Signal
      structure Error = PosixError
	 
      type sigset = word
      type signal = int
	 
      val r = ref 0w0

      local
	 fun make f = (f r; !r)
      in
	 val empty = make(Error.checkResult o Prim.sigemptyset)
	 val full = make(Error.checkResult o Prim.sigfillset)
      end

      local
	 fun make f (sigset, signal) =
	    (r := sigset;
	     Error.checkResult(f(r, signal));
	     !r)
      in
	 val add = make Prim.sigaddset
	 val del = make Prim.sigdelset
      end

      fun isMember(sigset, signal) =
	 (r := sigset;
	  case Prim.sigismember(r, signal) of
	     1 => true
	   | 0 => false
	   | _ => Error.error())
   
      local
	 fun make how s = (r := s; Prim.sigprocmask(how, r))
      in val block = make Prim.block
	 val unblock = make Prim.unblock
      end
   end
