(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure SystemPrimitive =
   struct
      type cstring = C.CS.cs
      type cstringArray = C.CSS.css
	 
      structure CommandLine =
	 struct
	    val argc = _prim "MLTON_argc": unit -> int;
	    val argv = _prim "MLTON_argv": unit -> cstringArray;
	    val commandName = _prim "MLTON_commandName": unit -> cstring;
	 end

      structure Date =
	 struct
	    type time = int
	    type size = int

	    structure Tm =
	       struct
		  type tm = pointer

		  val buffer = _prim "(uint)&MLTON_tm": tm;

		  val sec = _prim "MLTON_tm_sec": tm -> int;
		  val min = _prim "MLTON_tm_min": tm -> int;
		  val hour = _prim "MLTON_tm_hour": tm -> int;
		  val mday = _prim "MLTON_tm_mday": tm -> int;
		  val mon = _prim "MLTON_tm_mon": tm -> int;
		  val year = _prim "MLTON_tm_year": tm -> int;
		  val wday = _prim "MLTON_tm_wday": tm -> int;
		  val yday = _prim "MLTON_tm_yday": tm -> int;
		  val isdst = _prim "MLTON_tm_isdst": tm -> int;

		  val setSec = _prim "MLTON_set_tm_sec": tm * int -> unit;
		  val setMin = _prim "MLTON_set_tm_min": tm * int -> unit;
		  val setHour = _prim "MLTON_set_tm_hour": tm * int -> unit;
		  val setMday = _prim "MLTON_set_tm_mday": tm * int -> unit;
		  val setMon = _prim "MLTON_set_tm_mon": tm * int -> unit;
		  val setYear = _prim "MLTON_set_tm_year": tm * int -> unit;
		  val setWday = _prim "MLTON_set_tm_wday": tm * int -> unit;
		  val setYday = _prim "MLTON_set_tm_yday": tm * int -> unit;
		  val setIsdst = _prim "MLTON_set_tm_isdst": tm * int -> unit;
	       end
	       
	    val ascTime = _prim "MLTON_asctime": Tm.tm -> cstring;
	    val localTime = _prim "MLTON_localtime": time ref -> Tm.tm;
	    val gmTime = _prim "MLTON_gmtime": time ref -> Tm.tm;
	    val mkTime = _prim "MLTON_mktime": Tm.tm -> time;
	    val localOffset = _prim "MLTON_localOffset": unit -> int;

	    val strfTime =
	       _prim "MLTON_strftime":
	       char array * size * nullString * Tm.tm -> size;
	 end

      structure OS =
	 struct
	    structure FileSys =
	       struct
		  val tmpnam = _prim "MLTON_tmpnam": cstring -> cstring;
	       end
	    structure Process =
	       struct
		  val halt = _prim "MLTON_halt": int -> unit;
	       end
	 end

      structure Time =
	 struct
	    structure Timeval =
	       struct
		  type timeval = pointer

		  val buffer = _prim "(uint)&MLTON_timeval": timeval;
		  val sec = _prim "MLTON_timeval_sec": timeval -> int;
		  val usec = _prim "MLTON_timeval_usec": timeval -> int;
	       end
	    
	    val gettimeofday =
	       _prim "MLTON_gettimeofday": Timeval.timeval -> int;
	 end
   end
