(* Modified from SML/NJ sources by sweeks@research.nj.nec.com on 1999-1-7. *)

(* internal-timer.sml
 *
 * COPYRIGHT (c) 1995 AT&T Bell Laboratories.
 *
 *)

structure Timer : sig

    include TIMER
    val resetTimers : unit -> unit

  end = struct

    datatype cpu_timer = CPUT of {
	usr : Time.time, sys : Time.time, gc : Time.time
      }
    datatype real_timer = RealT of Time.time

    local
      fun mkTime (s, us) = Time.T{sec = s, usec = us}
    in
    fun getTime () =
       let val {utime, stime, ...} = Posix.ProcEnv.times()
       in
	  {
	   usr = utime,
	   sys = stime,
	   gc = Time.zeroTime (* a lie for now *)
	  }
       end
    end (* local *)

    fun startCPUTimer () = CPUT(getTime())
    fun checkCPUTimer (CPUT{usr=u0, sys=s0, gc=g0}) = let
	  val {usr, sys, gc} = getTime()
	  in
	    { usr = Time.-(usr, u0),
	      sys = Time.-(sys, s0),
	      gc = Time.-(gc, g0)
	    }
	  end
    val initCPUTime = ref(startCPUTimer ())
    fun totalCPUTimer () = !initCPUTime

    fun startRealTimer () = RealT(Time.now())
    fun checkRealTimer (RealT t) = Time.-(Time.now(), t)
    val initRealTime = ref(startRealTimer ())
    fun totalRealTimer () = !initRealTime

    fun resetTimers () = (
	  initCPUTime := startCPUTimer ();
	  initRealTime := startRealTimer ())

  end (* Timer *)


(*
 * $Log$
 *)
