(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
structure Char0 =
   struct
      open Primitive.Int Primitive.Char
	 
      type char = char
      type string = string

      val minChar = #"\000"
      val numChars = 256
      val maxOrd = 255
      val maxChar = #"\255"

      fun succ c =
	 if safe andalso c = maxChar
	    then raise Chr
	 else unsafeChr(ord c + 1)

      fun pred c =
	 if safe andalso c = minChar
	    then raise Chr
	 else unsafeChr(ord c - 1)

      fun chrOpt c =
	 if safe andalso Primitive.Int.gtu(c, maxOrd)
	    then NONE
	 else SOME(unsafeChr c)

      fun chr c =
	 case chrOpt c of
	    NONE => raise Chr
	  | SOME c => c

      val {compare, ...} = Util.makeCompare(op <)

      structure String = String0

      fun contains s =
	 let
	    val a = Array.array(numChars, false)
	    val n = String.size s
	    fun loop i =
	       if i = n then ()
	       else (Array.update(a, ord(String.sub(s, i)), true) ;
		     loop(i + 1))
	 in loop 0 ;
	    fn c => Array.sub(a, ord c)
	 end

      fun notContains s = not o (contains s)

      fun memoize (f: char -> 'a): char -> 'a =
	 let val a = Array.tabulate(numChars, f o chr)
	 in fn c => Array.sub(a, ord c)
	 end
	 
      local
	 val not = fn f => memoize(not o f)
	 infix or andd
	 fun f or g = memoize(fn c => f c orelse g c)
	 fun f andd g = memoize(fn c => f c andalso g c)
      in
	 val isLower = contains "abcdefghijklmnopqrstuvwxyz"
	 val isUpper = contains "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	 val isDigit = contains "0123456789"
	 val isAlpha = isUpper or isLower
	 val isHexDigit = isDigit or (contains "abcdefABCDEF")
	 val isAlphaNum = isAlpha or isDigit
	 val isPrint = memoize(fn c => #" " <= c andalso c <= #"~")
	 val isSpace = contains " \t\r\n\v\f"
	 val isGraph = (not isSpace) andd isPrint
	 val isPunct = isGraph andd (not isAlphaNum)
	 val isCntrl = not isPrint
	 val isAscii = memoize(fn c => c < #"\128")

	 local
	    fun make(lower, upper, diff) =
	        memoize(fn c => if lower <= c andalso c <= upper
				   then chr(ord c + diff)
				else c)
	    val diff = ord #"A" - ord #"a"
	 in
	    val toLower = make(#"A", #"Z", ~diff)
	    val toUpper = make(#"a", #"z", diff)
	 end
      end
   end

