/* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 */
/*
 * Macros and procedure declarations used by the IntInf support (bignums).
 */

#ifndef	_MLTON_INT_INF_H
#define	_MLTON_INT_INF_H


#include "mlton-lib.h"


/*
 * A pointer to a struct intInfRes_t is used to communicate the state of the
 * world back from some of the C support routines and the running MLton world.
 * The frontier slot holds the new heap frontier and the value slot holds the
 * result.  In cases where some storage might be needed, the ML code allocates
 * the maximum amount which might be needed and calls the C routine.  It then
 * uses what it must, possibly rolling the heap frontier back.
 */
struct	intInfRes_t {
	pointer	frontier,
		value;
};


/*
 * MLTON_createIntInfs() is passed an array of struct intInfInit's (along
 * with a pointer to the current GC_state) at the start of the program.
 * The array is terminated by an intInfInit with mlstr field NULL.
 * For each other entry, the globalIndex'th entry of the globals array in
 * the GC_state structure is set to the IntInf.int whose value corresponds
 * to the mlstr string.
 * On return, the GC_state must have been adjusted to account for any space
 * used.
 */
struct intInfInit {
	uint	globalIndex;
	char	*mlstr;
};


/*
 * Check if an IntInf.int is small (i.e., a fixnum).
 */
#define	MLTON_IntInf_isSmall(arg)						\
	(((uint)(arg) & 0x1) != 0)

/*
 * Check if two IntInf.int's are both small (i.e., fixnums).
 * This is a gross hack, but uses only one test.
 */
#define	MLTON_IntInf_areSmall(lhs, rhs)					\
	(((uint)(lhs) & (uint)(rhs) & 0x1) != 0)

#define MLTON_IntInf_add(lhs, rhs, space)	(				\
	intInfRes = MLTON_do_add((lhs), (rhs), (space), frontier),	\
	frontier = intInfRes->frontier,					\
	intInfRes->value)

#define MLTON_IntInf_sub(lhs, rhs, space)	(				\
	intInfRes = MLTON_do_sub((lhs), (rhs), (space), frontier),	\
	frontier = intInfRes->frontier,					\
	intInfRes->value)

#define MLTON_IntInf_toString(arg, base, str) (				\
	intInfRes = MLTON_do_toStr(arg, base, str, frontier),		\
	frontier = intInfRes->frontier,					\
	intInfRes->value)

#define	MLTON_IntInf_mul(lhs, rhs, space) (				\
	intInfRes = MLTON_do_mul((lhs), (rhs), (space), frontier),	\
	frontier = intInfRes->frontier,					\
	intInfRes->value)

#define MLTON_IntInf_neg(arg, space) (					\
	intInfRes = MLTON_do_negate(arg, (space), frontier),		\
	frontier = intInfRes->frontier,					\
	intInfRes->value)

#define	MLTON_IntInf_quot(num, den, space) (				\
	intInfRes = MLTON_do_quot((space), (num), (den), frontier),	\
	frontier = intInfRes->frontier,					\
	intInfRes->value)

#define	MLTON_IntInf_rem(num, den, space) (				\
	intInfRes = MLTON_do_rem((space), (num), (den), frontier),	\
	frontier = intInfRes->frontier,					\
	intInfRes->value)


extern void	MLTON_createIntInfs(GC_state *state, struct intInfInit inits[]);
extern struct intInfRes_t	*MLTON_do_add(pointer lhs,
					     pointer rhs,
					     pointer rspace,
					     pointer frontier),
				*MLTON_do_sub(pointer lhs,
					     pointer rhs,
					     pointer rspace,
					     pointer frontier),
				*MLTON_do_mul(pointer lhs,
					     pointer rhs,
					     pointer rspace,
					     pointer frontier),
				*MLTON_do_toStr(pointer arg,
					       int base,
					       pointer space,
					       pointer frontier),
				*MLTON_do_negate(pointer arg,
						pointer space,
						pointer frontier),
				*MLTON_do_quot(pointer space,
					      pointer num,
					      pointer den,
					      pointer frontier),
				*MLTON_do_rem(pointer space,
					     pointer num,
					     pointer den,
					     pointer frontier);
extern uint	MLTON_smallMul(uint lhs, uint rhs, pointer carry);
extern int	MLTON_IntInf_compare(pointer lhs, pointer rhs),
		MLTON_IntInf_equal(pointer lhs, pointer rhs);

#endif	/* #ifndef _MLTON_INT_INF_H */
