(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature AST_CORE_STRUCTS = 
   sig
      include AST_ATOMS
   end

signature AST_CORE = 
   sig
      include AST_CORE_STRUCTS
      
      (* controls layout of applications of Vid's *)
      val isInfix: (Vid.t -> bool) ref
      
      structure Fixity:
	 sig
	    datatype t =
	       Infix of int option
	     | Infixr of int option
	     | Nonfix

	    val isInfix: t -> bool
	    val layout: t -> Layout.t
	 end
	 
      structure Fixop:
	 sig
	    datatype t = Op | None
	 end

      structure Pat:
	 sig
	    type t
	       
	    structure Item:
	       sig
		  type pat
		  datatype t =
		     Field of Record.Field.t * pat
		   | Vid of Vid.t * Type.t option * pat option 
		     (* vid <:ty> <as pat> *)
	       end
	    sharing type Item.pat = t

	    datatype node =
	       Wild
	     | Var of {fixop: Fixop.t, name: Longvid.t}
	     | Const of Const.t
	     | Tuple of t list
	     | Record of {items: Item.t list, flexible: bool}
	     | List of t list
	     | FlatApp of t list
	     | App of Longcon.t * t
	     | Constraint of t * Type.t
	     | Layered of {fixop: Fixop.t,
			   var: Var.t,
			   constraint: Type.t option,
			   pat: t}
	       
	    include WRAPPED sharing type node' = node
			    sharing type obj = t

	    val wild: t
	    val longvid: Longvid.t -> t
	    val var: Var.t -> t
	    val const: Const.t -> t
	    val tuple: t list -> t
	    (*	    val record: {record: t Record.t, flexible:bool} -> t*)
	    val con: Con.t -> t
	    val app: Con.t * t -> t
	    val constraint: t * Type.t -> t
	    val layered: {fixop: Fixop.t,
			   var: Var.t,
			   constraint: Type.t option,
			   pat: t} -> t

	    val layout: t -> Layout.t
	 end
      
      structure Exp:
	 sig
	    type dec
	    type t
	    datatype node =
	       Const of Const.t
	     | Var of {name: Longvid.t, fixop: Fixop.t}
	     | Fn of (Pat.t * t) list
	     | FlatApp of t list
	     | App of t * t
	     | Case of t * (Pat.t * t) list
	     | Let of dec * t
	     | Seq of t list
	     | Record of t Record.t
	     | List of t list
	     | Selector of Record.Field.t
	     | Constraint of t * Type.t
	     | Handle of t * (Pat.t * t) list
	     | Raise of t
	     | If of t * t * t
	     | Andalso of t * t
	     | Orelse of t * t
	     | While of {test: t, expr: t}
	     | Prim of {name: string, ty: Type.t}

	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	       
	    val const: Const.t -> t
	    val var: Var.t -> t
	    val longvid: Longvid.t -> t
	    val fnn: (Pat.t * t) list -> t
	    val app: t * t -> t
	    val casee: t * (Pat.t * t) list -> t
	    val constraint: t * Type.t -> t
	    val con: Con.t -> t
	    val handlee: t * (Pat.t * t) list -> t
	    val raisee: t -> t
	    val tuple: t list -> t
	    val select: {tuple: t, offset: int} -> t
	    val record: t Record.t -> t
	    val lett: dec list * t -> t

	    val iff: t * t * t -> t
	    val unit: t

	    val layout: t -> Layout.t
	 end

      structure EbRhs:
	 sig
	    type t
	    datatype node =
	       Gen of Type.t option
	     | Def of Longcon.t
	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	 end

      structure Dec:
	 sig
	    type t
	    datatype node =
	       Val of Tyvar.t list * (Pat.t * Exp.t) list
	     | Valrec of Tyvar.t list * {var: Var.t,
					 ty: Type.t option,
					 fixity: Vid.t option,
					 rules: (Pat.t * Exp.t) list} list
	     | Fun of Tyvar.t list * {pats: Pat.t list,
				      resultType: Type.t option,
				      body: Exp.t} list list
	     | Type of TypBind.t
	     | Datatype of DatatypeRhs.t
	     | Abstype of {datBind: DatBind.t, body: t}
	     | Exception of (Con.t * EbRhs.t) list
	     | SeqDec of t list
	     | Local of t * t
	     | Open of Longstrid.t list
	     | Overload of Var.t * Type.t * Longvar.t list
	     | Fix of {fixity: Fixity.t, ops: Vid.t list}
	    include WRAPPED sharing type node' = node
			    sharing type obj = t
	       
            val empty: t
	    val vall: Tyvar.t list * Var.t * Exp.t -> t
	    val funn:
	       Tyvar.t list
	       * (Var.t * (Pat.t * Exp.t) list * Type.t option) list -> t
	    val exceptionn: Con.t * Type.t option -> t
	    val openn: Longstrid.t list -> t
	    val datatypee:
	        {tyvars: Tyvar.t list,
		 tycon: Tycon.t,
		 cons: (Con.t * Type.t option) list} list -> t
	       
	    val layout: t -> Layout.t
	 end
      sharing type Dec.t = Exp.dec

      val layoutLet: Layout.t * Layout.t -> Layout.t
      val layoutLocal: Layout.t * Layout.t -> Layout.t
   end
