(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature LONGID_STRUCTS =
   sig
      structure Strid: AST_ID
      structure Id: AST_ID
   end

signature LONGID =
   sig
      include LONGID_STRUCTS
      include T
      datatype node = T of {strids: Strid.t list, id: Id.t}

      include WRAPPED sharing type node' = node
		      sharing type obj = t

      (* constructors *)
      val fromString: string * Region.t -> t
      val long: Strid.t list * Id.t -> t
      val short: Id.t -> t

      (* destructors *)
      val split: t -> Strid.t list * Id.t
      val toId: t -> Id.t
	 
      (* prepend with a path: 
       prepend(([B, C], x), A) = ([A, B, C], x)
       prepends(([C, D], x), [A, B]) = ([A, B, C, D], x) *)
      val prepend: t * Strid.t -> t
      val prepends: t * Strid.t list -> t

      (* isLong returns true if the list of strids is nonempty *)
      val isLong: t -> bool

      (* the name to print out for error messages *)
      val className: string

      val bogus: t
	 
      val toString: t -> string

      (* report an unbound error message *)
      val unbound: t -> unit
   end
