(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature RECORD_STRUCTS = 
   sig
      val isSorted: bool
      structure Field: FIELD
   end

signature RECORD = 
   sig
      include RECORD_STRUCTS
	 
      type 'a t

      (* constructor *)
      val fromList: (Field.t * 'a) list -> 'a t

      (* destructor *)
      val toList: 'a t -> (Field.t * 'a) list
	 
      (* tuple[a, b, c] creates {1 = a, 2 = b, 3 = c} *)
      val tuple: 'a list -> 'a t

      val isTuple: 'a t -> bool

      (* detuple r returns the components, if r is a tuple *)
      val detupleOpt: 'a t -> 'a list option

      (* range{1 = a, 2 = b, 3 = c} returns [a, b, c] *)
      val range: 'a t -> 'a list
	 
      (* usual list operations *)
      val map: 'a t * ('a -> 'b) -> 'b t
      val foreach: 'a t * ('a -> unit) -> unit
      val exists: 'a t * ('a -> bool) -> bool
      val fold: 'a t * 'b * ('a * 'b -> 'b) -> 'b
      val peek: 'a t * Field.t -> 'a option
      val unzip: 'a t -> Field.t list * 'a list
      val zip: Field.t list * 'a list -> 'a t
	 
      (* create a record with the same fields but a new range.
       also return some additional info *)
      val change: 'a t * ('a list -> 'b list * 'c) -> 'b t * 'c
	 
      val layout: {record: 'a t,
		    separator: string,
		    extra: string,
		    layoutTuple: 'a list -> Layout.t,
		    layoutElt: 'a -> Layout.t} -> Layout.t
   end
