(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Const(S: CONST_STRUCTS): CONST = 
struct

open S

local open Ast
in structure Aconst = Const
end

datatype node =
   Char of char
 | Int of Int32.int
 | SmallIntInf of SmallIntInf.t
 | IntInf of string
 | Real of string
 | String of string
 | Word of Word32.word

datatype t = T of {node: node,
		   tycon: Tycon.t}

fun make(n, t) = T{node = n, tycon = t}

fun node(T{node, ...}) = node
fun tycon(T{tycon, ...}) = tycon

local
   fun newHash() = Word.fromInt(Int.random())
   val char = newHash()
in
   fun hash(c: t): word =
      case node c of
	 Char c =>
	    Word.xorb(char,
		      Word.fromLargeWord(Word8.toLargeWord(Byte.charToByte c)))
       | Int i =>
	    (* the div is to make sure it fits in NJ's 31 bit word *)
	    Word.fromInt(Int32.toInt(i div 2))
       | SmallIntInf s => SmallIntInf.hash s
       | IntInf s => HashString.hashString s
       | Real s => HashString.hashString s
       | String s => HashString.hashString s
       | Word w => Word.fromLargeWord w
end
   
fun 'a toAst(make: Ast.Const.t -> 'a, constrain: 'a * Ast.Type.t -> 'a) c =
   let
      fun maybeConstrain(defaultTycon, aconst) =
	 let val t = tycon c
	 in if Tycon.equals(t, defaultTycon)
	       then make aconst
	    else constrain(make aconst, Ast.Type.con(Tycon.toAst t, []))
	 end
      fun int s = maybeConstrain(Tycon.defaultInt, Aconst.Int s)
   in case node c of
      Char c => make(Aconst.Char c)
    | Int n => int(Int32.toString n)
    | SmallIntInf i => int(SmallIntInf.toMLstring i)
    | IntInf s => int s
    | Real s => make(Aconst.Real s)
    | String s => make(Aconst.String s)
    | Word w => maybeConstrain(Tycon.defaultWord, Aconst.Word w)
   end

val toAstExp = toAst(Ast.Exp.const, Ast.Exp.constraint)
val toAstPat = toAst(Ast.Pat.const, Ast.Pat.constraint)

fun equals(c, c') =
   Tycon.equals(tycon c, tycon c')
   andalso
   case (node c, node c') of
      (Char c, Char c') => c = c'
    | (Int n, Int n') => n = n'
    | (SmallIntInf n, SmallIntInf n') => SmallIntInf.equals(n, n')
    | (IntInf s, IntInf s') => String.equals(s, s')
    | (Real s, Real s') => String.equals(s, s')
    | (String s, String s') => String.equals(s, s')
    | (Word w, Word w') => w = w'
    | _ => false

fun fromInt n = T{node = Int(Int32.fromInt n),
		  tycon = Tycon.defaultInt}

fun fromString s = T{node = String s,
		     tycon = Tycon.string}

local
   open Layout
   fun wrap(pre, post, s) = str(pre ^ String.toString s ^ post)
in
   fun layout c =
      case node c of
	 Int n => str(Int32.toString n)
       | SmallIntInf i => str(SmallIntInf.toMLstring i)
       | IntInf s => str s
       | Real l => String.layout l
       | Word w => seq[str "0wx", str(Word32.toString w)]
       | Char c => wrap("#\"", "\"", String.implode[c])
       | String s => wrap("\"", "\"", s)
end

end
