(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor GenericScheme(S: GENERIC_SCHEME_STRUCTS): GENERIC_SCHEME =
struct

open S

type ty = Type.t
type tyvar = Tyvar.t
   
datatype t = T of {tyvars: Tyvar.t list, ty: Type.t}

fun ty(T{ty, ...}) = ty
fun tyvars(T{tyvars, ...}) = tyvars

fun isPoly(T{tyvars, ...}) = not(List.isEmpty tyvars)

fun fromType t = T{tyvars = [], ty = t}

val equals = fn _ => Error.unimplemented "Scheme.equals"

fun layout(T{tyvars, ty}) =
   let open Layout
      val ty = Type.layout ty
   in case tyvars of
      [] => ty
    | _ => align[seq[str "Forall ",
		     tuple(List.map(tyvars, Tyvar.layout)),
		     str "."],
		 ty]
   end

fun apply(T{tyvars, ty}, args) =
   case (tyvars, args) of
      ([], []) => ty  (* must special case this, since don't want to substitute
		       in monotypes *)
    | _ => Type.substitute(ty, List.zip(tyvars, args))

fun makeGen(numTyvars, equality, makeType): t =
   let val tyvars = List.duplicate(numTyvars, fn () =>
				   Tyvar.newNoname{equality = equality})
   in T{tyvars = tyvars,
	ty = makeType(List.map(tyvars, Type.var))}
   end

val make0 = fromType
fun make1 makeType = makeGen(1, false,
			     fn [a] => makeType a
			      | _ => Error.bug "make1")
fun make2 makeType = makeGen(2, false,
			     fn [a, b] => makeType(a, b)
			      | _ => Error.bug "make2")
fun makeEqual1 makeType = makeGen(1, true,
				 fn [a] => makeType a
				  | _ => Error.bug "makeEqual1")

end

