(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature HASH_TYPE_STRUCTS = 
   sig
      include ATOMS
   end

signature HASH_TYPE = 
   sig
      include HASH_TYPE_STRUCTS
      include TYPE_OPS sharing type tycon = Tycon.t

      val hash: t -> word
	 
      val var: Tyvar.t -> t

      val tycon: t -> Tycon.t
      val isUnit: t -> bool
      val ofConst: Const.t -> t
      val fromPrim: Prim.Type.t -> t
      val toPrim: t -> Prim.Type.t
      val variant: Tycon.t * Con.t -> t

      structure Dest :
	 sig
	    datatype dest =
	       Var of Tyvar.t
	     | Con of Tycon.t * t list
	     | Variant of Tycon.t * Con.t (* datatype variant *)
	    val dest: t -> dest
	 end

      (* substitute(t, [(a1, t1), ..., (an, tn)]) performs simultaneous
       * substitution of the ti for ai in t.
       * The ai's are not required to contain every free variable in t
       *)
      val substitute: t * (Tyvar.t * t) list -> t
	 
      (* conversion to Ast *)
      val toAst: t -> Ast.Type.t
      val optionToAst: t option -> Ast.Type.t option

      (* SOME if they are all =, else NONE.  Error for empty list. *)
      val equalss: t list -> t option
	 
      (* for reporting type errors *)
      val error: string * Layout.t -> 'a

      val containsTycon: t * Tycon.t -> bool

      val plist: t -> PropertyList.t

      (* O(1) time *)
      val equals: t * t -> bool

      val makeHomVariant:
	 {var: t * Tyvar.t -> 'a,
	  con: t * Tycon.t * 'a list -> 'a,
	  variant: Tycon.t * Con.t -> 'a}
	 -> {hom: t -> 'a,
	     destroy: unit -> unit}
	 
      val makeHom:
	 {var: t * Tyvar.t -> 'a,
	  con: t * Tycon.t * 'a list -> 'a}
	 -> {hom: t -> 'a,
	     destroy: unit -> unit}

      val makeMonoHom :
	 {con: t * Tycon.t * 'a list -> 'a}
	 -> {hom: t -> 'a,
	     destroy: unit -> unit}
	 
      val hom: {ty: t,
		var: Tyvar.t -> 'a,
		con: Tycon.t * 'a list -> 'a} -> 'a

      val layout: t -> Layout.t
      val stats: unit -> Layout.t
   end
