(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature PRIM_STRUCTS = 
   sig
      structure Longid: LONGID
      structure Scheme: SCHEME
      structure Type: TYPE
      sharing Type = Scheme.Type
   end

signature PRIM = 
   sig
      include PRIM_STRUCTS

      structure Name:
	 sig
	    datatype t =
	       (* implemented in backend.fun *)
	       Array | ArraySub | ArrayUpdate
	     | VectorSub | VectorFromArray
	     | StringSub 
	     | Ref | Deref | Assign
	     | SaveWorld
	     | SaveStack
	     (* implemented in implement-exceptions.fun *)
	     | ExnName
	     (* implemented in poly-equal.fun *)
	     | Equal
	     (* implemented in machine.fun *)
	     | Bogus (*: unit -> 'a.  Make a bogus value of any type *)
	     | ArrayLength
	     | VectorLength
	     | Eq
	     | Size
	     | CPrim of string
	 end

      type t
	 
      (* these are essential primitives that must appear in the basis *)
      val array: t
      val bogus: t
      val deref: t
      val eq: t    (* pointer equality *)
      val equal: t (* polymorphic equality *)
      val exit: t
      val intInfEqual: t
      val intPlus: t
      val print: t
      val reff: t
      val saveStack: t
      val saveWorld: t
      val size: t
      val stringEqual: t
      val stringToIntInf: t
      val vectorLength: t
      val vectorSub: t

      (* constructors *)
      val new: string * Scheme.t -> t
      val newNullary: string -> t (* new of type unit -> unit *)

      (* destructors *)
      val longid: t -> Longid.t
      val name: t -> Name.t
      val numArgs: t -> int option
      val scheme: t -> Scheme.t

      val checkApp: {
		     prim: t,
		     targs: 'a list,
		     args: 'a list,
		     con: Type.Tycon.t * 'a list -> 'a,
		     equals: 'a * 'a -> bool,
		     dearrowOpt: 'a -> ('a * 'a) option,
		     detupleOpt: 'a -> 'a list option,
		     isUnit: 'a -> bool
		     } -> 'a option

      (* example: Update, + (because of exceptions)
       * not example: Ref, Array, !, Subscript
       *)
      val maySideEffect: t -> bool

      (* examples: Ref, Array
       * not example: Update, Subscript, :=, !, +
       *)
      val isMoveable: t -> bool

      (*
       * Always returns same result when given same args.
       * isFunctional implies isMoveable.
       * isFuntional implies not maySideEffect.
       * examples: ArrayLength, Equal, +
       * not examples: Ref, !, Array, Subscript
       *)
      val isFunctional: t -> bool
	 
      val isArrayAllocation: t -> bool

      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
