(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYPE_OPS_STRUCTS =
   sig
      structure Tycon: TYCON

      type t
      val con: Tycon.t * t list -> t
      val deconOpt: t -> (Tycon.t * t list) option
   end

signature TYPE_OPS =
   sig
      (* Don't want to include TYPE_OPS_STRUCTS because don't want to propagate
       * the Tycon structure, which will cause duplicate specifications later on.
       *)
      type tycon
      type t
      val con: tycon * t list -> t
      val deconOpt: t -> (tycon * t list) option

      (* constants *)
      val bool: t
      val char: t
      val exn: t
      val int: t
      val intInf: t
      val real: t
      val stack: t
      val string: t
      val unit: t
      val unitRef: t
      val word8: t
      val word: t

      val defaultInt: t
      val defaultWord: t

      (* constructors *)
      val arrow: t * t -> t
      val array: t -> t
      val vector: t -> t
      val list: t -> t
      val reff: t -> t
      val tuple: t list -> t

      (* destructors *)
      val detycon: t -> tycon
      val dearray: t -> t
      val devector: t -> t
      val dearrow: t -> t * t
      val arg: t -> t    (* arg = #1 o dearrow *)
      val result: t -> t (* result = #2 o dearrow *)
      val dearrowOpt: t -> (t * t) option
      val dearrayOpt: t -> t option
      val deref: t -> t
      val derefOpt: t -> t option
      val detuple: t -> t list
      val detupleOpt: t -> t list option
      val nth: t * int -> t 

      (* predicates *)
      val isTuple: t -> bool
   end
