(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Type(S: TYPE_STRUCTS): TYPE =
struct

open S

structure Type =
   struct
      datatype t =
	 Var of var
       | Con of con
       | Record of record
      withtype var = Tyvar.t
      and con = Tycon.t * t list
      and record = t Record.t
      datatype t' = datatype t

      val var = Var
      fun con(c, ts) =
	 if Tycon.equals(c, Tycon.tuple)
	    then (case ts of
		     [t] => t
		   | ts => Record(Record.tuple ts))
	 else Con(c, ts)
      val record = Record

      val deconOpt =
	 fn Con(c, ts) => SOME(c, ts)
	  | Record r => (case Record.detupleOpt r of
			    NONE => NONE
			  | SOME ts => SOME(Tycon.tuple, ts))
	  | _ => NONE
   end

structure Ops = TypeOps(structure Tycon = Tycon
			open Type)
open Ops Type

structure Tyvars = UnorderedSet(Tyvar)
   
fun tyvars t =
   case t of
      Var a => Tyvars.singleton a
    | Con(_, ts) => Tyvars.unions(List.map(ts, tyvars))
    | Record r => Tyvars.unions(List.map(Record.range r, tyvars))
	 
val tyvars = Tyvars.toList o tyvars

fun toAst t =
   case t of
      Var a => Ast.Type.var a
    | Con(c, ts) => Ast.Type.con(Tycon.toAst c, List.map(ts, toAst))
    | Record r => Ast.Type.record(Record.map(r, toAst))

val layout = Ast.Type.layout o toAst

fun optionToAst z = Option.map(z, toAst)

fun substitute(t, sub) =
   let
      fun loop t =
	 case t of
	    Var a => let fun lookup [] = Error.error "substitute"
			   | lookup((a', t) ::s) =
			     if Tyvar.equals(a, a') then t else lookup s
		     in lookup sub
		     end
	  | Con(c, ts) => Con(c, List.map(ts, loop))
	  | Record r => Record(Record.map(r, loop))
   in loop t
   end

fun hom{ty, var, con} =
   let
      val rec loop =
	 fn Var a => var a
	  | Con(c, ts) => con(c, loops ts)
	  | Record r =>
	       (case Record.detupleOpt r of
		   SOME ts => con(Tycon.tuple, loops ts)
		 | NONE => Error.bug "Type.hom")
      and loops = fn ts => List.map(ts, loop)
   in loop ty
   end

end
