(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYPE_STRUCTS =
   sig
      structure Ast: AST
      structure Record: RECORD
      structure Tycon: TYCON
      structure Tyvar: TYVAR
      sharing Tyvar = Ast.Tyvar
      sharing Ast.Tycon = Tycon.AstId
      sharing Record = Ast.SortedRecord
   end

signature TYPE = 
   sig
      include TYPE_STRUCTS
      include TYPE_OPS where type tycon = Tycon.t
	    
      datatype t' =
	 Var of Tyvar.t
       | Con of Tycon.t * t' list
       | Record of t' Record.t
      sharing type t = t'
	 
      (* constructors *)
      val var: Tyvar.t -> t
      val record: t Record.t -> t

      (* a list (without duplicates) of all the type variables in a type *)
      val tyvars: t -> Tyvar.t list

      (* substitute(t, [(a1, t1), ..., (an, tn)]) performs simultaneous
       substitution of the ti for ai in t *)
      val substitute: t * (Tyvar.t * t) list -> t

      (* conversion to Ast *)
      val toAst: t -> Ast.Type.t
      val optionToAst: t option -> Ast.Type.t option

      val hom: {ty: t,
		var: Tyvar.t -> 'a,
		con: Tycon.t * 'a list -> 'a} -> 'a

      val layout: t -> Layout.t
   end

