(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(*
 *
 *)
signature ALLOCATE_REGISTERS_STRUCTS = 
   sig
      structure Cps: CPS
      structure Machine: MACHINE
   end

signature ALLOCATE_REGISTERS = 
   sig
      include ALLOCATE_REGISTERS_STRUCTS

      structure Info:
	 sig
	    datatype t =
	       T of {
		     (* Where do variables defined and/or used in this block
		      * reside. Will only be called on variables x such that
		      * shouldAllocate x. 
		      *
		      * Funcnames have the restriction that you can't assign
		      * stack offset operands to formal parameters since the
		      * stack pointer won't be set until after the parameter
		      * passing. Must use a register.
		      *)
		      varOperand: Cps.Var.t -> Machine.Operand.t,

		      gcInfo: Machine.GCInfo.t,
		      
		      (* Appear at the beginning of the block. *)
		      prelude: Machine.Statement.t list,
		      
		      (* Appear before the transfer at the end of the block. *)
		      postlude: Machine.Statement.t list
		      }
	 end
      
      (* Use Machine.Chunk.newRegister with the appropriate chunk to get a new
       * register.
       *)
      val allocate:
	 {
	  program: Cps.Program.t,

	  varType: Cps.Var.t -> Machine.Type.t,

	  funcChunk: Cps.Func.t -> Machine.Chunk.t,
	  jumpChunk: Cps.Jump.t -> Machine.Chunk.t,

	  jumpHandlers: Cps.Jump.t -> Cps.Jump.t list,

	  (* Whether or not a stack slot/register needs to be allocated for a
	   * given variable
	   *)
	  shouldAllocate: Cps.Var.t -> bool
	 }
	 ->
	 {
          (* GCInfo at the binding of a var to an array allocation. *)
	  arrayInfo: Cps.Var.t -> Machine.GCInfo.t,

	  funcInfo:
	  Cps.Func.t -> {
			 info: Info.t,
			 (* If handlers are used, this gives the stack offset
			  * where the old exnStack and handler should be stored
			  *)
			 handlerOffset: int option
			 },
	  
	  contInfo:
	  Cps.Jump.t -> {
			 (* number of bytes in frame, including return address *)
			 size: int,
			 (* offsets of live pointers in frame *)
			 liveOffsets: int list
			 },

	  handlerInfo:
	  Cps.Jump.t -> {
			 (* amount to subtract to get to the next frame *)
			 size: int,
			 (* offsets of live pointers in frame *)
			 liveOffsets: int list
			 },

	  jumpInfo: Cps.Jump.t -> Info.t
	  }
   end
