(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Mtype(S: MTYPE_STRUCTS): MTYPE = 
struct

open S

datatype dest =
   Char
 | Double
 | Int
 | Pointer
 | Uint
 | Void

datatype t = T of {dest: dest}

fun dest(T{dest, ...}) = dest

fun equals(t, t') = dest t = dest t'

local
   fun new dest = T{dest = dest}
in
   val char = new Char
   val double = new Double
   val int = new Int
   val pointer = new Pointer
   val uint = new Uint
   val void = new Void
end

val all = [char, double, int, pointer, uint, void]

fun memo f =
   let val all = List.map(all, fn t => (t, f t))
   in fn t => #2(List.lookup(all, fn (t', _) => equals(t, t')))
   end

val label = uint
  
fun isVoid t =
   case dest t of
      Void => true
    | _ => false

fun isPointer t =
   case dest t of
      Pointer => true
    | _ => false
	 
local
   val byte = 1
   val word = 4
   val double = 8
in
   fun size t =
      case dest t of
	 Char => byte
       | Double => double
       | Int => word
       | Pointer => word
       | Uint => word
       | Void => 0
end

fun toString t =
   case dest t of
      Char => "uchar"
    | Double => "double"
    | Int => "int"
    | Pointer => "pointer"
    | Uint => "uint"
    | Void => "void"

val layout = Layout.str o toString

fun align(n: int, ty: t): int =
   let val s = size ty
   in if n mod s = 0
	 then n
      else n - (n mod s) + s
   end

end
