(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature REPRESENTATION_STRUCTS = 
   sig
      structure Cps: CPS
      structure Mtype: MTYPE
   end

signature REPRESENTATION = 
   sig
      include REPRESENTATION_STRUCTS

      structure TyconRep:
	 sig
	    datatype t =
	       Prim of Mtype.t
	     | Enum
	     | EnumDirect
	     | EnumIndirect
	     | EnumIndirectTag
	     | IndirectTag

	    val void: t

	    val toMtype: t -> Mtype.t

	    val layout: t -> Layout.t
	    val equals: t * t -> bool
	 end

      (* How a constructor variant of a datatype is represented. *)
      structure ConRep:
	 sig
	    datatype t =
	       (* need no representation *)
	       Void
	     (* an integer *)
	     | Int of int
	     (* an integer, but of Pointer type *)
	     | IntCast of int
	     (* just keep the value itself *)
	     | Transparent of Mtype.t
	     (* box the arg(s) *)
	     | Tuple
	     (* box the arg(s) and add the integer tag as the first word *)
	     | TagTuple of int

	    val layout: t -> Layout.t
	 end
      
      val compute:
	 Cps.Program.t
	 -> {
	     tyconRep: Cps.Tycon.t -> TyconRep.t,
	     conRep: Cps.Con.t -> ConRep.t,
	     toMtype: Cps.Type.t -> Mtype.t
	    }
   end
