(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ABSTRACT_VALUE_STRUCTS = 
   sig
      structure Sxml: SXML
   end

signature ABSTRACT_VALUE = 
   sig
      include ABSTRACT_VALUE_STRUCTS

      structure Lambda:
	 sig
	    type t
	       
	    val dest: t -> Sxml.Lambda.t

	    val layout: t -> Layout.t
	 end

      structure Lambdas :
	 sig
	    type t

	    val equals: t * t -> bool
	    val toList: t -> Lambda.t list
	    val plist: t -> PropertyList.t
	 end

      type t

      (* constructors *)
      val fromType: Sxml.Type.t -> t
      val lambda: Sxml.Lambda.t -> t
	 
      (* operators *)
      val tuple: t list -> t (* there must be one argument that is not Type _ *)
      val select: t * int -> t
      val dearray: t -> t
      val deref: t -> t
      val primApply: {prim: Sxml.Prim.t,
		      args: t list,
		      resultTy: Sxml.Type.t} -> t
	 
      datatype dest =
	 Type of Sxml.Type.t (* type doesn't contain any arrows *)
       | Ref of t
       | Array of t
       | Vector of t
       | Tuple of t list
       | Lambdas of Lambdas.t
      val dest: t -> dest
	 
      (* no possible values correspond to me *) 
      val isEmpty: t -> bool

      val addHandler: t * (Lambda.t -> unit) -> unit

      val unify: t * t -> unit
      val coerce: {from: t, to: t} -> unit

      val typeIsFirstOrder: Sxml.Type.t -> bool

      (* destroy info associated with Sxml.Type used to keep track of arrows *)
      val destroy: unit -> unit

      val plist: t -> PropertyList.t
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
