(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
(* Wherever there is a use of Value.coerce here,
 * there must be a corresponding coercion in closure-convert.fun.
 *)
functor FlowAnalysis(S: FLOW_ANALYSIS_STRUCTS): FLOW_ANALYSIS = 
struct

val wordEqual: word * word -> bool = op =
   
open S
open Sxml

structure Value = AbstractValue(structure Sxml = Sxml)

fun analyze(p as Program.T{datatypes, body}) =
   let
      val {get = conArg: Con.t -> Value.t option, set = setConArg} =
	 Property.new(Con.plist, Property.initRaise("conArg", Con.layout))

      val _ =
	 List.foreach
	 (datatypes, fn {cons, ...} =>
	  List.foreach
	  (cons, fn {con, arg} =>
	   setConArg(con, (case arg of
			      NONE => NONE
			    | SOME t => SOME(Value.fromType t)))))

      val {set = setValue, get = value} =
	 Property.new(Var.plist, Property.initRaise("value", Var.layout))

      val varExp = value o VarExp.var
      fun varExps xs = List.map(xs, varExp)
	 
      fun merge(vs: Value.t list, ty: Type.t): Value.t =
	 let val v = Value.fromType ty
	 in List.foreach(vs, fn v' => Value.coerce{from = v', to = v})
	    ; v
	 end

(*       val traceEvalPrimExp =
 * 	 Trace.trace2("evalPrimExp", PrimExp.layout, Var.layout, Value.layout)
 *)
      
      fun evalExp(e: Exp.t): Value.t =
	 let val {decs, result} = Exp.dest e
	 in evalDecs decs; varExp result
	 end
      and evalDecs(ds: Dec.t list): unit =
	 List.foreach
	 (ds,
	  fn Dec.Fun{decs, ...} =>
                (List.foreach(decs, fn {var, ty, ...} =>
			      setValue(var, Value.fromType ty));
		 List.foreach(decs, fn {var, lambda, ...} =>
			      Value.unify(value var, evalLambda lambda)))
	   | Dec.MonoVal{var, ty, exp} =>
		let fun set v = setValue(var, v)
		   fun new() = let val v = Value.fromType ty
			       in set v; v
			       end
		in case exp of
		   PrimExp.Var x => set(varExp x)
		 | PrimExp.Const c => (new(); ())
		 | PrimExp.ConApp{con, arg, ...} =>
		      (case (arg,    conArg con) of
			  (NONE,   NONE)       => ()
			| (SOME x, SOME v)     => Value.coerce{from = varExp x,
							       to = v}
			| _                    => Error.bug "constructor mismatch"
		       ; new(); ())
		 | PrimExp.Tuple xs =>
		      if Value.typeIsFirstOrder ty
			 then (new(); ())
		      else set(Value.tuple(varExps xs))
		 | PrimExp.Select{tuple, offset} =>
		      set(Value.select(varExp tuple, offset))
		 | PrimExp.Lambda lambda => set(evalLambda lambda)
		 | PrimExp.PrimApp{prim, args, ...} =>
		      set(Value.primApply{prim = prim,
					  args = varExps args,
					  resultTy = ty})
		 | PrimExp.App{func, arg} =>
		      let val argValue = varExp arg
			 val result = new()
		      in Value.addHandler
			 (varExp func, fn l =>
			  let val lambda = Value.Lambda.dest l
			     val {arg = formal, body, ...} =
				Lambda.dest lambda
			  in Value.coerce{from = argValue, to = value formal}
			     ; Value.coerce{from = varExp(Exp.result body),
					    to = result}
			  end)
		      end
		 | PrimExp.Case{test, cases, default, ...} =>
		      let val result = new()
			 fun branch e =
			    Value.coerce{from = evalExp e, to = result}
		      in List.foreach(cases, fn (Pat.T{con, arg, ...}, e) =>
				      ((case (arg,        conArg con) of
					   (NONE,       NONE)       => ()
					 | (SOME(x, _), SOME v)     =>
					      setValue(x, v)
					 | _ => Error.bug "constructor mismatch")
				       ; branch e))
			 ; (case default of
			       NONE => ()
			     | SOME e => branch e)
		      end
		 | PrimExp.Raise _ => (new(); ())
		 | PrimExp.Handle{try, catch = (x, t), handler} =>
		      let val result = new()
		      in Value.coerce{from = evalExp try, to = result}
			 ; setValue(x, Value.fromType t)
			 ; Value.coerce{from = evalExp handler, to = result}
		      end
		end
	   | _ => Error.bug "flow analysis saw bogus Dec")

      and evalLambda lambda : Value.t =
	 let val {arg, argType, body} = Lambda.dest lambda
	 in setValue(arg, Value.fromType argType)
	    ; evalExp body
	    ; Value.lambda lambda
	 end

   in evalExp body
      ; {
	 value = value,
	 conArg = conArg
	}
   end

end
