(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Globalize(S: GLOBALIZE_STRUCTS): GLOBALIZE = 
struct

open S
open Dec PrimExp

fun globalize{program = Program.T{datatypes, body}, lambdaFree} =
   let
      local
	 val {get: Tycon.t -> bool, set} =
	    Property.new(Tycon.plist, Property.initConst false)
	 fun makeBig tycon = set(tycon, true)
	 val _ = (List.foreach(datatypes, makeBig o #tycon)
		  ; makeBig Tycon.array
		  ; makeBig Tycon.vector)
      in val tyconIsBig = get
      end

      fun typeIsSmall t =
	 let open Type
	 in case dest t of
	    Con(c, ts) =>
	       not(tyconIsBig c)
	       andalso if (Tycon.equals(c, Tycon.tuple)
			   orelse Tycon.equals(c, Tycon.reff))
			  then List.forall(ts, typeIsSmall)
		       else true
	  | _ => Error.bug "typeIsSmall saw type variable"
	 end
   
      val {get = varIsGlobal, set = setVarGlobal} =
	 Property.new(Var.plist, Property.initConst false)

      val isGlobal = varIsGlobal o VarExp.var
	 
      fun areGlobal xs = List.forall(xs, isGlobal)

      fun loopExp(e: Exp.t, once: bool): unit =
	 List.foreach
	 (Exp.decs e,
	  fn MonoVal{var, ty, exp} =>
	     let
		val global =
		   case exp of
		      Const _ => true
		    | Var x => isGlobal x
		    | Tuple xs => areGlobal xs
		    | Select{tuple, ...} => isGlobal tuple
		    | Lambda l => (loopLambda l;
				   List.forall(lambdaFree l, varIsGlobal))
		    | ConApp{arg, ...} =>
			 (case arg of
			     NONE => true
			   | SOME x => isGlobal x)
 		    | PrimApp{prim, args, ...} =>
 			 areGlobal args
 			 andalso (once
				  andalso (case Prim.name prim of
					      Prim.Name.Ref => typeIsSmall ty
					    | _ => false))
		    | Case{cases, default, ...} =>
			 (List.foreach(cases, fn (_, e) =>
				       loopExp(e, once));
			  case default of
			     NONE => ()
			   | SOME e => loopExp(e, once);
			  false)
		    | Handle{try, handler, ...} =>
			 (loopExp(try, once); loopExp(handler, once);
			  false)
		    | _ => false
	     in if global then setVarGlobal(var, true) else ()
	     end
	   | Fun{decs, ...} =>
		(case decs of
		    {lambda, ...} :: _ =>
		       (if List.forall(lambdaFree lambda, varIsGlobal)
			   then List.foreach(decs, fn {var, ...} => setVarGlobal(var, true))
			else ())
		  | _ => ();
		 List.foreach(decs, loopLambda o #lambda))
	   | _ => Error.bug "globalize saw strange dec")

      and loopLambda(l: Lambda.t): unit =
	 loopExp(Lambda.body l, false)

      val body = loopExp(body, true)

   in
      {isGlobal = varIsGlobal}
   end

end
