(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature CONTROL =
   sig
      val instrumentSxml: bool ref
	 
      (* set all flags to their default values *)
      val defaults: unit -> unit

      (*------------------------------------*)
      (*            Begin Flags             *)
      (*------------------------------------*)

      (* whether or not to print auxiliary files *)
      val aux: bool ref

      datatype chunk =
	 OneChunk
       | ChunkPerFunc
       | Coalesce of {limit: int}

      val chunk: chunk ref

      (* SOME n means that the executable should use a fixed heap of size n *)
      val fixedHeap: int option ref

      val flattenRounds: int ref

      (* The .h files that should be #include'd in the .c file. *)
      val includes: string list ref
	 
      datatype inline =
	 NonRecursive of {size: int option}
       | Leaf of {size: int option}
       | LeafNoLoop of {size: int option}
       | Recursive of {size: int}

      val inline: inline ref
      val layoutInline: inline -> Layout.t
      val setInlineSize: int -> unit

      val inlineRounds: int ref
	 
      (* *)
      val inputFile: File.t ref

      (* call count instrumentation *)
      val instrument: bool ref

      val keepCps: bool ref

      (* used by the match compiler *)
      datatype match = LeftToRight | Rotate
      val match: match ref

      (* how to use MLton *)
      datatype mode =
	 Defunctorize
(*       | Monomorphise *)
       | Compile
      val setMode: mode -> unit
      val getMode: unit -> mode

      val polyvariance:
	 {
	  rounds: int,
	  (* Only duplicate big functions when
	   * (size - small) * (number of occurrences - 1) <= product
	   *)
	  small: int,
	  product: int
	 } option ref

      (* elaborator inserts code to print a message on entry to each function *)
      val printAtFunEntry: bool ref

      (* profiling of the C code (at the granularity of Cps function). *)
      val profile: bool ref

      (* overflow, bounds, div by zero checking.
       * controls the value of the variable MLton.safe available in the Basis
       *)
      val safe: bool ref
	 
      (* run typeChecker on ILs *)
      val setTypeCheck: bool -> unit
      val shouldTypeCheck: unit -> bool

      (*------------------------------------*)
      (*             End Flags              *)
      (*------------------------------------*)

      (* tracing and other informative messages*)
      val messagesAreOn: unit -> bool
      val messagesOn: unit -> unit
      val messagesOff: unit -> unit
      val message: (unit -> Layout.t) -> unit
      val messageLay: Layout.t -> unit
      val messageStr: string -> unit
      val traceCall: string -> ('a -> 'b) -> 'a -> 'b
      val sizeMessage: string * 'a -> Layout.t

      val displays: string * ((Layout.t -> unit) -> unit) -> unit
      val display: string * (unit -> Layout.t) -> unit

      (*------------------------------------*)
      (*          Error Reporting           *)
      (*------------------------------------*)
      val reset: unit -> unit
      val anyErrors: unit -> bool

      (* abort compilation once this many errors reached *)
      val errorThreshhold: int ref

      val error: Region.t * Layout.t -> unit
      val errorStr: Region.t * string -> unit

      (*------------------------------------*)
      (*          Compiler Passes           *)
      (*------------------------------------*)
      datatype style = No | C | ML

      datatype 'a display =
	 NoDisplay
       | Layout of 'a -> Layout.t
       | Layouts of 'a * (Layout.t -> unit) -> unit

      val header: style -> Layout.t

      val pass: {name: string,
		 suffix: string,
		 style: style,
		 thunk: unit -> 'a,
		 display: 'a display} ->'a
	 
      val passTypeCheck: {name: string,
			   suffix: string,
			   style: style,
			   thunk: unit -> 'a,
			   display: 'a display,
			   typeCheck: 'a -> unit} -> 'a
	 
      val passSimplify: {name: string,
			  suffix: string,
			  style: style,
			  thunk: unit -> 'a,
			  display: 'a display,
			  simplify: 'a -> 'a,
			  typeCheck: 'a -> unit} -> 'a
   end
