(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Region(S: REGION_STRUCTS): REGION = 
struct

open S

datatype t = T of {left: int, right: int}

fun left(T{left, ...}) = left
fun right(T{right, ...}) = right
   
val bogus = T{left = ~1, right = ~1}

fun list(xs, reg) =
   case xs of	       
      [] => bogus
    | x :: _ => T{left = left(reg x),
		right = right(reg(List.last xs))}

val isBogus =
   fn (T{left = ~1, ...}) => true
    | _ => false

val lineNum = ref 1
val lineIndexes = ref[0]

fun current() =
   case !lineIndexes of
      index :: _ => T{left = index, right = index}
    | _ => Error.bug "Region.current"

fun newline index =
   (IntRef.inc lineNum ;
    ListRef.push(lineIndexes, index))

fun indexToPos index =
   let
      fun loop(indexes, lineNum) =
	 case indexes of
	    i :: indexes =>
	       if index >= i
		  then {line = lineNum,
			column = index - i}
	       else loop(indexes, lineNum-1)
	  | _ => {line = ~1, column = ~1} (*Error.bug "indexToPos"*)
   in loop(!lineIndexes, !lineNum)
   end

local open Layout
   fun layoutIndex index =
      let val {line, column} = indexToPos index
	 open Layout
      in seq[Int.layout line, str ".", Int.layout column]
      end
in   
   fun layout(T{left, right}) =
      if left = ~1
	 then str "<bogus>"
      else seq[layoutIndex left, str "-", layoutIndex right]
end

fun reset() = (lineNum := 1 ;
	       lineIndexes := [0])

val make = T

structure Wrap =
   struct
      type region = t
      datatype 'a t = T of {node: 'a, region: region}

      fun node(T{node, ...}) = node
      fun region(T{region, ...}) = region
      fun makeRegion(node, region) = T{node = node, region = region}
      fun makeRegion'(node, left, right) = T{node = node,
					   region = make{left = left,
							 right = right}}
      fun make node = makeRegion(node, bogus)
      fun dest(T{node, region}) = (node, region)
      val left = fn T{region, ...} => left region
      val right = fn T{region, ...} => right region
   end
   
end

structure Region = Region()
