(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor DeadCode(S: DEAD_CODE_STRUCTS): DEAD_CODE = 
struct

open S
open CoreML
open Pat Dec Exp

fun deadCode{basis, user} =
   let
      val {get = varIsUsed, set = setVarUsed, destroy} =
	 Property.newDest(Var.plist, Property.initConst false)

      fun foreachDefinedVar(d, f) =
	 case d of
	    Val{pat, ...} => Pat.foreachVar(pat, f)
	  | Fun{decs, ...} => List.foreach(decs, f o #var)
	  | Overload{var, ...} => f var
	  | _ => ()
		

      fun patVarIsUsed p =
	 let
	    exception Used
	 in (Pat.foreachVar(p, fn x => if varIsUsed x
					 then raise Used
				      else ());
	     false) handle Used => true
	 end	    
	 
      fun decIsNeeded d =
	 case d of
	    Val{pat, ...} => patVarIsUsed pat
	  | Fun{decs, ...} => List.exists(decs, varIsUsed o #var)
	  | Datatype _ => true
	  | Exception _ => true
	  | Overload{var, ...} => varIsUsed var

      fun useVar x = setVarUsed(x, true)
      fun useExp e = Exp.foreachVar(e, useVar)
      fun useDec d = 
	 case d of
	    Val{exp, ...} => useExp exp
	  | Fun{decs, ...} => List.foreach(decs, fn {rules, ...} =>
					   List.foreach(rules, useExp o #2))
	  | Datatype _ => ()
	  | Exception _ => ()
	  | Overload{ovlds, ...} => List.foreach(ovlds, useVar)
	       
   in List.foreach(user, useDec);

      List.fold(List.reverse basis, [], fn (d, b) =>
		if decIsNeeded d
		   then (useDec d; d :: b)
		else b)
      before destroy()
   end

end

