(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Global(S: GLOBAL_STRUCTS): GLOBAL = 
struct

open S
open PrimExp

fun equalss(cs, cs') = List.equals(cs, cs', Var.equals)

val primExpEquals =
   fn (Const c, Const c') => Const.equals(c, c')
    | (PrimApp{prim = p, targs = ts, ...}, PrimApp{prim = p', targs = ts', ...}) =>
	 (case Prim.numArgs p of
	     NONE => Prim.equals(p, p') andalso List.equals(ts, ts', Type.equals)
	   | _ => false)
    | (ConApp{con = c, args}, ConApp{con = c', args = args'}) =>
	 Con.equals(c, c') andalso equalss(args, args')
    | (Tuple xs, Tuple xs') => equalss(xs, xs')
    | _ => false

(* A really simple implementation using a list of all globals to do
 * the interning.
 *)
fun make() =
   let
      val binds: {var: Var.t, ty: Type.t, exp: PrimExp.t} list ref = ref []
	 	 
      fun all() = List.reverse(!binds) before binds := []

      fun new(ty: Type.t, exp: PrimExp.t): Var.t =
	 case List.peek(!binds, fn {exp = exp', ...} =>
			primExpEquals(exp, exp')) of
	    SOME{var, ...} => var
	  | NONE =>
	       let val var = Var.newString "global"
	       in ListRef.push(binds, {var = var, ty = ty, exp = exp})
		  ; var
	       end

(*       val new = Trace.trace2("Global.new", Type.layout, PrimExp.layout,
 * 			     Var.layout) new
 *)
	 
   in {new = new, all = all}
   end

(* A trickier implementation using hash consing.
 *)

structure Table = HashTable

local
   fun newHash() = Word.fromInt(Int.random())
   val tuple = newHash()
   val conApp = newHash()
   val primApp = newHash()
   fun hashVars(xs: Var.t list, w: word): word =
      List.fold(xs, tuple, fn (x, w) => Word.xorb(w, Var.hash x))
in
   val hash: PrimExp.t -> word =
      fn Const c => Const.hash c
       | Tuple xs => hashVars(xs, tuple)
       | ConApp{args, ...} => hashVars(args, conApp)
       | PrimApp{args, ...} => hashVars(args, primApp)
       | _ => Error.bug "hash not implemented for prim exp"
end

fun make() =
   let
      val binds: {var: Var.t, ty: Type.t, exp: PrimExp.t} list ref = ref []
      fun all() = List.reverse(!binds) before binds := []
	 
      val size = 10000
      exception NotFound
      val table = Table.mkTable (hash, primExpEquals) (size, NotFound)

      fun new(ty: Type.t, exp: PrimExp.t): Var.t =
	 case Table.find table exp of
	    SOME x => x
	  | NONE => let val x = Var.newString "global"
		    in ListRef.push(binds, {var = x, ty = ty, exp = exp})
		       ; Table.insert table (exp, x)
		       ; x
		    end	   
   in {new = new, all = all}
   end
	    
end


