(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor RemoveUnusedConstructors(S: REMOVE_UNUSED_CONSTRUCTORS_STRUCTS)
   : REMOVE_UNUSED_CONSTRUCTORS = 
struct

open S
open Dec PrimExp Transfer

fun remove(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = conIsUseful: Con.t -> bool,
	   set = setConUseful,
	   destroy = destroyConUseful} =
	 Property.newDest(Con.plist, Property.initConst false)

      fun loopBind{var, ty, exp} =
	 case exp of
	    ConApp{con, ...} => setConUseful(con, true)
	  | _ => ()

      val _ =
	 (setConUseful(Con.truee, true)
	  ; setConUseful(Con.falsee, true)
	  ; List.foreach(globals, loopBind)
	  ; List.foreach(functions, fn {body, ...} =>
			 Exp.foreachBind(body, loopBind)))

      val {get = conNumCons: Con.t -> int,
	   set = setConNumCons, destroy = destroyConNumCons} =
	 Property.newDest(Con.plist,
			  Property.initRaise("num cons", Con.layout))
	 
      val datatypes =
	 List.revMap
	 (datatypes, fn {tycon, cons} =>
	  let val cons = List.keepAll(cons, conIsUseful o #con)
	     val n = List.length cons
	  in List.foreach(cons, fn {con, ...} => setConNumCons(con, n))
	     ; {tycon = tycon,
		cons = cons}
	  end)
	 
      fun loopTransfer(t: Transfer.t): Transfer.t =
	 case t of
	    Case{cause, test, cases, default} =>
	       let val cases = List.keepAll(cases, conIsUseful o #1)
		  fun keep default = Case{cause = cause,
					  test = test,
					  cases = cases,
					  default = default}
		  fun none() = keep NONE
	       in case default of
		  NONE => none()
		| SOME j => 
		     case cases of
			[] => Jump{dst = j, args = []}
		      | (c, _) :: _ =>
			   if conNumCons c = List.length cases
			      then none()
			   else keep(SOME j)
	       end
	  | _ => t
	 
      fun loopExp(e: Exp.t): Exp.t =
	 let val {decs, transfer} = Exp.dest e
	 in Exp.make{decs = List.map(decs, loopDec),
		     transfer = loopTransfer transfer}
	 end
      and loopDec(d: Dec.t): Dec.t =
	 case d of
	    Fun{name, args, body} => Fun{name = name,
					 args = args,
					 body = loopExp body}
	  | _ => d

      val functions =
	 List.revMap
	 (functions, fn {name, args, body, returns} =>
	  {name = name, args = args, returns = returns,
	   body = loopExp body})
	 
   in destroyConUseful()
      ; destroyConNumCons()
      ; Program.T{datatypes = datatypes,
		  globals = globals,
		  functions = functions,
		  main = main}
   end

end
