(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor RemoveUnusedGlobals(S: REMOVE_UNUSED_GLOBALS_STRUCTS): REMOVE_UNUSED_GLOBALS = 
struct

open S
open Dec PrimExp

fun remove(Program.T{datatypes, globals, functions, main}) =
   let
      val {get = isGlobal: Var.t -> bool, set = setGlobal,
	   destroy = destroyGlobal} =
	 Property.newDest(Var.plist, Property.initConst false)
	 
      val {get = isUseful: Var.t -> bool, set = setUseful,
	   destroy = destroyUseful} =
	 Property.newDest(Var.plist, Property.initConst false)

      fun var x = if isGlobal x then setUseful(x, true) else ()
      fun handlePrimExp e = PrimExp.foreachVar(e, var)
      val handleDec =
	 fn Bind{exp, ...} => handlePrimExp exp
	  | _ => ()
      fun handleTransfer t = Transfer.foreachVar(t, var)
   in
      List.foreach(globals, fn {var, ...} => setGlobal(var, true))
      ; List.foreach(functions, fn {body, ...} =>
		     Exp.foreach'(body, {handleDec = handleDec,
					 handleTransfer = handleTransfer}))
      ; List.foreach(List.reverse globals, fn {var, exp, ...} =>
		     if isUseful var
			then handlePrimExp exp
		     else ())
      ; (Program.T{datatypes = datatypes,
		   globals = List.keepAll(globals, isUseful o #var),
		   functions = functions,
		   main = main}
	 before (destroyGlobal()
		 ; destroyUseful()))
   end
   
end
