(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor SaveWorld(S: SAVE_WORLD_STRUCTS): SAVE_WORLD = 
struct

open S

open Dec PrimExp Transfer

fun usesSaveWorld(Program.T{functions, ...}): bool =
   let exception Yes
   in (List.foreach
       (functions, fn {body, ...} =>
	Exp.foreachBind(body, fn {exp, ...} =>
			case exp of
			   PrimApp{prim, ...} =>
			      if (case Prim.name prim of
				     Prim.Name.SaveWorld => true
				   | Prim.Name.SaveStack => true
				   | Prim.Name.Size => true
				   | _ => false)
				 then raise Yes
			      else ()
			 | _ => ()));
       false) handle Yes => true
   end

fun saveWorld(Program.T{datatypes, globals, functions, main}) =
   let
      val sizeFunctions = ref []
      val {get = sizeName, destroy, ...} =
	 Property.newDest
	 (Type.plist,
	  Property.initFun
	  (fn t =>
	   let val f = Func.newString "size"
	      val arg = Var.newNoname()
	      val s = Var.newNoname()
	   in ListRef.push
	      (sizeFunctions,
	       {name = f,
		args = [(arg, t)],
		body = Exp.make{decs =
				[Dec.Bind{var = s,
					  ty = Type.int,
					  exp = PrimApp{prim = Prim.size,
							targs = [t],
							args = [arg]}}],
				transfer = Return[s]},
		returns = [Type.int]});
	      f
	   end))

      val saveStack = Func.newNoname()
	   
      fun replacePrimApps(e: Exp.t): Exp.t =
	 let val {decs, transfer} = Exp.dest e
	 in List.foldr
	    (decs, Exp.make{decs = [], transfer = transfer}, fn (d, e) =>
	     let fun normal() = Exp.prefix(e, d)
	     in case d of
		Fun{name, args, body} =>
		   Exp.prefix(e, Fun{name = name,
				     args = args,
				     body = replacePrimApps body})
	      | Bind{var, exp = PrimExp.PrimApp{prim, targs, args}, ...} =>
		   let
		      fun doit(name, formals) =
			 let val j = Jump.newNoname()
			 in Exp.make
			    {decs = [Dec.Fun{name = j, args = formals, body = e}],
			     transfer = Transfer.Call{func = name, args = args,
						      cont = SOME j}}
			 end
		   in case (Prim.name prim, targs) of
		      (Prim.Name.SaveWorld, []) =>
			 doit(Func.saveWorld, [(var, Type.bool)])
		    | (Prim.Name.SaveStack, []) =>
			 doit(saveStack, [(var, Type.stack)])
		    | (Prim.Name.Size, [t]) =>
			 doit(sizeName t, [(var, Type.int)])
		    | _ => normal()
		   end
	      | _ => normal()
	     end)
	 end

      val functions = 
	 List.map(functions, fn {name, args, body, returns} =>
		  {name = name,
		   args = args,
		   body = replacePrimApps body,
		   returns = returns})
	 
      fun make(func: Func.t, formals, return, prim, actuals) =
	 {name = func,
	  args = formals,
	  body = let val b = Var.newNoname()
		 in Exp.make{decs = [Bind{var = b,
					  ty = return,
					  exp = PrimApp{prim = prim,
							targs = [],
							args = actuals}}],
			     transfer = Return[b]}
		 end,
	  returns = [return]}
      val functions =
	 let val arg = Var.newNoname()
	 in make(Func.saveWorld,
		 [(arg, Type.string)],
		 Type.bool,
		 Prim.saveWorld,
		 [arg])
	 end
	 :: make(saveStack,
		 [],
		 Type.stack,
		 Prim.saveStack,
		 [])
	 :: (!sizeFunctions @ functions)
	 
   in destroy();
      Program.T{datatypes = datatypes,
		globals = globals,
		functions = functions,
		main = main}
   end

val saveWorld =
   fn p => if usesSaveWorld p then saveWorld p else p

end
