(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor TwoPointLattice(S: TWO_POINT_LATTICE_STRUCTS): TWO_POINT_LATTICE = 
struct

open S

structure Set = DisjointSet
   
datatype t = T of value Set.t
and value =
   Bottom of (unit -> unit) list ref  (* If I become Top, then run these. *)
  | Top of Top.t

fun value(T s) = Set.value s

fun dest v =
   case value v of
      Bottom _ => Dest.bottom
    | Top t => Dest.top t

fun toString e =
   case value e of
      Bottom _ => bottom
    | Top _ => top

val layout = Layout.str o toString

fun new(): t = T(Set.singleton(Bottom(ref [])))

fun equals(T s, T s') = Set.equals(s, s')

fun addHandler(e, h) =
   case value e of
      Bottom hs => ListRef.push(hs, h)
    | Top _ => h()

fun isTop s =
   case value s of
      Top _ => true
    | _ => false

fun isBottom s =
   case value s of
      Bottom _ => true
    | _ => false

fun runHandlers hs = List.foreach(!hs, fn h => h())

fun makeTop(T s) =
   case Set.value s of
      Top _ => ()
    | Bottom hs => (Set.setValue(s, Top(Top.new()));
		    runHandlers hs)

fun from <= to =
   if equals(from, to)
      then ()
   else
      case (value from, value to) of
	 (_, Top _) => ()
       | (Top _, _) => makeTop to
       | (Bottom hs, _) => ListRef.push(hs, fn () => makeTop to)

fun ==(T s, T s') =
   if Set.equals(s, s')
      then ()
   else
      let val e = Set.value s
	 val e' = Set.value s'
      in Set.union(s, s');
	 case (e, e') of
	    (Top _, Top _) => ()
	  | (Bottom hs, Top _) => (Set.setValue(s, e'); runHandlers hs)
	  | (Top _, Bottom hs) => (Set.setValue(s, e); runHandlers hs)
	  | (Bottom hs, Bottom hs') => Set.setValue(s, Bottom(ref(!hs @ !hs')))
      end

end
