(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Basis(S: BASIS_STRUCTS): BASIS = 
struct

open S

local open Ast
in structure Strid = Strid
end
local open Env
in structure Interface = Interface
end

structure SigEnv = MonoEnv(structure Domain = Ast.Sigid
			   structure Range = Interface)
   
structure FctEnv = PolyEnv(structure Domain = Ast.Fctid)

datatype t = Basis of {E: Env.t,
		       G: SigEnv.t,
		       F: functorClosure FctEnv.t}
and functorClosure = T of {arg: Strid.t,
			   argSig: Interface.t,
			   body: Ast.Strexp.t,
			   basis: t,
			   counter: int ref}
fun layout(Basis{E, G, ...}) =
   let open Layout
   in record[("E", Env.layout E),
	     ("G", SigEnv.layout G)]
   end   

val empty = Basis{E = Env.empty,
		  G = SigEnv.empty,
		  F = FctEnv.empty()}
val bogus = empty

structure FunctorClosure =
   struct
      type basis = t
      datatype t = datatype functorClosure

      val bogus = T{arg = Strid.bogus,
		    argSig = Interface.bogus,
		    body = Ast.Strexp.var Ast.Longstrid.bogus,
		    basis = bogus,
		    counter = ref 0}
      val equals = fn _ => Error.unimplemented "FunctorClosure.equals"
      val layout = fn _ => Layout.str "<functor closure>"
   end

val op + =
   fn (Basis{E, F, G}, Basis{E=E', F=F', G=G'}) =>
   Basis{E = Env.+(E, E'),
	 G = SigEnv.+(G, G'),
	 F = FctEnv.+(F, F')}

fun toEnv(Basis{E, ...}) = E
fun fromEnv E = Basis{E = E, F = FctEnv.empty(), G = SigEnv.empty}

fun prim() =
   let val (E, cs) = Env.prim()
   in (fromEnv E, cs)
   end
   
fun consolidate(Basis{E, F, G}) =
   Basis{F = F, G = G, E = Env.consolidate E}

fun plusEnv(Basis{E, G, F}, E') = Basis{E = Env.+(E, E'), G = G, F = F}

fun replaceEnv(Basis{G, F, ...}, E) = Basis{E = E, G = G, F = F}

fun extendStrid(Basis{E, G, F}, x, y) =
   Basis{G = G, F = F, E = Env.extendStrid(E, x, y)}

fun extendFctid(Basis{E, F, G}, x, y) =
   Basis{E = E, G = G, F = FctEnv.extend(F, x, y)}

fun extendSigid(Basis{E, F, G}, x, y) =
   Basis{E = E, F = F, G = SigEnv.extend(G, x, y)}

fun lookupLongstrid(Basis{E, ...}, x) = Env.lookupLongstrid(E, x)
fun lookupFctid(Basis{F, ...}, x) =
   case FctEnv.peek(F, x) of
      SOME fct => fct
    | NONE => (Ast.Fctid.unbound x ;
	       FunctorClosure.bogus)
    
fun lookupSigid(Basis{G, ...}, x) =
   case SigEnv.peek(G, x) of
      SOME I => I
    | NONE => (Ast.Sigid.unbound x ;
	       Interface.bogus)

end

