(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature BASIS_STRUCTS = 
   sig
      structure Ast: AST
      structure Decs: DECS
      structure Env: ELABORATE_ENV
      structure Interface: INTERFACE
      sharing Ast = Decs.Ast = Env.Ast = Interface.Ast
      sharing Decs.CoreML = Env.CoreML = Interface.CoreML
   end

signature BASIS = 
   sig
      include BASIS_STRUCTS

      (* the value of a functor *)
      structure FunctorClosure :
	 sig
	    type basis
	    datatype t = T of {arg: Ast.Strid.t,
			       argSig: Interface.t,
			       body: Ast.Strexp.t,
			       basis: basis,     (* for free variables *)
			       counter: int ref} (* the number of applications of me *)

	    val layout: t -> Layout.t
	 end

      type t = FunctorClosure.basis

      val empty: t
      val prim: unit -> t * Decs.CoreML.Constructors.t
	 
      val + : t * t -> t
      val toEnv: t -> Env.t
      val fromEnv: Env.t -> t
      val plusEnv: t * Env.t -> t
      val replaceEnv: t * Env.t -> t
	 
      val consolidate: t -> t
	 
      val extendStrid: t * Ast.Strid.t * Env.t -> t
      val extendSigid: t * Ast.Sigid.t * Interface.t -> t
      val extendFctid: t * Ast.Fctid.t * FunctorClosure.t -> t
	 
      val lookupLongstrid: t * Ast.Longstrid.t -> Env.t
      val lookupSigid: t * Ast.Sigid.t -> Interface.t
      val lookupFctid: t * Ast.Fctid.t -> FunctorClosure.t
	 
      val layout: t -> Layout.t
   end
