(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature ELABORATE_ATOMS_STRUCTS = 
   sig
      structure Ast: AST
      structure TypeStr: TYPE_STR
      sharing Ast = TypeStr.Ast
	 
      structure Decs:
	 sig
	    type t

	    val empty: t
	    val datatypes :
	       {
		tycon: TypeStr.Tycon.t,
		tyvars: Ast.Tyvar.t list,
		cons: {
		       con: TypeStr.Con.t,
		       arg: TypeStr.Type.t option
		      } list
	       } list -> t
	 end
      
      structure Env:
	 sig
	    type t
	    val empty: t
	    val extendTycon: t * Ast.Tycon.t * TypeStr.t -> t
	    val + : t * t -> t
	    val cons: TypeStr.Cons.t -> t
	    val lookupLongtycon: t * Ast.Longtycon.t -> TypeStr.t
	    val peekLongtycon: t * Ast.Longtycon.t -> TypeStr.t option
	 end
   end

signature ELABORATE_ATOMS = 
   sig
      include ELABORATE_ATOMS_STRUCTS

      structure Lookup:
	 sig
	    type t
	    val fromFun: (Ast.Longtycon.t -> TypeStr.t) -> t
	    val fromEnv: Env.t -> t
	    val plusEnv: t * Env.t -> t
	    val lookup: t * Ast.Longtycon.t -> TypeStr.t
	 end

      val elaborateType: Ast.Type.t * Lookup.t -> TypeStr.Scheme.t

      val elaborateTypeOpt: Ast.Type.t option * Env.t -> TypeStr.Type.t option

      val elaborateTypBind: Ast.TypBind.t * Lookup.t -> Env.t
	 
      val elaborateScheme:
 	 Ast.Tyvar.t list * Ast.Type.t * Lookup.t
	 -> TypeStr.Scheme.t * Ast.Tyvar.t list

      val elaborateClosedScheme:
	 Ast.Tyvar.t list * Ast.Type.t * Lookup.t -> TypeStr.Scheme.t

      val elaborateCons:
	 {cons: (Ast.Con.t * Ast.Type.t option) list,
	  tyvars: Ast.Tyvar.t list,
	  tycon: TypeStr.Tycon.t,
	  resultType: Ast.Type.t,
	  lookup: Lookup.t} -> TypeStr.Cons.t

      val elaborateDatBind:
	 {datBind: Ast.DatBind.t,
	  lookup: Lookup.t,
	  newTycon: Ast.Tycon.t -> TypeStr.Tycon.t}
	 -> {VE: Env.t, TE: Env.t, decs: Decs.t}

      val elaborateDatatypeRhs:
	 {rhs: Ast.DatatypeRhs.t,
	  lookup: Lookup.t,
	  newTycon: Ast.Tycon.t -> TypeStr.Tycon.t} -> Decs.t * Env.t
   end
