(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor Elaborate(S: ELABORATE_STRUCTS): ELABORATE= 
struct

open S

local open Ast
in structure FctArg = FctArg
   structure Longstrid = Longstrid
   structure Topdec = Topdec
   structure SigConst = SigConst
   structure Sigexp = Sigexp
   structure Strdec = Strdec
   structure Strid = Strid
   structure Strexp = Strexp
end

local open CoreML
in structure Con = Con
   structure Prim = Prim
   structure Scheme = Scheme
   structure Tycon = Tycon
   structure Type = Type
end

structure Decs = Decs(structure Ast = Ast
		      structure CoreML = CoreML)

structure TypeStr = TypeStr(structure Ast = Ast
			    structure Tycon = Tycon
			    structure Con = Con
			    structure Type =
			       struct
				  open Type
				  datatype t = datatype t'
			       end
			    structure Scheme = Scheme)

structure Interface = Interface(structure Ast = Ast
				structure CoreML = CoreML
				structure TypeStr = TypeStr)

structure Env = ElaborateEnv(structure Ast = Ast
			     structure CoreML = CoreML
			     structure Interface = Interface
			     structure TypeStr = TypeStr)

structure ElaborateAtoms =
   ElaborateAtoms(structure Ast = Ast
		  structure Decs =
		     struct
			open Decs
			val datatypes = single o CoreML.Dec.Datatype
		     end
		  structure Env = Env
		  structure TypeStr = TypeStr)
structure Lookup = ElaborateAtoms.Lookup

structure Basis = Basis(structure Ast = Ast
			structure Decs = Decs
			structure Env = Env
			structure Interface = Interface)
   
local open Basis
in structure FunctorClosure = FunctorClosure
end

structure ElaborateSigexp = ElaborateSigexp(structure Ast = Ast
					    structure Basis = Basis
					    structure Interface = Interface)
local open ElaborateSigexp
in val elaborateSigexp = elaborateSigexp
end

structure ElaborateCore = ElaborateCore(structure Ast = Ast
					structure CoreML = CoreML
					structure Decs = Decs
					structure ElaborateAtoms = ElaborateAtoms
					structure Env = Env)

val counters: int ref list ref = ref []
fun clearCounters() = counters := []
fun newCounter() = let val r = ref 0
		   in counters := r :: (!counters) ; r
		   end

val useOpaque = ref false
   
fun elaborateSigexpConstraint(cons: SigConst.t, B: Basis.t): Env.t -> Env.t =
   let
      fun s(sigexp, opaque) =
	 let val interface = elaborateSigexp(sigexp, B)
	 in fn E => Env.cut{env = E,
			    interface = interface,
			    opaque = opaque,
			    region = Sigexp.region sigexp}
	 end
   in case cons of
      SigConst.None => (fn E => E)
    | SigConst.Transparent sigexp => s(sigexp, false)
    | SigConst.Opaque sigexp => s(sigexp, !useOpaque)
   end	 
   
(* Elaborate d in basis B, accumulating core decs onto decs
 * and accumulating bindings into env.
 *)

(* val traceElaborateStrdec =
 *    Trace.trace4("elaborateStrdec" , Strdec.layout,
 * 		Layout.ignore: Basis.t -> Layout.t (*Basis.layout*),
 * 		Layout.ignore: Decs.t -> Layout.t (*Decs.layout*),
 * 		Layout.ignore: Env.t -> Layout.t (*Env.layout*),
 * 		Layout.ignore: Decs.t * Env.t -> Layout.t)
 * 		(*Layout.tuple2(Decs.layout, Env.layout))*)
 *)
   
fun elaborateStrdec arg: Decs.t * Env.t =
(*   traceElaborateStrdec *)
   (fn (d: Strdec.t, B: Basis.t, decs, E: Env.t) =>
    case Strdec.node d of
       Strdec.Core d => (* rule 56 *)
	  ElaborateCore.elaborateDec(d, Basis.toEnv B, decs, E)
     | Strdec.Structure strbinds => (* rules 57, 61 *)
	  List.fold
	  (strbinds, (decs, E), fn ({name, def, constraint}, (decs, E)) =>
	   let val (decs, E') = elaborateStrexp(def, B, decs)
	   in (decs,
	       Env.extendStrid
	       (E, name, elaborateSigexpConstraint(constraint, B) E'))
	   end)
     | Strdec.Local(d, d') => (* rule 58 *)
	  let val (decs, E') = elaborateStrdec(d, B, decs, Basis.toEnv B)
	  in elaborateStrdec(d', Basis.replaceEnv(B, E'), decs, E)
	  end
     | Strdec.Seq ds => (* rule 60 *)
	  let
	     val (_, decs, E) =
		List.fold
		(ds, (B, decs, E), fn (d, (B, decs, E)) =>
		 let val (decs, E') = elaborateStrdec(d, B, decs, Env.empty)
		 in (Basis.plusEnv(B, E'), decs, Env.+(E, E'))
		 end)
	  in (decs, E)
	  end
       ) arg

and elaborateStrexp(e: Strexp.t, B: Basis.t, decs): Decs.t * Env.t =
   case Strexp.node e of
      Strexp.Var p => (* rule 51 *)
	 (decs, Basis.lookupLongstrid(B, p))
    | Strexp.Struct d => (* rule 50 *)
	 elaborateStrdec(d, B, decs, Env.empty)
    | Strexp.Constrained(e, c) => (* rules 52, 53 *)
	 let val (decs, E) = elaborateStrexp(e, B, decs)
	 in (decs, elaborateSigexpConstraint(c, B) E)
	 end
    | Strexp.App(fctid, strexp) => (* rules 54, 154 *)
	 let
	    val (decs, E) = elaborateStrexp(strexp, B, decs)
	    val FunctorClosure.T{arg, argSig, body, basis=B', counter} =
	       Basis.lookupFctid(B, fctid)
	 in IntRef.inc counter ;
	    elaborateStrexp
	    (body, 
	     Basis.extendStrid(B', arg,
			       Env.cut{env = E,
				       interface = argSig,
				       opaque = false,
				       region = Strexp.region strexp}),
	     decs)
	 end
    | Strexp.Let(d, e) => (* rule 55 *)
	 let val (decs, E) = elaborateStrdec(d, B, decs, Basis.toEnv B)
	 in elaborateStrexp(e, Basis.replaceEnv(B, E), decs)
	 end

fun elaborateTopdec(d: Topdec.t, B: Basis.t, decs, B': Basis.t) =
   case Topdec.node d of
      Topdec.Strdec d =>
	 let val (decs, E) = elaborateStrdec(d, B, decs, Basis.toEnv B')
	 in (decs, Basis.replaceEnv(B', E))
	 end
    | Topdec.Signature sigbinds =>
	 (decs,
	  List.fold(sigbinds, B', fn ((sigid, sigexp), B') =>
		    Basis.extendSigid(B', sigid, elaborateSigexp(sigexp, B))))
    | Topdec.Functor funbinds =>
	 (* Appendix A, p.58 *)
	 (decs,
	  List.fold(funbinds, B', fn ({name, arg, result, body}, B') =>
		    let val body = Strexp.constrained(body, result)
		       val (arg, argSig, body) =
			  case FctArg.node arg of
			     FctArg.Structure(arg, argSig) => (arg, argSig, body)
			   | FctArg.Spec spec =>
				let val strid = Strid.fromString "ZZZNewStridZZZ"
				in (strid,
				    Sigexp.spec spec,
				    Strexp.lett
				    (Strdec.openn[Longstrid.short strid],
				     body))
				end
		    in Basis.extendFctid
		       (B', name,
			FunctorClosure.T{arg = arg,
					 argSig = elaborateSigexp(argSig, B),
					 body = body,
					 basis = B,
					 counter = newCounter()})
		    end))

(* val elaborateTopdec =
 *    Trace.trace4("elaborateTopdec",
 * 		Topdec.layout,
 * 		Layout.ignore (*Basis.layout*),
 * 		Layout.ignore,
 * 		Layout.ignore (*Basis.layout*),
 * 		Layout.ignore (*Layout.tuple2(Decs.layout, Layout.ignore)*))
 *    elaborateTopdec
 *)

structure Accum =
   struct
      datatype t = T of {decs: Decs.t,
			 basis: Basis.t}

      val empty = T{decs = Decs.empty,
		    basis = Basis.empty}

      fun prim() = T{decs = Decs.empty,
		     basis = #1(Basis.prim())}
	 
      fun decs(T{decs, ...}) = decs
      fun basis(T{basis, ...}) = basis

      val layoutDecs = Decs.layout o decs

      val layout = Basis.layout o basis
   end

	  
fun elaborateProgram(Ast.Program.T decs, B: Basis.t,
		     Accum.T{decs = decs', basis = B'}) =
   let
      val (B, decs, B') =
	 List.fold(decs, (B, decs', B'), fn (d, (B, decs, B')) =>
		   let val (decs, B'') = elaborateTopdec(d, B, decs, Basis.empty)
		   in (Basis.+(B, B''), decs, Basis.+(B', B''))
		   end)
   in (B,
       Accum.T{decs = decs,
	       basis = B'})
   end

end
