(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature INTERFACE_STRUCTS = 
   sig
      structure Ast: AST
      structure CoreML: CORE_ML
      structure TypeStr: TYPE_STR
      sharing Ast = CoreML.Ast = TypeStr.Ast
      sharing CoreML.Tycon = TypeStr.Tycon
      sharing CoreML.Con = TypeStr.Con
      sharing CoreML.Type = TypeStr.Type
      sharing CoreML.Scheme = TypeStr.Scheme
   end

signature INTERFACE = 
   sig
      include INTERFACE_STRUCTS

      structure ShapeId: UNIQUE_ID

      structure LazyTypeStr :
	 sig
	    include TYPE_STR
            val fromStrict: TypeStr.t -> t
	 end
      sharing Ast = LazyTypeStr.Ast

      structure Status :
	 sig
	    datatype t = Var | Con | Exn
	 end

      type interface
      type t = interface
	 
      structure Element :
	 sig
	    datatype t =
	       Val of {name: Ast.Vid.t,
		       scheme: CoreML.Scheme.t,
		       status: Status.t}
	     | Type of {name: Ast.Tycon.t,
			typeStr: TypeStr.t}
	     | Str of {name: Ast.Strid.t,
		       interface: interface}
	 end
      
      (* constructors *)
      val empty: t
      val vals: {name: Ast.Vid.t,
		  scheme: LazyTypeStr.Scheme.t,
		  status: Status.t} list -> t
      val types: {name: Ast.Tycon.t,
		   typeStr: LazyTypeStr.t} list -> t
      val strs: {name: Ast.Strid.t,
		  interface: t} list -> t
      val extendTycon: t * Ast.Tycon.t * LazyTypeStr.t -> t
      val cons: LazyTypeStr.Cons.t -> t
      val excons: LazyTypeStr.Cons.t -> t

      val shapeId: t -> ShapeId.t
	 
      val +: t * t -> t
	 
      val lookupLongtycon: t * Ast.Longtycon.t -> LazyTypeStr.t
      val peekLongtycon: t * Ast.Longtycon.t -> LazyTypeStr.t option

      val realize: t * (Ast.Longtycon.t -> TypeStr.TypeFcn.t) -> t
	 
      (* elements(I, f).
       f is called on the flexible tycons to get the realization *)
      val fold: t * 'a * (Element.t * 'a -> 'a) -> 'a
	 
      (* lazy.  rename all flexible tycons. *)
      val copy: t -> t
	 
      val shareType: t * Ast.Longtycon.t * Ast.Longtycon.t -> unit
      val share: t * Ast.Longstrid.t * Ast.Longstrid.t -> unit
      val wheres: t * (Ast.Longtycon.t * LazyTypeStr.TypeFcn.t) list -> unit
	 
      (* call when all copying is done in order to gc forward pointers *)
      val completelyForce: t -> unit
	 
      (* same up to names of flexible tycons *)
      val sameShape: t * t -> bool
	 
      val layout: t -> Layout.t
   end
