(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor TypeStr(S: TYPE_STR_STRUCTS): TYPE_STR = 
struct

open S

local open Ast
in structure Tyvar = Tyvar
end

structure TypeFcn =
   struct
      datatype t =
	 Tycon of Tycon.t
       | Fun of Scheme.t

      val isFun =
	 fn Fun _ => true
	  | _ => false

      val isTycon =
	 fn Tycon _ => true
	  | _ => false
	 
      val equals = fn _ => Error.unimplemented "TypeFcn.equals"
	 
      fun layout t =
	 let open Layout
	 in case t of
	    Tycon c => Tycon.layout c
	  | Fun s => Scheme.layout s
	 end

      fun funn(s as Scheme.T{tyvars, ty}): t =
	 case ty of
	    Type.Con(tycon, tys) =>
	       if List.equals(tyvars, tys,
			 fn (a, Type.Var a') => Tyvar.equals(a, a')
			  | _ => false)
		  then Tycon tycon
	       else Fun s
	  | _ => Fun s

      fun apply(f: t, ts) =
	 case f of
	    Tycon c => Type.Con(c, ts)
	  | Fun s => Scheme.apply(s, ts)
   end

structure Cons =
   struct
      datatype t = T of {name: Ast.Con.t,
			 con: Con.t,
			 scheme: Scheme.t} list
	 
      val empty = T []

      fun isEmpty(T l) = List.isEmpty l

      val isEmpty =
	 fn T[] => true
	  | _ => false
	 
      fun toList(T l) = l
	 
      fun layout(T l) =
	 let open Layout
	 in List.layout(fn {name, con, scheme} =>
			record[("name", Ast.Con.layout name),
			       ("con", Con.layout con),
			       ("scheme", Scheme.layout scheme)]) l
	 end
   end

datatype t =
   Fun of TypeFcn.t
 | Datatype of {tycon: Tycon.t, cons: Cons.t}

val bogus =
   Fun(TypeFcn.Fun
       (Scheme.T{tyvars = [],
		 ty = Type.Var(Ast.Tyvar.newNoname{equality = false})}))
   
fun make(typeFcn, cons) =
   if Cons.isEmpty cons
      then Fun typeFcn
   else case typeFcn of
      TypeFcn.Tycon tycon =>
	 Datatype{tycon = tycon, cons = cons}
    | TypeFcn.Fun _ => Error.error "TypeStr.make: non-well-formed typeStr"

val dest = 
   fn Fun typeFcn => (typeFcn, Cons.empty)
    | Datatype{tycon, cons} => (TypeFcn.Tycon tycon, cons)

val typeFcn = #1 o dest
val cons = #2 o dest
   
fun def s = Fun(TypeFcn.funn s)
val fromTypeFcn = Fun
fun data(tycon, cons) = Datatype{tycon = tycon, cons = cons}

val tycon = Fun o TypeFcn.Tycon

fun layout typeStr =
   let open Layout
   in case typeStr of
      Fun f => TypeFcn.layout f
    | Datatype{tycon, cons} =>
	 record[("tycon", Tycon.layout tycon),
		("cons", Cons.layout cons)]
   end

val equals = fn _ => Error.unimplemented "TypeStructure.equals"

end
