(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature TYPE_STR_STRUCTS = 
   sig
      structure Ast: AST
      structure Tycon :
	 sig
	    type t
	    val newNoname: unit -> t
	    val exn: t
	    val layout: t -> Layout.t
	 end
      structure Con :
	 sig
	    type t
	    val fromAst: Ast.Con.t -> t
	    val layout: t -> Layout.t
	 end
      structure Type :
	 sig
	    datatype t =
	       Var of Ast.Tyvar.t
	     | Con of Tycon.t * t list
	     | Record of t Ast.SortedRecord.t
	    val var: Ast.Tyvar.t -> t
	    val record: t Ast.SortedRecord.t -> t
	    val arrow: t * t -> t
(* 	    val hom: t * {var: Ast.Tyvar.t -> 'a,
 * 			  con: Tycon.t * 'a list -> 'a,
 * 			  record: 'a Ast.SortedRecord.t -> 'a} -> 'a
 *)
	 end
      structure Scheme :
	 sig
	    datatype t = T of {tyvars: Ast.Tyvar.t list, ty: Type.t}
	    val ty: t -> Type.t
	    val apply: t * Type.t list -> Type.t
	    val layout: t -> Layout.t
	 end
   end

signature TYPE_STR = 
   sig
      include TYPE_STR_STRUCTS
      
      structure TypeFcn :
	 sig
	    datatype t =
	       Tycon of Tycon.t
	     | Fun of Scheme.t
	       
	    val funn: Scheme.t -> t

	    val isFun: t -> bool
	    val isTycon: t -> bool

	    val apply: t * Type.t list -> Type.t

	    val equals: t * t -> bool
	    val layout: t -> Layout.t
	 end

      structure Cons :
	 sig
	    datatype t = T of {name: Ast.Con.t,
			       con: Con.t,
			       scheme: Scheme.t} list
	       
	    val empty: t
	    val isEmpty: t -> bool
	    val toList: t -> {name: Ast.Con.t,
			      con: Con.t,
			      scheme: Scheme.t} list
	    val layout: t -> Layout.t
	 end

      type t

      (* constructors *)
      val bogus: t
      val make: TypeFcn.t * Cons.t -> t
      val def: Scheme.t -> t
      val fromTypeFcn: TypeFcn.t -> t
      val data: Tycon.t * Cons.t -> t
      val tycon: Tycon.t -> t
	 
      (* destructors *)
      val dest: t -> TypeFcn.t * Cons.t
      val typeFcn: t -> TypeFcn.t
      val cons: t -> Cons.t
	 
      val equals: t * t -> bool
      val layout: t -> Layout.t
   end
