functor Array1D(A: BASIC_ARRAY_1D): ARRAY_1D =
struct

open Int A
structure Index = Int
structure Shape = Int

local
   structure L = LazyListUnmemo(Int)

   fun convert a =
      let val maxIndex = A.shape a
      in L.iterate(Int.zero, fn i =>
		   if Int.equals(i, maxIndex) then L.Empty
		   else L.First(sub(a, i), Int.add1 i))
      end
   
   structure S =
      ConvertToSequence(structure S = L
			type 'a t = 'a A.t
			val convert = convert)
in
   open S
end

val size = A.shape
val length = size
val nth = sub

fun foldi(a, b, f) =
   Int.fold(0, size a - 1, b, fn (i, b) => f(i, sub(a, i), b))

fun foreach(a, f) = fold(a, (), f o #1)

fun foldr(a, b, f) =
   Int.foldDown(size a - 1, 0, b, fn (i, b) => f(sub(a, i), b))
   
fun copy a = tabulate(size a, fn i => sub(a, i))

fun toList a = foldr(a, List.empty(), List.cons)
   
fun swap(a, i, j) =
   let val t = sub(a, i)
   in (update(a, i, sub(a, j)) ;
       update(a, j, t))
   end

fun maxIndex a = sub1(size a)

fun foreachDown(a, f) =
   Int.forDown(maxIndex a, zero, fn i => f(i, sub(a, i)))
   
fun shuffle a =
   foreachDown(a, fn (i, _) => swap(a, i, randomRange(zero, i)()))

fun getAndSet a = (fn i => sub(a, i),
		   fn (i, x) => update(a, i, x))
   
end

structure Array1D = Array1D(BasicArray1D())
structure Array = Array1D
