(*-------------------------------------------------------------------*)
(*                              Array2D                              *)
(*-------------------------------------------------------------------*)

structure Array2D =
   struct
      structure Shape = Pair(structure X = Int
			     structure Y = Int)
      structure Index = Shape
	 
      datatype 'a t =
	 T of {rows: int,
	       columns: int,
	       elts: 'a Array.t}

      fun unspot(columns, i:int) =
	 let val c = i mod columns
	    val r = (i - c) div columns
	 in (r, c)
	 end
      
      fun spot(T{columns, ...}, r, c) = r * columns + c

      fun computeSize(r, c) =
	 if r < 0 orelse c < 0 then Error.error "Array2D.size"
	 else r * c
	    
      fun new(r, c, x) =
	 T{rows = r, columns = c, elts = Array.new(computeSize(r, c), x)}

      fun tabulate(r, c, f) =
	 T{rows = r, columns = c,
	   elts = Array.tabulate
	   (computeSize(r, c), fn i => f(unspot(c, i)))}

      fun shape(T{rows, columns, ...}) = (rows, columns)

      fun elts(T{elts, ...}) = elts
	 
      fun sub(a, r, c) = Array.sub(elts a, spot(a, r, c))

      fun update(a, r, c, x) = Array.update(elts a, spot(a, r, c), x)

      fun copy(T{rows, columns, elts}) =
	 T{rows = rows, columns = columns,
	   elts = Array.copy elts}

      fun foldi(a as T{rows, columns, elts}, b, f) =
	 Int.fold(0, rows - 1, b, fn (r, b) =>
		  Int.fold(0, columns - 1, b, fn (c, b) =>
			   f(r, c, sub(a, r, c), b)))

      fun foreachi(a as T{rows, columns, elts}, f) =
	 Int.foreach(0, rows - 1, fn r =>
		     Int.foreach(0, columns - 1, fn c =>
				 f(r, c, sub(a, r, c))))
   end

(*
functor Array2D(structure X: INDEX_SHAPE
		structure Y: INDEX_SHAPE
		structure A: ARRAY_1D
		sharing X.I = Y.I = A.I): ARRAY =
struct

structure I = A.I
structure Iter = Iterate(I)
structure IS = PairIndexShape(structure X = X
			      structure Y = Y)
structure A = Array(structure IS = IS
		    structure A = A)

open A

structure XS = X.Shape
structure YS = Y.Shape
   
fun output(a, outputElt, out) =
   let fun outputAt(r, c) = outputElt(sub(a, (r, c)), out)
      val (xs, ys) = shape a
      val xa = XS.all xs
      val SOME(y0, ya) = YS.all ys
   in XS.S.foreach(xa, fn x =>
		   (outputAt(x, y0) ;
		    YS.S.foreach(ya, fn y =>
				 (Out.output(out, " ") ;
				  outputAt(x, y))) ;
		    Out.newline out))
   end 

end 
*)
