(*-------------------------------------------------------------------*)
(*                              Array                                *)
(*-------------------------------------------------------------------*)

functor Array(structure IS: INDEX_SHAPE
	      structure A: ARRAY_1D
	      sharing IS.I = A.I): ARRAY =
struct

structure Index = IS.Index
structure Shape = IS.Shape
structure L = List

datatype 'a t = T of {shape: Shape.t,
		      array: 'a A.t}

fun shape(T{shape, ...}) = shape
fun array(T{array, ...}) = array
   
exception New = Size

fun new(s, x) = T{shape = s,
		 array = A.new(Shape.size s, x)}

fun tabulate(s, f) = T{shape = s,
		      array = A.tabulate(Shape.size s, fn i =>
					 f(IS.intTo(i, s)))}

fun sub(T{array, shape}, i) = A.sub(array, IS.toInt(i, shape))

fun update(T{array, shape}, i, x) =
   A.update(array, IS.toInt(i, shape), x)

fun swap(T{array=a, shape=s}, i, j) =
   A.swap(a, IS.toInt(i, s), IS.toInt(j, s))

fun foreach(a, f) = A.foreach(array a, f)

fun foreachIndex(T{shape=s, array=a}, f) =
   A.foreachIndex(a, fn (i, x) => f(IS.intTo(i, s), x))

fun shuffle(T{array=a, ...}) = A.shuffle a

fun getAndSet(T{array=a, shape=s}) =
   let val {get, set} = A.getAndSet a
   in {get = fn i => get(IS.toInt(i, s)),
       set = fn (i, x) => set(IS.toInt(i, s), x)}
   end

fun output(T{array=a, shape=s}, outputElt, out) =
   A.output(a, fn(i, x, out) => outputElt(IS.intTo(i, s), x, out), out)

end
