(*-------------------------------------------------------------------*)
(*                            Bool Array                             *)
(*-------------------------------------------------------------------*)

functor BoolArray(I: INTEGER): BOOL_ARRAY =
struct

val {error, ...} = Error.errors("array", "bool")

structure I = I   
structure A = ByteArray
structure B = Byte
   
datatype t = T of {size: int,
		   array: A.bytearray}

fun array(T{array, ...}) = array
fun size(T{size, ...}) = size

val byteSize = 8

fun computeSize size = (size + byteSize - 1) div byteSize

exception New = Size

fun new(size, b) =
   let val size = I.toInt size
   in T{size = size,
	array = A.array(computeSize size,
			if b then B.allOnes else B.allZeros)}
   end
       
fun index(T{size, ...}, i) =
   let val i = I.toInt i
   in if 0 <= i andalso i < size
	 then let val byteNum = i div byteSize
		  val spot = i mod byteSize
	      in (byteNum, spot)
	      end
      else error "index"
   end

fun sub(a as T{array, ...}, i) =
   let val (i, s) = index(a, i)
   in B.isBitSet(A.sub(array, i), s)
   end

fun update(a as T{array, ...}, i, b) =
   let val (i, s) = index(a, i)
      val oper = if b then B.setBit else B.clearBit
   in A.update(array, i, oper(A.sub(array, i), s))
   end

fun set(a, i) = update(a, i, true)
fun clear(a, i) = update(a, i, false)

end

