(*-------------------------------------------------------------------*)
(*                             ArrayNest                             *)
(*-------------------------------------------------------------------*)

functor ArrayNest(structure A: ARRAY
		  structure A': ARRAY): ARRAY =
struct

val {error, unimplemented, ...} = Error.errors("array", "nest")

structure Shape = Pair(structure X = A.Shape
		       structure Y = A'.Shape)
structure Index = Pair(structure X = A.Index
		       structure Y = A'.Index)

datatype 'a t = T of {shape: Shape.t,
		      array: 'a A'.t A.t}

fun shape(T{shape, ...}) = shape
fun array(T{array, ...}) = array

exception New

fun new((s, s'), x) =
   T{shape = (s, s'),
     array = A.tabulate(s, fn _ => A'.new(s', x))}
   handle A.New => raise New
	| A'.New => raise New
	| e => raise e

fun tabulate((s, s'), f) =
   T{shape = (s, s'),
     array = A.tabulate(s, fn i =>
			A'.tabulate(s', fn j => f(i, j)))}
   handle A.New => raise New
	| A'.New => raise New
	| e => raise e

fun sub(a, (i, j)) = A'.sub(A.sub(array a, i), j)

fun update(a, (i, j), x) = A'.update(A.sub(array a, i), j, x)

fun swap(a, (i, j), (i', j')) =
   let val a = array a
      val b = A.sub(a, i)
      val b' = A.sub(a, i')
      val t = A'.sub(b, j)
   in (A'.update(b, j, A'.sub(b', j')) ;
       A'.update(b', j', t))
   end

fun foreach(a, f) = A.foreach(array a, fn (i, a') =>
			     A'.foreach(a', fn (j, x) => f((i, j), x)))

fun shuffle _ = unimplemented "shuffle"

fun getAndSet a = {get = fn i => sub(a, i),
		   set = fn (i, x) => update(a, i, x)}
   
fun output(a, outputElt, out) =
   A.output(array a,
	    fn (i, a', out) =>
	    (A'.output(a', fn (j, x, out) => outputElt((i, j), x, out),
		       out) ;
	     Out.output(out, "\n")),
	    out)
   
end
