(*-------------------------------------------------------------------*)
(*                           Packed Array                            *)
(*-------------------------------------------------------------------*)

functor PackedArray(P: PACKED_BITS): PACKED_ARRAY =
struct

structure A = BoolArray
structure Pack = P
structure Unpack =
   struct
      datatype t = T of {offset: int,
			 array: A.t}
      val size = P.size
      fun get(T{offset, array}, bit) = A.sub(array, offset + bit)
   end
   
datatype t = T of {size: int,
		   array: A.t}

fun array(T{array, ...}) = array
fun size(T{size, ...}) = size

fun offset i = P.size * i

fun sub(a, i) = Unpack.T{offset = offset i, array = array a}

fun update(a, i, x) =
   let val offset = offset i
   in Iterate.for(0, P.size - 1,
		  fn bit =>
		  A.update(array a, offset + bit, P.get(x, bit)))
   end

exception New = Size
fun new(size, f) =
   let val a = T{size = size,
		 array = A.new(P.size * size, true)}
   in (Iterate.for(0, size-1, fn i => update(a, i, f i)) ;
       a)
   end

fun listTo l =
   let val lr = ref l
   in new(List.length l,
	  fn _ => Ref.pop lr)
   end

end
