(*-------------------------------------------------------------------*)
(*                             RowMajor                              *)
(*-------------------------------------------------------------------*)

functor RowMajor(structure A: ARRAY_1D
		 structure R: ARRAY
		 structure C: ARRAY
		 sharing A.I = R.Shape.I = C.Shape.I): ARRAY = 
struct

val {error, ...} = Error.errors("array", "row-major")
   
structure I = A.I

structure Shape =
   struct
      structure P = Pair(structure X = R.Shape
			 structure Y = C.Shape)
      open P

      structure I = I

      fun size(r, c) = I.mul(R.Shape.size r, C.Shape.size c)
      val convert = size
   end

structure Index =
   struct
      structure P = Pair(structure X = R.Index
			 structure Y = C.Index)
      open P
      structure I = I
      structure S = Shape

      fun toInt((r, c), (r', c')) = I.add(I.mul(R.Index.toInt(r, r'),
					     C.Shape.size c'),
				       C.Index.toInt(c, c'))
      val from = toInt
      fun intTo(n, (r, c)) = let val m = C.Shape.size c
			   in (R.Index.intTo(I.divide(n, m), r),
			       C.Index.intTo(I.modulo(n, m), c))
			   end
      val to = intTo
   end

structure A = ConvertArray(structure A = A
			   structure Shape = Shape
			   structure Index = Index)
   
open A

structure Iter = Iterate(I)
   
fun output(a, outputElt: Index.t * '1a * Outstream.t -> unit, out) =
   let val (_, c) = shape a
      val sub1NumCols = I.sub1(C.Shape.size c)
      fun isEndOfRow ((_, c):Index.t) = I.eq(c, sub1NumCols)
   in A.output(a, fn (i, x, out) =>
	       (outputElt(i, x, out) ;
		if isEndOfRow i then Out.newline out else ()),
	       out)
   end
(*
   
   let val shape as (r, c) = shape a
      val numCols = C.Shape.size c
      val a = array a
   in 
      Iter.for
      (I.zero, I.sub1(Shape.size shape), fn n =>
       outputElt(A.sub(a,n), out) ;
       if I.divides(
		  
      Iter.for
      (I.zero, I.sub1(R.Shape.size r), fn r =>
       (Iter.for(I.zero, I.sub1(C.Shape.size c), fn c =>
		 (if I.isPositive c
		     then Out.output(out, " ")
		  else () ;
		  outputElt(sub(a, (r, c)), out))) ;
	Out.output(out, "\n")))
   end
		    *)

end
