(*-------------------------------------------------------------------*)
(*                             Instream                              *)
(*-------------------------------------------------------------------*)

structure Instream: INSTREAM =
struct

structure String = String0
structure Char = String.Char
structure I = Pervasive.IO
open I
   
type t = I.instream

val standard = stdIn
val close = closeIn
val inputChar = input1
val peekChar = lookahead   
val endOf = endOfStream

(*fun inputAll i =
   let
      fun loop accum = 
	 if endOf i then accum
	 else (input i) :: accum
   in String.concat(rev (loop []))
   end
*)
fun outputAll(i, out) =
   let
      fun loop() = if endOf i then ()
		   else (Out.output(out, input i) ; loop())
   in loop()
   end

fun inputTo(i, p) =
   let
      fun loop accum =
	 case peekChar i of
	    NONE => accum
	  | SOME c => if p c then accum else (inputChar i ; loop(c :: accum))
   in String.implode(rev(loop []))
   end

fun sameContents(in1, in2) =
   let
      fun loop() =
	 case (input1 in1, input1 in2) of
	    (NONE, NONE) => true
	  | (SOME c1, SOME c2) => Char.equals(c1, c2) andalso loop()
	  | _ => false
   in loop()
   end

fun inputToSpace i = inputTo(i, Char.isSpace)
fun inputToChar(i, c) = inputTo(i, fn c' => Char.equals(c, c'))
fun ignoreSpaces i = (inputTo(i,not o Char.isSpace) ; ())

(*
fun duplicate i n =
   let val fis = LU.duplicate n I.openChanIn
      val fs = LU.map fis #1
      val is = LU.map fis #2
      fun dup() = (case inputChar i of
		      NONE => LU.foreach is close
		    | SOME c => let val s = C.toString c
				in (LU.foreach fs (fn f => f s) ;
				    dup())
				end)
   in (CML.spawn dup ; is)
   end
*)
fun inputNothing _ = ()

fun layout _ = Layout.str "<instream>"
   
end

structure In = Instream
