signature LAYOUT =
   sig
      type t

      val empty: t

      val isEmpty: t -> bool
	 
      (* convert a string to a layout object *)
      val str: string -> t

      (* layout the objects on the same line *)
      val seq: t list -> t

      (* layout the objects on separate lines*)
      val align: t list -> t

      (* layout the objects on separate lines, if necessary *)
      val mayAlign: t list -> t

      (* indent the entire object *)
      val indent: t * int -> t

      (* put string between elements *)
      val separate: t list * string -> t list

      (* adds string at beginning of all objects except first *)
      val separateLeft: t list * string -> t list

      (* adds string at the end of all objects except last *) 
      val separateRight: t list * string -> t list

      val alignPrefix: string * t list -> t

      val ignore: 'a -> t
	 
      (* semi-standard layouts *)
      val option: ('a -> t) -> 'a option -> t
      val tuple2: ('a -> t) * ('b -> t) -> 'a * 'b -> t
      val tuple3: ('a -> t) * ('b -> t) * ('c -> t) -> 'a * 'b * 'c -> t
      val tuple4: ('a -> t) * ('b -> t) * ('c -> t) * ('d -> t)
	 -> 'a * 'b * 'c * 'd -> t
      val list: t list -> t
      val tuple: t list -> t
      val vector: t list -> t
      val paren: t -> t
      val record: (string * t) list -> t
      val namedRecord: string * (string * t) list -> t

      val toString: t -> string
	 
      (* print the object *)
      val print: t * (string -> unit) -> unit
      val outputWidth: t * int * Out.t -> unit
      val output: t * Out.t -> unit
      val outputTree: t * Out.t -> unit

      val makeOutput: ('a -> t) -> 'a * Out.t -> unit

      (* whether or not to print things in detail -
       * routines that create layouts should use this flag to decide
       * how detailed to print
       *)
      val detailed: bool ref
      val switch: {detailed: 'a -> t,
		    normal: 'a -> t} -> 'a -> t
   end
