signature STRING0 =
   sig
      structure Char: CHAR0

      include ORDER0 where type t = string

      val empty: t
      val newline: t
	 
      val isEmpty: t -> bool

      val contains: t * Char.t -> bool
		    
      val fromChar: Char.t -> t
      val toChar: t -> Char.t

      val make: int * Char.t -> t
      val last: t -> Char.t

      (* beginning at start, with length chars *)
      val substring1: t * {start: int, length: int} -> t

      (* inclusive of start, exclusive of finish *)
      val substring2: t * {start: int, finish: int} -> t

      val isSubstring: {string: string, substring: string} -> bool
	 
      val prefix: t * int -> t
      val suffix: t * int -> t
      val dropPrefix: t * int -> t
      val dropSuffix: t * int -> t
      val dropLast: t -> t

      val isPrefix: {string: t, prefix: t} -> bool
      val isSuffix: {string: t, suffix: t} -> bool

      (* splits the string into substrings broken at char,
       e.g. split("foo$bar$baz", #"$") = ["foo", "bar", "baz"] *)
      val split: t * Char.t -> t list
	 
      val output: t * TextIO.outstream -> unit

      (* Basis Library *)
      val maxSize: int
      val size: t -> int
      val sub: t * int -> Char.t
      val concat: t list -> t
      val ^ : t * t -> t
      val implode: Char.t list -> t
      val explode: t -> Char.t list
      val tokens: (Char.t -> bool) -> t -> t list 
      val fields: (Char.t -> bool) -> t -> t list 
      val translate: (Char.t -> string) -> string -> string

      val fromString: t -> t option
      val toString: t -> t
      val fromCString: t -> t option
      val toCString: t -> t
   end
