(*-------------------------------------------------------------------*)
(*                             Dijkstra                              *)
(*-------------------------------------------------------------------*)

functor Dijkstra (structure Graph: DIRECTED_GRAPH
		  structure Weight: WEIGHT): SHORTEST_PATH =
struct

structure Graph = Graph
structure Node = Graph.Node
structure Edge = Graph.Edge
structure Weight = Weight
structure Heap = FibonacciHeap(Weight)
structure Heap = Heap
structure Elt = Heap.E
structure O = Option

fun shortestPath {graph, weight, source} =
   let
      val {get = eltOption, set = setElt} =
	 Graph.nodeInfo(graph, fn _ => NONE)
      val elt = O.projector eltOption
      val distanceOption = (O.optionize Elt.key) o eltOption
      val distance = O.projector distanceOption 
      val fringe = Heap.empty()
      fun addToFringe(n, d) = setElt(n, SOME (Heap.insert(fringe, d,n)))
      fun relax e =
	 let val (n,n') = Edge.nodes e
	    val d = Weight.+(distance n, weight e)
	 in case distanceOption n' of
	    NONE => addToFringe(n', d)
	  | SOME d' => if Weight.<(d, d')
			   then Heap.decreaseKey(fringe, elt n', d)
		       else ()
	 end
   in (addToFringe(source, Weight.zero) ;
       Iterate.whileDo
       (fn () => not(Heap.isEmpty fringe),
	fn () => List.foreach(Graph.adjacent(graph,
					     Heap.deleteMin fringe),
			      relax)) ;
       distanceOption)
   end

end
