signature DIRECTED_GRAPH_STRUCTS = 
   sig
   end

signature DIRECTED_GRAPH = 
   sig
      include DIRECTED_GRAPH_STRUCTS
      
      structure Node :
	 sig
	    include T

	    val successors: t -> t list
	    val plist: t -> PropertyList.t
	 end
      
      structure Edge :
	 sig
	    type t = {from: Node.t, to: Node.t}
	 end

      (* the main graph type *)
      type t

      val new: unit -> t
      val newNode: t -> Node.t
      val addEdge: t * Edge.t -> unit

      (* error if edge isn't there *)
      val removeEdge: t * Edge.t -> unit

      val nodes: t -> Node.t list

      val foreachNode: t * (Node.t -> unit) -> unit
	       
      val foreachEdge: t * (Edge.t -> unit) -> unit

(*      val isCyclic: t -> bool*)

(*      val random: {numNodes: int, numEdges: int} -> t*)

      (* depth first search *)
	 
      structure DfsParam :
	 sig
	    type t = {startNode: Node.t -> unit,
		      finishNode: Node.t -> unit,
		      handleTreeEdge: Edge.t -> unit,
		      handleNonTreeEdge: Edge.t -> unit,
		      startTree: Node.t -> unit,
		      finishTree: Node.t -> unit,
		      finishDfs: unit -> unit}
	    val finishNode: (Node.t -> unit) -> t
	    val ignore: 'a -> unit
	    val combine: t * t -> t
	 end

      val dfsNodes: t * Node.t list * DfsParam.t -> unit
      val dfs: t * DfsParam.t -> unit

      val removeBackEdges: t -> unit

      val foreachDescendent: t * Node.t * (Node.t -> unit) -> unit

      val discoverFinishTimes: t -> {discover: Node.t -> int,
				     finish: Node.t -> int,
				     destroy: unit -> unit,
				     param: DfsParam.t}

      val display:
	 {graph: t,
	  layoutNode: Node.t -> Layout.t,
	  display: Layout.t -> unit} -> unit

      (* strongly-connected components.
       * Each component is given as a list of nodes.
       * The components are returned topologically sorted.
       *)
      val stronglyConnectedComponents: t -> Node.t list list
	 
      exception TopologicalSort
      val topologicalSort: t -> Node.t list
(*
      val transpose: t -> t
*)
      (* input/output *)
(*
      exception Input
      val input: In.t * (In.t -> 'a)* (In.t -> 'b)
	  -> t * 'a * (Edge.t -> 'b)

      val output: t * Out.t -> unit*)
   end
