(*-------------------------------------------------------------------*)
(*                        Shortest Path Check                        *)
(*-------------------------------------------------------------------*)

functor ShortestPathCheck(S: SHORTEST_PATH_CHECK_STRUCTS) :
   SHORTEST_PATH_CHECK =
struct

open S
structure Node = Graph.Node
structure Edge = Graph.Edge

structure S = DisjointSet

structure Answer =
   struct      
      datatype t =
	 Shortest
       | SourceNonZero
       | PredecessorReachable of Edge.t
       | Relaxable of Edge.t
       | NoPath of Node.t
	 
      fun toString Shortest =
	 "The distances are correct shortest path distances.\n"
       | toString SourceNonZero =
	 "The distance to the source must be zero.\n"
	| toString (PredecessorReachable e) =
	  let val (n,n') = Edge.nodes e
	  in String.concat
	     ["The distances are contradictory.\n",
	      "1. There is an edge from ", Node.toString n, " to ", Node.toString n',
	      "\n2. ", Node.toString n, " has a finite distancEdge.",
	      "\n3. ", Node.toString n', " has an infinite distancEdge.\n"]
	  end
	| toString (NoPath n) =
	     String.concat
	     ["There is not a valid predecessor path from ",
	      Node.toString n, " to the sourcEdge.\n"]
	| toString (Relaxable e) =
	  let val (n,n') = Edge.nodes e
	  in String.concat
	     ["The distances are not shortest path distances.\n",
	      "The edge from ", Node.toString n, " to ", Node.toString n',
	      " can be relaxed.\n"]
	  end

      fun output(a, out) = out(toString a)
   end
structure A = Answer
	 
exception Answer of Answer.t

fun check {graph, source, weight, distance} =
   case distance source of
      NONE => A.SourceNonZero
    | SOME d => 
	  if not (Weight.equals(Weight.zero, d)) then A.SourceNonZero
	  else
	  let val {get=set, ...} = Graph.nodeInfo(graph, S.singleton)
	      fun union(n,n') = S.union(set n, set n')
	      fun checkRelax e =
		  let val (n,n') = Edge.nodes e
		  in case distance n of
		      NONE => () 
		    | SOME d =>
		      case distance n' of
			  NONE => raise Answer (A.PredecessorReachable e)
			| SOME d' =>
			      let val d'' = Weight.+(d, weight e)
			      in if Weight.<(d'', d')
				     then raise Answer (A.Relaxable e)
				 else if Weight.equals(d', d'')
					  then union(n,n')
				      else ()
			      end
		  end
	      in (Graph.foreachEdge(graph, checkRelax) ;
		  let val sourceSet = set source
		     fun canReachSource n =
			let val equiv = S.equals(set n, sourceSet)
			in case distance n of
			   NONE => ()
			 | SOME _ =>
			      if equiv then () else raise Answer (A.NoPath n)
			end
		  in Graph.foreachNode(graph, canReachSource) end ;
	       A.Shortest)
	      end handle Answer a => a
end
