functor Env(S: ENV_STRUCTS): ENV =
struct

open S

datatype 'a t = T of (Domain.t * 'a) List.t

fun size(T l) = List.length l
   
fun eq d d' = Domain.equals(d, d')
fun member(d, ds) = List.exists(ds, eq d)
   
fun invariant(T drs) =
   let fun loop([], _) = true
	 | loop((d, _) :: drs, seen) =
	      not(member(d, seen)) andalso loop(drs, d :: seen)
   in loop(drs, [])
   end

fun domain(T drs) = List.map(drs, #1)
   
val fromList = T
fun toList(T l) = l
   
fun empty() = T(List.empty())

fun single(d, r) = T[(d, r)]

fun isEmpty(T l) = List.isEmpty l

fun singleton dr = T(List.single dr)

fun new(ds, f) = T(List.map(ds, fn d => (d, f d)))

fun map(T drs, f) = T(List.map(drs, fn (d, r) => (d, f r)))
fun mapi(T drs, f) = T(List.map(drs, fn (d, r) => (d, f(d, r))))

fun fold(T drs, b, f) = List.fold(drs, b, fn ((_, r), b) => f(r, b))
fun foldi(T drs, b, f) = List.fold(drs, b, fn ((d, r), b) => f(d, r, b))
   
fun equal d (d', _) = Domain.equals(d, d')

fun remove(T drs, d) = T(List.remove(drs, equal d))
   
fun extend(T drs, d, r) =
   T(List.cons((d, r), List.remove(drs, equal d)))

fun env + (T l) = List.fold(l, env, fn ((d, r), env) => extend(env, d, r))

fun plus es = List.fold(es, empty(), fn (e, accum) => accum + e)
   
fun peek(T l, d) =
   case List.peek(l, equal d) of
      NONE => NONE
    | SOME (_, r) => SOME r
fun lookup(env, d) = case peek(env, d) of
   SOME r => r
 | NONE => (Layout.output(Domain.layout d, Out.error) ;
	    Out.newline Out.error ;
	    Error.error "lookup")

fun restrict(env, ds) = new(ds, fn d => lookup(env, d))
(* is this faster?
fun restrict(T srs, ss) =
	 T(let fun loop(_, []) = []
		 | loop([], _) = Error.error "didn't find everything"
		 | loop((s, r) :: srs, ss) =
		   let fun loop'([], _) = loop(srs, ss)
			 | loop'(s' :: ss', ss) =
			      if Domain.equals(s, s')
				 then (s, r) :: (loop(srs, ss'@ss))
			      else loop'(ss', s' :: ss)
		   in loop'(ss, [])
		   end
	   in loop(srs, ss)
	   end)
*)

fun multiExtend(env, [], []) = env
  | multiExtend(env, d :: ds, r :: rs) = multiExtend(extend(env, d, r), ds, rs)
  | multiExtend _ = Error.error "multiExtend"

fun foreach(e, f) = List.foreach(toList e, f o #2)
fun foreachi(e, f) = List.foreach(toList e, f)

fun forall(e, f) = List.forall(toList e, f o #2)
fun foralli(e, f) = List.forall(toList e, f)
   
fun equals rangeEqual (e1, e2) =
   Int.equals(size e1, size e2)
   andalso foralli(e1, fn (d, r) =>
		   case peek(e2, d) of
		      NONE => false
		    | SOME r' => rangeEqual(r, r'))
   
fun layout layoutR (T ps) =
   let open Layout
   in seq[str "[",
	  align(List.map(ps, fn (d, r) =>
			 seq[Domain.layout d, str " -> ", layoutR r])),
	  str"]"]
   end

fun maybeLayout (name, layoutR) env =
   if isEmpty env then Layout.empty
   else let open Layout
	in seq[str name, str " = ", layout layoutR env]
	end

end

functor PolyEnv(S: ENV_STRUCTS): ENV = Env(S)
