functor AppendList(): APPEND_LIST =
struct

datatype 'a t =
   Base of 'a list
 | Append of 'a t list

val empty = Base[]
fun single x = Base[x]
val fromList = Base
fun append(t1, t2) = Append[t1, t2]
val appends = Append
fun cons(x, l) = append(single x, l)

fun fold(l, b, f) =
   let
      fun loop(l, b) =
	 case l of
	    Base l => List.fold(l, b, f)
	  | Append l => List.fold(l, b, loop)
   in loop(l, b)
   end

fun foreach(l, f) = fold(l, (), fn (x, ()) => f x)

fun foldr(l, b, f) =
   let
      fun loop(l, b) =
	 case l of
	    Base l => List.foldr(l, b, f)
	  | Append l => List.foldr(l, b, loop)
   in loop(l, b)
   end

fun map(l, f) =
   let
      val rec loop =
	 fn Base l => Base(List.map(l, f))
	  | Append l => Append(List.map(l, loop))
   in loop l
   end

fun toList ds =
   let fun loop(Base xs, rest) = List.append(xs, rest)
	 | loop(Append ts, rest) = List.foldr(ts, rest, loop)
   in loop(ds, [])
   end

local open Layout
in fun layout (sep, layoutX) xs =
   let fun loop(Base xs) = seq[List.layoutSep (sep, layoutX) xs, str sep]
	 | loop(Append ts) = seq(List.map(ts, loop))
   in loop xs
   end
end

end

structure AppendList = AppendList()
