(*-------------------------------------------------------------------*)
(*                           BasicLazyList                           *)
(*-------------------------------------------------------------------*)

functor BasicLazyList(P: PROMISE): BASIC_LAZY_LIST =
struct

datatype 'a t = T of ('a * 'a t) option P.t
   
fun destruct(T p) = P.force p
      
datatype ('a, 'b) iter =
   Empty
  | First of 'a * 'b
  | Rest of 'a t
       
fun empty() = T(P.delay(fn () => NONE))

fun lazyCons th = T(P.delay th)

fun cons(x, l) = T(P.delay(fn () => SOME(x, l)))
   
fun iterate(b, f) =
   let
      fun iter b =
	 T(P.delay(fn () =>
		 case f b of
		    Empty => NONE
		  | First(a, b) => SOME(a, iter b)
		  | Rest s => destruct s))
   in iter b
   end

end
