(*-------------------------------------------------------------------*)
(*                             LazyList                              *)
(*-------------------------------------------------------------------*)

functor LazyList(structure L: BASIC_LAZY_LIST): LAZY_LIST =
struct

structure L' = List(structure L = L)
   
open Int L L'
   
fun appendReverse(l, r) =
   appendThunk(l, fn () => reverse r)

fun appendReverseIncremental(l, l') =
   iterate((l, l', empty()), fn (l, l', a) =>
	   case (destruct l, destruct l') of
		(NONE, NONE) => Rest(empty())
	      | (NONE, SOME(x, _)) => Rest(cons(x, a))
	      | (SOME(x, l), NONE) => Rest(cons(x, append(l, a)))
	      | (SOME(x, l), SOME(x', l')) => First(x, (l, l', cons(x', a))))
   
fun appendAppendReverse c =
   let fun rot(L, R, A) =
	 let fun finish() = append(L, append(reverse R, A))
	 in case destruct L of
	    NONE => finish()
	  | SOME(x, L) =>
	       case splitAtMost(R, c) of
		  NONE => finish()
		| SOME(R, R') =>
		     lazyCons(fn () =>
			      SOME(x,
				   rot(L, R', append(reverse R, A))))
	 end
   in rot
   end
      
fun appendReverseNthTail c =
   let val aar = appendAppendReverse c
      fun rot(L, R,n) = 
	 if n >= c
	    then L.lazyCons(fn () =>
			    SOME(head L,
				 rot(tail L,
				     nthTail(R, c),
				     n - c)))
	 else aar(L, nthTail(R,n), empty())
   in rot
   end

end

functor StandardLazyList(P: PROMISE) =
   LazyList(structure L = BasicLazyList(P)
	    structure I = Int)

functor LazyListUnmemo(I: INTEGER) =
   LazyList(structure L = BasicLazyList(PromiseUnmemo)
	    structure I = I)

structure LazyListUnmemo = StandardLazyList(PromiseUnmemo)

structure LazyList = StandardLazyList(Promise)
