structure ListSet :>
   sig
      type 'a t = 'a list

      val new: {equal: 'a * 'a -> bool,
		 layout: 'a -> Layout.t}
	 -> {empty: 'a t,
	     isEmpty: 'a t -> bool,
	     foreach: 'a t * ('a -> unit) -> unit,
	     contains: 'a t * 'a -> bool,
	     <=: 'a t * 'a t -> bool,
	     add: 'a t * 'a -> 'a t,
	     subset: 'a t * ('a -> bool) -> 'a t,
	     layout: 'a t -> Layout.t,
	     union: 'a t * 'a t -> 'a t,
	     intersect: 'a t * 'a t -> 'a t,
	     difference: 'a t * 'a t -> 'a t,
	     equal: 'a t * 'a t -> bool}
   end =
   struct
      open List

      fun new{equal, layout} =
	 let
	    fun contains(vs, v) = exists(vs, fn v' => equal(v, v'))
	    fun vs <= vs' = forall(vs, fn v => contains(vs', v))
	    fun equal(vs, vs') = vs <= vs' andalso vs' <= vs
	    fun add(vs, v) = if contains(vs, v) then vs else cons(v, vs)
	    fun union(vs, vs') = fold(vs', vs, fn (v, vs) => add(vs, v))
	    val subset = keepAll
	    fun intersect(vs, vs') = subset(vs, fn v => contains(vs', v))
	    fun difference(s, s') = subset(s, fn x => not(contains(s', x)))
	    local open Layout
	    in
	       val layout = fn vs =>
		  seq[str "{",
		      align(separateRight(toList(map(vs, layout)), ", ")),
		      str "}"]
	    end
	 in {empty = empty(),
	     isEmpty = isEmpty,
	     foreach = foreach,
	     contains = contains,
	     <= = op <=,
	     add = add,
	     subset = subset,
	     layout = layout,
	     union = union,
	     intersect = intersect,
	     difference = difference,
	     equal = equal}
	 end
   end
