signature LIST =
   sig
      include DESTRUCTABLE_SEQUENCE

      (* include BASIC_LIST_ITER *)
      val empty: unit -> 'a t
      val cons: 'a * 'a t -> 'a t
      datatype ('a, 'b) iter =
	 Empty
       | First of 'a * 'b
       | Rest of 'a t
      val iterate: 'b * ('b -> ('a, 'b) iter) -> 'a t
      (* *)

      val head: 'a t -> 'a
      val tail: 'a t -> 'a t
      val nthTail: 'a t * Int.t -> 'a t
      val single: 'a -> 'a t
      val appendThunk: 'a t * (unit -> 'a t) -> 'a t
      val append: 'a t * 'a t -> 'a t
	 
      val flatten: 'a t t -> 'a t
	 
      val reverse: 'a t -> 'a t

      (* reverseAppend(l1, l2) = append(reverse l1, l2) *)
      val reverseAppend: 'a t * 'a t -> 'a t

      val zip: 'a t * 'b t -> ('a * 'b) t
      val zip3: 'a t * 'b t * 'c t -> ('a * 'b * 'c) t
      val unzip: ('a * 'b) t -> 'a t * 'b t
      val unzip3: ('a * 'b * 'c) t -> 'a t * 'b t * 'c t

      (* mapAppend(l1, f, l2) = append(map(l1, f), l2) *)
      val mapAppend: 'a t * ('a -> 'b) * 'b t -> 'b t
      val map: 'a t * ('a -> 'b) -> 'b t
      val mapi: 'a t * (int * 'a -> 'b) -> 'b t
      val map2: 'a t * 'b t * ('a * 'b -> 'c) -> 'c t
      val map3: 'a t * 'b t * 'c t * ('a * 'b * 'c -> 'd) -> 'd t
      val map4: 'a t * 'b t * 'c t * 'd t * ('a * 'b * 'c * 'd -> 'e) -> 'e t
      val tabulate: 'a t * (Int.t * 'a -> 'b) -> 'b t
      val partition: 'a t * ('a -> bool) -> {yes: 'a t, no: 'a t}
      val keepAll: 'a t * ('a -> bool) -> 'a t
      val keepAllMap: 'a t * ('a -> 'b option) -> 'b t
      val removeAll: 'a t * ('a -> bool) -> 'a t

      (* group according to equivalence relation *)
      val equivalence: 'a t * ('a * 'a -> bool) -> 'a t t

      val separate: 'a t * 'a -> 'a t

      (* removes firsts element if it exists
       returns NONE if pred is false on everything *)
      val peekRemove: 'a t * ('a -> bool) -> ('a t * 'a) option
	 
      (* error if predicate isn't true on some element *)
      val removeFirst: 'a t * ('a -> bool) -> 'a t

      (* removes first element on which predicate is true.
       returns original list if predicate is false on every elt. *)
      val remove: 'a t * ('a -> bool) -> 'a t

      val splitLast: 'a t -> 'a t * 'a
      val allButLast: 'a t -> 'a t
	  
      val flattenMap: 'a t * ('a -> 'b t) -> 'b t
      val duplicate: Int.t * (unit -> 'a) -> 'a t
      val firstN: 'a t * Int.t -> 'a t
      val splitAtMost: 'a t * Int.t -> ('a t * 'a t) option

      (*   cross [[1, 2, 3], [4, 5], [6, 7]] =
       [[1, 4, 6], [1, 4, 7], [1, 5, 6], [1, 5, 7],
       [2, 4, 6], [2, 4, 7], [2, 5, 6], [2, 5, 7],
       [3, 4, 6], [3, 4, 7], [3, 5, 6], [3, 5, 7]] *)
      val cross: 'a t t -> 'a t t
	 
      val prefixes: 'a t -> 'a t t
      val suffixes: 'a t -> 'a t t
	 
      (* extract each element paired with the elements before it
       and after it in the same order as in the list *)
      val each: 'a t -> ('a t * 'a * 'a t) t

      val listTo: 'a list -> 'a t

      val compare: 'a t * 'a t * ('a * 'a -> order) -> order

      val ordered :
	 {< : 'a * 'a -> bool}
	 -> {insert: 'a t * 'a -> 'a t,
	     insertionSort: 'a t  -> 'a t,
	     median: 'a t -> 'a,
	     orderStatistic: 'a t * Int.t -> 'a,
	     partition: 'a t * 'a -> 'a t * 'a t,
	     max: 'a t -> 'a,
	     min: 'a t -> 'a,
	     largest: 'a t * Int.t -> 'a t,
	     smallest: 'a t * Int.t -> 'a t}

      val power: 'a t -> 'a t t
      val subsets: 'a t * int -> 'a t t
	 
      val set: {equal: 'a * 'a -> bool,
		 layout: 'a -> Layout.t}
	 -> {empty: 'a t,
	     singleton: 'a -> 'a t,
	     size: 'a t -> int,
	     equals: 'a t * 'a t -> bool,
	     <= : 'a t * 'a t -> bool,
	     >= : 'a t * 'a t -> bool,
	     < : 'a t * 'a t -> bool,
	     > : 'a t * 'a t -> bool,
	     + : 'a t * 'a t -> 'a t,
	     - : 'a t * 'a t -> 'a t,
	     intersect: 'a t * 'a t -> 'a t,
	     unions: 'a t t -> 'a t,
	     add: 'a t * 'a -> 'a t,
	     remove: 'a t * 'a -> 'a t,
	     contains: 'a t * 'a -> bool,
	     power: 'a t -> 'a t t,
	     subsets: 'a t * int -> 'a t t,
	     areDisjoint: 'a t * 'a t -> bool,
	     subset: 'a t * ('a -> bool) -> 'a t, (* aka keepAll *)
	     subsetSize: 'a t * ('a -> bool) -> int,
	     replace: 'a t * ('a -> 'a option) -> 'a t,
	     map: 'a t * ('a -> 'a) -> 'a t,
	     layout: 'a t -> Layout.t}
   end 
