functor Mark(): MARK =
struct

structure File = FileSys.File

datatype t =
   T of {string: string,
	 pos: int}

fun pos(T{pos, ...}) = pos

local
   val numChars = 15
in
   fun output(T{string, pos}, out) =
      String.output(String.substring1(string, {start = pos, length = numChars}),
		    out)
end

fun length(T{string, ...}) = String.length string
fun isAtBeginning m = pos m = 0
fun isAtEnd m = pos m = length m - 1
fun diff(m, m') = pos m - pos m'

fun fromString s = T{string = s, pos = 0}
fun fromFile f = fromString(File.contents f)

fun beginning(T{string, ...}) = T{string = string, pos = 0}
   
exception BackwardChars
fun backwardChars(T{string, pos},n) =
   let val pos = pos - n
   in if pos < 0 then raise BackwardChars
      else T{string = string, pos = pos}
   end
fun backwardChar m = backwardChars(m, 1)
val backwardChar =
   Trace.trace("backwardChar", output, output) backwardChar

exception ForwardChars
fun forwardChars(T{string, pos},n) =
   let val pos = pos + n
   in if pos > String.length string then raise ForwardChars
      else T{string = string, pos = pos}
   end
fun forwardChar m = forwardChars(m, 1)
val forwardChar =
   Trace.trace("forwardChar", output, output) forwardChar

fun charAt(T{string, pos}) = String.sub(string, pos)
   
fun lookingAtChar(m, c) = Char.equals(charAt m, c)
   
local
   fun searchChar move (m, c) =
      let
	 fun loop m =
	    if lookingAtChar(m, c) then m else loop(move m)
      in loop m
      end
in 
   val searchCharForward = searchChar forwardChar
   fun searchCharBackward(m, c) =
      searchChar backwardChar (backwardChar m, c)
end

fun bol m = forwardChar(searchCharBackward(m, #"\n"))
   handle _ => beginning m
      
fun eol m = searchCharForward(m, #"\n")

fun whatColumn m = diff(m, bol m)

fun numColumns m = diff(eol m, bol m)
   
local
   fun moveLines move =
      let fun moves (m as T{string, pos},n) =
	 let val c = whatColumn m
	    fun loop(m,n) =
	       if n = 0 then forwardChars(m, Int.min(c,numColumns m))
	       else loop(move m, n - 1)
	 in loop(bol m, n)
	 end
	  fun move m = moves(m, 1)
      in (move, moves)
      end
in
   val (previousLine, previousLines) = moveLines(bol o backwardChar)
   val (nextLine,nextLines) = moveLines(forwardChar o eol)
end

fun lookingAtString(T{string, pos}, string') =
   let
      val len = String.length string
      val len' = String.length string'
      fun loop(pos, pos') =
	 not(pos = len)
	 andalso (pos' = len'
		  orelse (Char.equals(String.sub(string, pos),
				 String.sub(string', pos'))
			  andalso loop(pos + 1, pos' + 1)))
   in loop(pos, 0)
   end
val lookingAtString =
   Trace.trace2("lookingAtString", output, String.output, Bool.output)
   lookingAtString

exception Search
fun makeSearch move (m, s) =
   let
      fun search m =
	 if lookingAtString(m, s)
	    then forwardChars(m, String.length s)
	 else (search(move m) handle _ => raise Search)
   in search m
   end

val search = makeSearch forwardChar
val searchBackward = makeSearch backwardChar

fun skip p =
   let fun skip m = if p(charAt m) then skip(forwardChar m)
		    else m
   in skip
   end

fun skipUntil p = skip(not o p)

val skipSpaces = skip Char.isSpace
val skipUntilSpace = skipUntil Char.isSpace

fun substring(T{string, pos}, T{pos=pos', ...}) =
   String.substring2(string, {start=pos, finish=pos'})

fun num (fromString, exn) m =
   let val m = skipSpaces m
      val m' = skipUntilSpace m
   in case fromString(substring(m, m')) of
      NONE => raise exn
    | SOME n => (m',n)
   end

exception Int
val int = num(Int.fromString, Int)

exception Real
val real = num(Real.fromString, Real)

val real =
   Trace.trace("real", output, Out.output2(output, Real.output))
   real

val op < = fn (m, m') => pos m < pos m'

val equals = fn (m, m') => pos m = pos m'
   
val {>, >=, <=, min, max, compare} =
   Relation.lessEqual{< = op <, equal = op =}

end

structure Mark = Mark()
