structure MyDirs: MY_DIRS =
   struct
      open Dir
	 
      structure Host =
	 struct
	    datatype t = Eponym | Samson | Snake
	       
	    fun current() = 
	       case OS.Process.getEnv "HOST" of
		  NONE => Error.error "can't figure out host"
		| SOME s =>
		     case s of
			"eponym.nj.nec.com" => Eponym
		      | "samson.nj.nec.com" => Samson
		      | "snake" => Snake
		      | "snake.nj.nec.com" => Snake
		      | _ => Error.error "unrecognized host"
	 end
      open Host
	 
      fun dirs() =
	 let
	    val home =
	       fromString(case current() of
			     Eponym => "/home/sweeks/"
			   | Snake => "/home/sweeks/"
			   | Samson => "/usr/tmp/sweeks/")
	    val sml = relative(home, "sml")
	    val smlnj = relative(home, "sml/sml-nj-root")
	    val bin = relative(smlnj, "bin")
	    val binFiles = relative(smlnj, "bin.x86-unix")
	    val heap = relative(bin, ".heap")
	    val src = relative(smlnj, "src")
	    val compiler = relative(src, "sml-nj")
	 in {home = home, sml = sml, smlnj = smlnj, bin = bin,
	     binFiles = binFiles, heap = heap, src = src, compiler = compiler}
	 end
      
      fun exportFn(name, f) =
	 SMLofNJ.exportFn(File.toString(File.relative(#heap(dirs()),name)),
			  f)
      fun exportML name =
	 SMLofNJ.exportML(File.toString(File.relative(#heap(dirs()),name)))
   end
