(*-------------------------------------------------------------------*)
(*                                Tab                                *)
(*-------------------------------------------------------------------*)

functor Tab(): TAB =
struct

structure A = Array1D
   
val initialSize = 40
val growFactor = 2

fun make(out, tab) =
   let
      fun makeTabs size =
	 A.tabulate(size,
		    let val prev = ref ""
		    in fn _ =>
		       (let val cur = !prev
			in (prev := tab ^ cur ;
			    cur)
			end)
		    end)

      val tabs = ref(makeTabs initialSize)

      fun size() = A.size(!tabs)

      fun resize() = tabs := makeTabs(growFactor * size())

      val index = ref 0

      fun maybeResize() = if !index = size() then resize() else ()

      fun reset() = index := 0
	 
      fun indent() = Out.output(out, A.sub(!tabs, !index))

      fun right() = (index := !index + 1 ;
		     maybeResize())

      fun left() = if !index = 0 then Error.error "Tab.left"
		   else index := !index - 1
		      
      fun output x =
	 (indent() ;
	  Out.output(out, x) ;
	  Out.output(out, "\n"))
	 
   in {reset = reset,
       indent = indent,
       left = left,
       right = right}
   end
				    
end

structure Tab = Tab()
