(*-------------------------------------------------------------------*)
(*                           FixedInteger                            *)
(*-------------------------------------------------------------------*)

structure FixedInteger: FIXED_INTEGER =
struct

structure I =
   struct
      structure I = Pervasive.Int
      open I
	 
      type t = int
	 
      val zero = 0
      val one = 1
	 
      fun fromInt n = n
      fun toInt n = n
      val toReal = Pervasive.Real.fromInt
      val fromString = I.fromString
      val toString = I.toString
      val layout = Layout.str o toString
	     
      local
	 structure R = Random
	 val seeds = (13, 14)
	 val r = R.rand seeds
      in
	 fun randomRange(lo, hi) =
	    let val f = R.randRange (lo, hi)
	    in fn () => f r
	    end
      end

      val equals = op =
	 
      val {compare, ...} = Relation.lessEqual{< = op <, equals = equals}

      fun output(i, out) = Out.output(out, toString i)

      exception Input
      fun input i = (In.ignoreSpaces i ;
		     case fromString(In.inputToSpace i) of
			NONE => raise Input
		      | SOME n => n)
	    
      val maxInt = valOf maxInt
      val minInt = valOf minInt
   end

structure I' = Integer(I)
open I I'

val random = randomRange(minInt, maxInt)

end

structure Integer = FixedInteger
structure Int = Integer
