(*-------------------------------------------------------------------*)
(*                             FloatReal                             *)
(*-------------------------------------------------------------------*)

structure FloatReal: FLOAT_REAL =
struct

structure R =
   struct
      structure I = Int
	 
      open Pervasive.Real
      open Math
	 
      type t = real
	 
      val zero = 0.0
      val one = 1.0

      val ceiling = ceil
	 
      fun inverse x = 1.0/x
	 
      val equals = Pervasive.Real.==
(*      val equals = fn _ => Error.unimplemented "FloatReal.equals" *)
      val {compare, ...} = Relation.lessEqual{< = op <, equals = equals}

      val layout = Layout.str o toString
      fun output(x, out) = Out.output(out, toString x)

      exception Input 
      fun input i =
	 case fromString(In.inputToSpace i) of
	    SOME x => x
	  | NONE => raise Input
	    
      local
	 structure R = Random
	 val seeds = (13, 13)
	 val r = R.rand seeds
      in
	 fun random(lo, hi) = lo + ((hi - lo) * (R.randReal r))
      end

      val fromInt = real
      fun fromReal x = x
      fun toReal x = x

      structure Format =
	 struct
	    open StringCvt
	    type t = realfmt
	    val sci = SCI
	    val fix = FIX
	    val gen = GEN
	 end
      fun format(x, f) = Pervasive.Real.fmt f x
   end

structure R' = Real(R)
open R R'

end

structure Real = FloatReal
