(*-------------------------------------------------------------------*)
(*                              Integer                              *)
(*-------------------------------------------------------------------*)

functor Integer(I: BASIC_INTEGER): INTEGER =
struct

structure R = OrderedRing(I)
open I R

fun toCommaString n =
   let
      fun loop(chars, accum) =
	 case chars of
	    x1 :: x2 :: x3 :: (chars as _ :: _) =>
	       loop(chars, #"," :: x3 :: x2 :: x1 :: accum)
	  | _ => implode(rev chars @ accum)
   in loop(rev(explode(toString n)), [])
   end

fun choose(n, k) =
   let val k = max(k,n-k)
   in (prodFromTo(add1 k,n)) div (factorial(n-k))
   end

fun output(n, out) = Out.output(out, toString n)
   
fun isEven n = isZero(n mod two)
val isOdd = not o isEven

val (op ^) = fn (base, exp) =>
   let fun power exp = if isZero exp then one
		       else let val n = square(power(quot(exp, two)))
			    in if isEven exp then n
			       else base*n
			    end
   in power exp
   end
   
fun powerMod{base, exp, modulus} =
   let val m = modulus
      fun loop n =
	 if isZero n then one
	 else let val r = square(loop(n div two)) mod m
	      in if isEven n then r else (base*r) mod m
	      end
   in loop exp
   end			 

fun gcd(a, b) = if isZero b then a else gcd(b, a mod b)

fun lcm(a, b) = (a*b) div (gcd(a, b))

fun inputBetween{ins, error, min, max} =
   let
      fun loop() =
	 let fun continue() = (error() ; loop())
	 in let val n = input ins
	    in if min <= n andalso n <= max
		  then n
	       else continue()
	    end handle Input => continue()
	 end
   in loop()
   end

fun fromTo({start, stop, step}, s,nextS) =
    let val cmp = if isPositive step
		     then (op >)
		  else (op <)
    in #1(Iterate.iterate((s, start),
		  fn (_, i) => cmp(i, stop),
		  fn (s, i) => (nextS(i, s), i + step)))
    end

fun fold(start, stop, s,nextS) =
   fromTo({start = start, stop = stop, step = one}, s, nextS)

fun foldDown(start, stop, s,nextS) =
   fromTo({start = start,
	   stop = stop,
	   step = negOne},
	  s,
	  nextS)

fun forall(m,n, f) =
   let
      fun forall m = m > n
	 orelse (f m andalso forall(add1 m))
   in forall m
   end

fun forStep(sss, f) = fromTo(sss, (), fn (k, ()) => f k)
   
fun foreach(start, stop, f) = forStep({start = start,
				       stop = stop,
				       step = one},
				    f)
   
fun forDown(start, stop, f) =
   forStep({start = start,
	    stop = stop,
	    step = negOne},
	   f)

end

