functor Property(Plist: PROPERTY_LIST): PROPERTY =
struct

structure Plist = Plist
   
datatype ('sym, 'val) init =
   Const of 'val
 | Fun of 'sym -> 'val
 | Raise of string * ('sym -> Layout.t)
 | Rec of 'sym * ('sym -> 'val) -> 'val

val initConst = Const
val initFun = Fun
val initRaise = Raise
val initRec = Rec
   
fun print s = TextIO.output(TextIO.stdErr, s)

fun ('sym, 'val) make(plist: 'sym -> Plist.t,
		      i: ('sym, 'val) init,
		      destroy) =
   let
      val {add, peek, set = set', remove, ...} = Plist.newProperty()

      val (new, destroy) = destroy remove

      fun set(x: 'sym, v: 'val): unit =
	 let val pl = plist x
	 in set'(pl, v, fn () => new pl)
	 end

      fun get(x: 'sym): 'val =
	 let val pl = plist x
	    fun make v = (new pl; add(pl, v); v)
	 in case peek(plist x) of
	    SOME v => v
	  | NONE =>
	       case i of
		  Const v => v
		| Raise(name, layout) =>
		     Error.bug
		     (let open Layout
		      in toString(seq[layout x,
				      str " has no ", str name, str " property"])
		      end)
		| Fun f => make(f x)
		| Rec f => make(f(x, get))
	 end
			 
   in {get = get, set = set, destroy = destroy}
   end

fun new(plist, init) =
   let
      val {get, set, ...} =
	 make(plist, init,
	      fn _ =>
	      (fn _ => (),
	       fn () => Error.bug "attempt to destroy non-destructible property"))
   in {get = get, set = set}
   end

fun newDest(plist, init) =
   make(plist, init,
	fn remove => let val plists = ref []
		     in (fn plist => ListRef.push(plists, plist),
			 fn () => (List.foreach(!plists, remove);
				   plists := []))
		     end)

(* fun ('sym1, 'sym2, 'val) pair
 *    {property: ('sym1, ('sym2 * 'val) list) t,
 *     equals: 'sym2 * 'sym2 -> bool}: ('sym1 * 'sym2, 'val) t =
 *    fn i =>
 *    let
 *       fun equalTo b (b', _) = equals(b, b')
 *       val p = property(initConst [])
 * 	 
 *       fun get(s1: 'sym1, s2: 'sym2): 'val =
 * 	 let val alist = #get p s1
 * 	 in case List.peek(alist, equalTo s2) of
 * 	    NONE => init((s1, s2), i, get, 
 * 			 fn v => #set p (s1, (s2, v) :: alist))
 * 	  | SOME(_, v) => v
 * 	 end
 * 
 *       fun set((s1: 'sym1, s2: 'sym2), v: 'val): unit =
 * 	 #set p
 * 	 (s1, 
 * 	  let
 * 	     fun loop(s2vs: ('sym2 * 'val) list): ('sym2 * 'val) list =
 * 		case s2vs of
 * 		   [] => [(s2, v)]
 * 		 | ((s2v as (s2', _)) :: s2vs) =>
 * 		      if equals(s2, s2') then (s2, v) :: s2vs
 * 		      else s2v :: loop s2vs
 * 	  in loop(#get p s1)
 * 	  end)
 * 
 *    in {get = get,
 *        set = set,
 *        destroy = #destroy p}
 *    end
 *)
end
