(*  PROPERTY
 *  hides the underlying property and property lists and gives just
 *  the functions for manipulating "symbols" and "values"
 *  (as they are called in LISP property lists.
 * 
 *  To associate arbitrary information with a data structure, add an extra field
 *  to the data structure that holds a property list.  Then call 'wrap', passing
 *  the field selector and a property, and it will return the property
 *  operations that act directly on the data structure.   Or, call 'wrapNew' and
 *  it will create a new property and return the same operations.
 * 
 *  'pair' associates arbitrary information with pairs of values,
 *  where the first kind of value holds a property list and the second has an
 *  equality function.  'pair' stores an alist based on the equality function
 *  in the property.
 *)
							
signature PROPERTY =
   sig
      structure Plist: PROPERTY_LIST

      type ('sym, 'val) init

      val initConst: 'val -> ('sym, 'val) init
      val initFun: ('sym -> 'val) -> ('sym, 'val) init
      val initRec: ('sym * ('sym -> 'val) -> 'val) -> ('sym, 'val) init
      val initRaise: string * ('sym -> Layout.t) -> ('sym, 'val) init
	 
      val new:
	 ('sym -> Plist.t) * ('sym, 'val) init
	 -> {
	     get: 'sym -> 'val,
	     set: 'sym * 'val -> unit
	    }

      val newDest:
	 ('sym -> Plist.t) * ('sym, 'val) init
	 -> {
	     get: 'sym -> 'val,
	     set: 'sym * 'val -> unit,
	     destroy: unit -> unit
	    }

      (* associate a property with pairs of symbols
       * uses the property of the first and an alist of the second
       *)
(*       val pair :
 * 	 {property: ('sym1, ('sym2 * 'val) list) t,
 * 	  equals: 'sym2 * 'sym2 -> bool}
 * 	 -> ('sym1 * 'sym2, 'val) t
 *)
   end
