signature TRACE_CONTROL = 
   sig
      (* controls what tracing info is gathered *)
      val always: unit -> unit      (* always gather info *)
      val flagged: unit -> unit     (* only on flagged functions *)
      val never: unit -> unit       (* never gather info *)
	 
      (* value of newly created flag *)
      val default: bool ref

      (* turn on/off tracing for given functions *)
      val on: string list -> unit
      val off: string list -> unit

      (* turn on/off tracing for all functions *)
      val all: unit -> unit
      val none: unit -> unit

      (* returns functions that are on/off *)
      val whatsOn: unit -> string list
      val whatsOff: unit -> string list
   end

signature TRACE =
   sig
      structure Immediate :
	 sig
	    include TRACE_CONTROL

	    val message: Layout.t -> unit
	    val messageStr: string -> unit
	 end
      structure Delayed :
	 sig
	    include TRACE_CONTROL
	    val keepAll: bool ref
	 end
      structure Time: TRACE_CONTROL

      val never: unit -> unit
      val always: unit -> unit
      val flagged: unit -> unit
      val traceable: unit -> string list
      val outputTraceable: unit -> unit
      val reset: unit -> unit
	 
      (*---------- Delayed Feedback ----------*)
      structure Computation: COMPUTATION

      (* clear computation history *)
      val clear: unit -> unit

      (* get computation history *)
      val computation: unit -> Computation.t

      (* show computation history *)
      val history: unit -> unit

      (* show computation history, without arguments or results *)
      val calls: unit -> unit

      val times: unit -> unit

      (* enter the inspector *)
      val inspect: unit -> unit

      (*---------- Instrumentation ----------*)

      val traceCall: string -> ('a -> 'b) -> 'a -> 'b
	 
      val trace: string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 -> ('a -> 'b)
	 -> ('a -> 'b)

      val trace0: string * ('a -> Layout.t) -> (unit -> 'a) -> (unit -> 'a)

      val trace2: string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 * ('c -> Layout.t)
	 -> ('a * 'b -> 'c)
	 -> ('a * 'b -> 'c)

      val trace3: string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 * ('c -> Layout.t)
	 * ('d -> Layout.t)
	 -> ('a * 'b * 'c -> 'd)
	 -> ('a * 'b * 'c -> 'd)

      val trace4: string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 * ('c -> Layout.t)
	 * ('d -> Layout.t)
	 * ('e -> Layout.t)
	 -> ('a * 'b * 'c * 'd -> 'e)
	 -> ('a * 'b * 'c * 'd -> 'e)

      val traceRec :string
	 * ('a -> Layout.t)
	 * ('b -> Layout.t)
	 -> (('a -> 'b) -> ('a -> 'b))
	 -> 'a -> 'b 
   end
