(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor PrimEnv(S: PRIM_ENV_STRUCTS): PRIM_ENV = 
struct

open S

local open CoreML
in structure Constructors = Constructors
end
structure Env = Basis.Env

fun preprocessFile(f: File.t, B, A: Accum.t): Basis.t * Accum.t =
   (Control.inputFile := f;
    File.withIn(f, fn ins => preprocess(ins, B, A)))

fun preprocessFiles(fs: File.t list, B, A: Accum.t): Basis.t * Accum.t =
   List.fold(fs, (B, A), fn (f, (B, A)) => preprocessFile(f, B, A))

val root = ref(Dir.fromString "..")
   
fun lazy(name: string, thunk: unit -> 'a): unit -> 'a =
   Promise.lazy
   (fn () =>
    let val a = thunk()
    in Control.message(fn () => Control.sizeMessage(name, a));
       a
    end)

fun basisFile f =
   File.relative(Dir.relative(!root, "basis-library"), f)

fun basis() =
   Ref.fluidLet
   (Control.aux, false, fn () =>
    let
       fun files(f, B, A) =
	  preprocessFiles(List.map(List.keepAll(File.lines(basisFile f),
						fn s =>
						(s <> ""
						 andalso
						 #"#" <> String.sub(s, 0))),
				   basisFile),
			  B, A)
       val (B, cs) = Basis.prim()
       val (_, A) = files("build-basis", B, Accum.prim())
       val (_, A) =
	  files("bind-basis", Accum.basis A,
		Accum.T{decs = Accum.decs A,
			basis = Basis.fromEnv(Env.addEquals Env.empty)})
    in (A, cs)
    end)

(* val patchRecordTypes: Layout.t option ref = ref NONE
 * 
 * fun getPatchRecordTypes() =
 *    case !patchRecordTypes of
 *       NONE => Error.bug "getPatchRecordTypes"
 *     | SOME l => l
 *)

val defunctorize =
   lazy
   ("defunctorize", fn () =>
    let
       val (A, cs) =  Ref.fluidLet(useOpaque, true, basis)
       val B = Accum.basis A
	val (E, P, cs') = Env.identify(Basis.toEnv B)
(*	val _ = patchRecordTypes := SOME(Ast.Program.layout P) *)
	(* Uncomment the following to see the basis *)
(* 	val _ = File.withOut(File.fromString "/tmp/z",
 * 			     fn out => Layout.output(Env.layout E, out))
 *)
	val B = Basis.replaceEnv(B, E)
     in (Accum.T{decs = Decs.empty, basis = B},
	 Constructors.+(cs, cs'))
     end)

(* val monomorphise =
 *    lazy
 *    ("monomorphise",
 *     fn () =>
 *     let val (A, cs) = defunctorize()
 *     in (#2(preprocessFile(basisFile "top-level/overloads.sml", Accum.basis A, A)),
 * 	cs)
 *     end)
 *)

val compile =
   lazy
   ("compile env",
    fn () =>
    let val (A, cs) = basis()
       (* concatenated onto the end of the program for cleanup *)
       val (_, suffix) =
	  preprocessFiles([basisFile "misc/suffix.sml"],
			  Accum.basis A,
			  Accum.empty)
    in (A, Accum.decs suffix, cs)
    end)

fun forceAll d = (root := d
		  ; let open Control
		    in
		       Control.setMode Defunctorize; defunctorize()
(*		       ; Control.setMode Monomorphise; monomorphise() *)
		       ; Control.setMode Compile; compile()
		    end
		  ; ())
			       
end
