(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor NestedPat(S: NESTED_PAT_STRUCTS): NESTED_PAT = 
struct

open S

datatype t = T of {pat: node, ty: Type.t}
and node =
   Wild
  | Var of Var.t
  | Const of Const.t
  | Con of {con: Con.t, targs: Type.t list, arg: t option}
  | Tuple of t list
  | Layered of Var.t * t

fun node(T{pat, ...}) = pat
fun ty(T{ty, ...}) = ty
   
fun tuple[p] = p
  | tuple ps = T{pat = Tuple ps, ty = Type.tuple(List.map(ps, ty))}

local structure Pat = Ast.Pat
in fun toAst p =
   case node p of
      Wild => Pat.wild
    | Var x => Pat.var(Var.toAst x)
    | Const c => Const.toAstPat c
    | Con{con, arg, ...} =>
	 let val con = Con.toAst con
	 in case arg of
	    NONE => Pat.con con
	  | SOME p => Pat.app(con, toAst p)
	 end
    | Tuple ps => Pat.tuple(List.map(ps, toAst))
    | Layered(x, p) => Pat.layered{fixop = Ast.Fixop.None,
				  var = Var.toAst x,
				  constraint = NONE,
				  pat = toAst p}
end

val layout = Ast.Pat.layout o toAst

fun new(p, t) =
   case p of
      Tuple[p] => p
    | _ => T{pat = p, ty = t}

fun wild t = new(Wild, t)
   
fun isRefutable p =
   case node p of
      Wild => false
    | Var _ => false
    | Const _ => true
    | Con _ => true
    | Tuple ps => List.exists(ps, isRefutable)
    | Layered(_, p) => isRefutable p

val unit = T{pat = Tuple[], ty = Type.tuple[]}
   
fun varsAndTypes(p: t): (Var.t * Type.t) list =
   let
      fun loop(p: t, accum: (Var.t * Type.t) list) =
	 case node p of
	    Wild => accum
	  | Const _ => accum
	  | Var x => (x, ty p) :: accum
	  | Tuple ps => List.fold(ps, accum, loop)
	  | Con{arg, ...} => (case arg of
				NONE => accum
			      | SOME p => loop(p, accum))
	  | Layered(x, p) => loop(p, (x, ty p) :: accum)
   in loop(p, [])
   end

fun vars(p: t): Var.t list =
   List.map(varsAndTypes p, #1)

end
