(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
signature NESTED_PAT_STRUCTS = 
   sig
      include ATOMS
      structure Type :
	 sig
	    type t
	    val tuple: t list -> t
	 end
   end

signature NESTED_PAT = 
   sig
      include NESTED_PAT_STRUCTS
      
      datatype t = T of {pat: node, ty: Type.t}
      and node =
	 Wild
       | Var of Var.t
       | Const of Const.t
       | Con of {con: Con.t, targs: Type.t list, arg: t option}
       | Tuple of t list
       | Layered of Var.t * t

      (* constants *)
      val unit: t

      (* constructors *)
      val new: node * Type.t -> t
      val tuple: t list -> t
      val wild: Type.t -> t

      (* destructors *)
      val node: t -> node
      val ty: t -> Type.t

      (* true if the pattern contains a constant, constructor or variable *)
      val isRefutable: t -> bool

      (* a list of the variables in the pattern, along with their types.
       Used for match compilation in order to build a function that abstracts
       over the expression of a case rule p => e.
       See infer.fun *)
      val varsAndTypes: t -> (Var.t * Type.t) list
	 
      val layout: t -> Layout.t
   end
