(* Copyright (C) 1997-1999 NEC Research Institute.
 * Please see the file LICENSE for license information.
 *)
functor ImplementExceptions(S: IMPLEMENT_EXCEPTIONS_STRUCTS): IMPLEMENT_EXCEPTIONS = 
struct

open S
open Dec PrimExp
structure Dexp = DirectExp

fun doit(Program.T{datatypes, body}): Program.t =
   let
      val exnName = Var.newString "exnName"
      val exnNameType = Type.arrow(Type.exn, Type.string)
      val {get = exconInfo: Con.t -> {refVar: Var.t,
				      make: VarExp.t option -> Dexp.t} option,
	   set = setExconInfo} =
	 Property.new(Con.plist, Property.initConst NONE)

(*       val setExconInfo = Trace.trace2("setExconInfo", Con.layout,
 * 				      Layout.ignore, Unit.layout) setExconInfo
 * 
 *       val exconInfo =
 * 	 Trace.trace("exconInfo", Con.layout, Layout.ignore) exconInfo
 *)

      fun isExcon c =
	 case exconInfo c of
	    NONE => false
	  | SOME _ => true

      val exnValCons: {con: Con.t, arg: Type.t} list ref = ref []

      fun loopOpt e =
	 (case e of
	     NONE => NONE
	   | SOME e => SOME(loop e))

      and loops es = List.map(es, loop)

      and loop(e: Exp.t): Exp.t =
	 let val {decs, result} = Exp.dest e
	    val decs =
	       List.flatten
	       (List.reverse(List.fold(decs, [], fn (d, ds) => loopDec d :: ds)))
	 in Exp.new{decs = decs,
		    result = result}
	 end

      and loopDec(dec: Dec.t): Dec.t list =
	 case dec of
	    MonoVal b => loopMonoVal b
	  | Fun{decs, ...} =>
	       [Fun{tyvars = [],
		    decs = List.revMap(decs, fn {var, ty, lambda} =>
				       {var = var,
					ty = ty,
					lambda = loopLambda lambda})}]
	  | Exception{con, arg} =>
	       let open Dexp
		  val r = Var.newString "exnRef"
		  val uniq = monoVar(r, Type.unitRef)
		  val conApp =
		     fn arg =>
		     conApp{con = con, targs = [], ty = Type.exn, arg = SOME arg}
		  val (arg, decs, make) =
		     case arg of
			NONE =>
			   (* If the exception is not value carrying, then go
			    * ahead and make it now.
			    *)
			   let val exn = Var.newNoname()
			   in (Type.unitRef,
			       Dexp.vall{var = exn, exp = conApp uniq},
			       fn NONE => monoVar(exn, Type.exn)
				| _ => Error.bug "nullary excon applied to arg")
			   end
		      | SOME t =>
			   let val tupleType = Type.tuple[Type.unitRef, t]
			   in (tupleType,
			       [],
			       fn SOME x =>
			       conApp(tuple{exps = [uniq, varExp(x, t)],
					    ty = tupleType})
				| _ => Error.bug "unary excon not applied to arg")
			   end
	       in setExconInfo(con, SOME{refVar = r, make = make});
		  ListRef.push(exnValCons, {con = con, arg = arg});
		  vall{var = r, exp = reff(unit())} @ decs
	       end
	  | _ => Error.bug "implement exceptions saw unexpected dec"

      and loopMonoVal{var, ty, exp} : Dec.t list =
	 let fun primExp e = [MonoVal{var = var, ty = ty, exp = e}]
	    fun keep() = primExp exp
	    fun makeExp e = Dexp.vall{var = var, exp = e}
	 in case exp of
	    Lambda l => primExp(Lambda(loopLambda l))
	  | PrimApp{prim, targs, args} =>
	       primExp
	       (case (Prim.name prim, args) of
		   (Prim.Name.ExnName, [arg]) =>
		      App{func = VarExp.mono exnName, arg = arg}
		 | _ => PrimApp{prim = prim, targs = targs, args = args})
	  | ConApp{con, arg, ...} =>
	       (case exconInfo con of
		   NONE => keep()
		 | SOME{make, ...} => makeExp(make arg))
	  | Handle{try, catch = (catch, ty), handler} =>
	       primExp(Handle{try = loop try,
			      catch = (catch, ty),
			      handler = loop handler})
	  | Case{test, cases, default} =>
	       let
		  fun normal() =
		     primExp(Case
			     {test = test,
			      cases = List.revMap(cases, fn (pat, exp) =>
						  (pat, loop exp)),
			      default = loopOpt default})
	       in case cases of
		  [] => normal()
		| (Pat.T{con, ...}, _) :: _ =>
		     if not(isExcon con)
			then normal()
		     else (* convert to an exception match *)
			let
			   open Dexp
			   val defaultVar = Var.newString "default"
			   fun callDefault() =
			      app{func = monoVar(defaultVar,
						 Type.arrow(Type.unit, ty)),
				  arg = unit(),
				  ty = ty}
			   val unit = Var.newString "unit"
			in makeExp
			   (lett
			    {decs = vall{var = defaultVar,
					 exp =
					 lambda
					 {arg = unit,
					  argType = Type.unit,
					  bodyType = ty,
					  body =
					  case default of
					     SOME e => fromExp(loop e, ty)
					   | NONE =>
						Error.bug "no default for exception case"}},
			     
			     body = 
			     casee
			     {test = varExp(test, Type.exn),
			      ty = ty,
			      default = SOME(callDefault()),
			      cases =
			      List.revMap
			      (cases, fn (Pat.T{con, arg, ...}, e) =>
			       let
				  val refVar = Var.newNoname()
				  val body =
				     iff{test =
					 equal(monoVar(refVar, Type.unitRef),
					       monoVar
					       (#refVar(valOf(exconInfo con)),
						Type.unitRef)),
					 ty = ty,
					 thenn = fromExp(loop e, ty),
					 elsee = callDefault()}
				  fun make(arg, body) = 
				     (Pat.T{con = con, targs = [],
					    arg = SOME arg},
				      body)
			       in case arg of
				  NONE => make((refVar, Type.unitRef), body)
				| SOME(x, t) =>
				     let val tuple = (Var.newNoname(),
						      Type.tuple[Type.unitRef, t])
				     in make(tuple,
					     detupleBind
					     {tuple = monoVar tuple,
					      components = [refVar, x],
					      body = body})
				     end
			       end)}})
			end
	       end
	  | _ => keep()
	 end

      and loopLambda l =
	 let val {arg, argType, body} = Lambda.dest l
	 in Lambda.new{arg = arg,
		       argType = argType,
		       body = loop body}
	 end

      val body =
	 (* Wrap the body in a default handler that prints the exception name
	  * and exits with status 1.
	  *)
	 let val exn = (Var.newNoname(), Type.exn)
	 in Exp.fromPrimExp
	    (Handle
	     {try = loop body,
	      catch = exn,
	      handler =
	      let
		 open Dexp
		 fun string s = const(Const.fromString s)
		 fun print(e: Dexp.t): Dexp.t = 
		    primApp{prim = Prim.print,
			    targs = [],
			    args = [e],
			    ty = Type.unit}
	      in toExp
		 (sequence
		  [print(string "unhandled exception "),
		   print(app{func = monoVar(exnName, exnNameType),
			     arg = monoVar exn,
			     ty = Type.string}),
		   print(string "\n"),
		   primApp{prim = Prim.exit,
			   targs = [],
			   ty = Type.unit,
			   args = [const(Const.fromInt 1)]}])
	      end},
	     Type.unit)
	 end

      val (datatypes, body) =
	 case !exnValCons of
	    [] => (datatypes, body)
	  | cons =>
	       (* Add a datatype declaration for exceptions
		* and implement exnName.
		*)
	       ({tycon = Tycon.exn,
		 tyvars = [],
		 cons = List.revMap(cons, fn {con, arg} =>
				    {con = con, arg = SOME arg})}
		:: datatypes,
		Exp.prefix
		(body, 
		 MonoVal
		 {var = exnName,
		  ty = exnNameType,
		  exp =
		  Lambda
		  (Lambda.new
		   let val exn = Var.newNoname()
		   in {arg = exn,
		       argType = Type.exn,
		       body =
		       let open Dexp
		       in toExp
			  (casee
			   {test = monoVar(exn, Type.exn),
			    cases =
			    List.revMap
			    (cons, fn {con, arg} =>
			     (Pat.T{con = con,
				    targs = [],
				    arg = SOME(Var.newNoname(), arg)},
			      const
			      (Const.fromString(Con.originalName con)))),
			    default = NONE,
			    ty = Type.string})
		       end}
		   end)}))

      val program =
	 Program.T{datatypes = datatypes,
		   body = body}
	 
   in Program.clear program;
      program
   end
   
end
