
functor MLLrValsFun(structure Token : TOKEN
			     structure Ast : AST) = 
struct
structure ParserData=
struct
structure Header = 
struct
(* Heavily modified from SML/NJ sources by sweeks@research.nj.nec.com *)

(* ml.grm
 *
 * Copyright 1989,1992 by AT&T Bell Laboratories
 *)

fun reg(left,right) = Region.T{left = left, right = right}
fun error(reg,msg) = Control.error(reg, Layout.str msg)

open Ast
structure Field = Record.Field
structure Srecord = SortedRecord

structure Type =
   struct
      open Type
      val tuple = Record o Srecord.tuple
      val unit = tuple[]

      fun arrow(t1,t2) = Con(Longtycon.arrow, [t1,t2])
   end

structure Pat =
   struct
      open Pat
      val tuple =
	 fn [p] => node p
	  | ps => Tuple ps
      val unit = tuple[]

      val bogus = unit

      fun var id =
	 Var{name = Longvid.short(Vid.fromString id),
	     fixop = Fixop.None}

      fun makeAs(p1 : t, p2 : t) : node =
	 let
	    fun err() =
	       (error(Pat.region p1, "must have variable to left in as pattern") ;
		bogus)

	    fun fixopVar(p : t) =
	       case node p of
		  FlatApp[p] =>
		     (case node p of
			 Var{fixop,name} =>
			    (case Longvid.split name of
				([],vid) =>
				   SOME(fixop, Vid.toVar vid)
			      | _ =>
				   (err() ;
				    SOME(Fixop.None,Var.bogus)))
		       | _ => NONE)
		| _ => NONE
		     
	 in case fixopVar p1 of
	    SOME(fixop,var) =>
	       Layered{fixop = fixop, var = var,
		       constraint = NONE,
		       pat = p2}
	  | NONE =>
	       case node p1 of
		  Pat.Constraint(p,t) =>
		     (case fixopVar p of
			 SOME(fixop,var) =>
			    Layered{fixop = fixop, var = var,
				    constraint = SOME t,
				    pat = p2}
		       | _ => err())
		| _ => err()
	 end
   end

structure Exp =
   struct
      open Exp
      val tuple =
	 fn [e] => node e
	  | es => Record(Record.tuple es)
      val unit = tuple[]
   end

structure Dec =
   struct
      open Dec
      fun sequence(d1 : t, d2 : t) : t =
	 makeRegion'(case (node d1, node d2) of
			(SeqDec d1, SeqDec d2) => SeqDec(d1@d2)
		      | (SeqDec d1, _) => SeqDec(d1@[d2])
		      | (_, SeqDec d2) => SeqDec(d1::d2)
		      | _ => SeqDec[d1,d2],
		     left d1, right d2)
   end

structure Spec =
   struct
      open Spec

      (* some of this mess is so that a sharing equation captures as
       many specs as possible in its scope *)
      fun seq(s: t, s': t): t =
	 let fun reg s'' = makeRegion'(s'', left s, right s')
	 in case (node s, node s') of
	    (Empty,  _)     => s'
	  | (_,      Empty) => s
	  | (_,      Seq(s1, s2)) => reg(Seq(seq(s, s1), s2))
	  | (_,      Sharing{spec, equations}) =>
	       reg(Sharing{spec = seq(s, spec), equations = equations})
	  | _               => reg(Seq(s, s'))
	 end

(*      val seq = Trace.trace2("Spec.seq", layout, layout, layout) seq *)
   end

type rule = Pat.t * Exp.t
type rules = rule list
type clause = {pats : Pat.t list,
	       resultType : Type.t option,
	       body : Exp.t}
type eb = Con.t * EbRhs.t
type db = {tyvars : Tyvar.t list,
	   tycon : Tycon.t,
	   cons : (Con.t * Type.t option) list}

type strdesc = Strid.t * Sigexp.t

type wherespec = {tyvars : Tyvar.t list,
		  longtycon : Longtycon.t,
		  ty : Type.t}

type typdesc =  {tyvars : Tyvar.t list,
		 tycon : Tycon.t,
		 ty : Type.t option}

type valdesc = Var.t * Type.t

type exndesc = Con.t * Type.t option

type strbind = {name: Strid.t,
		def: Strexp.t,
		constraint: SigConst.t}
   
type sigbind = Sigid.t * Sigexp.t

type funbind = {name : Fctid.t,
		arg : FctArg.t,
		result : SigConst.t,
		body : Strexp.t}

fun ensureNonqualified(s,region) =
   if String.contains(s, #".")
      then (error(region, "expected nonqualified id") ;
	    "<bogus>")
   else s

fun cons1(x,(l,r,y)) = (x::l,r,y)

fun augment(id, sigexp, (wherespecs,right,binds)) =
   (id, Sigexp.makeRegion'(Sigexp.Where(sigexp,wherespecs),
			   Sigexp.left sigexp, right))
   :: binds

fun 'a augment1((strexp : Strexp.t,
		 makesigconst : Sigexp.t -> SigConst.t,
		 sigexp : Sigexp.t),
		(wherespecs : wherespec list,
		 right : int,
		 z : 'a)) : Strexp.t * 'a =
   (Strexp.makeRegion'
    (Strexp.Constrained(strexp, makesigconst(Sigexp.makeRegion'
					     (Sigexp.Where(sigexp,wherespecs),
					      Sigexp.left sigexp,
					      right))),
     Strexp.left strexp,
     right),
    z)

type 'a whereAnd = wherespec list * int * 'a list
   
  
end
structure LrTable = Token.LrTable
structure Token = Token
local open LrTable in 
val table=let val actionRows =
"\
\\001\000\001\000\000\000\000\000\
\\001\000\001\000\052\002\002\000\052\002\010\000\052\002\011\000\137\001\
\\016\000\052\002\019\000\052\002\022\000\052\002\026\000\052\002\
\\027\000\052\002\031\000\052\002\033\000\052\002\034\000\052\002\
\\036\000\052\002\037\000\052\002\040\000\052\002\041\000\052\002\
\\046\000\052\002\048\000\052\002\050\000\052\002\051\000\052\002\
\\052\000\141\001\058\000\055\002\059\000\055\002\066\000\052\002\000\000\
\\001\000\001\000\052\002\002\000\052\002\010\000\052\002\011\000\237\001\
\\016\000\052\002\019\000\052\002\022\000\052\002\026\000\052\002\
\\027\000\052\002\031\000\052\002\033\000\052\002\034\000\052\002\
\\036\000\052\002\037\000\052\002\040\000\052\002\041\000\052\002\
\\046\000\052\002\048\000\052\002\050\000\052\002\051\000\052\002\
\\052\000\141\001\058\000\069\002\059\000\069\002\066\000\052\002\000\000\
\\001\000\001\000\052\002\002\000\052\002\010\000\052\002\011\000\237\001\
\\016\000\052\002\019\000\052\002\022\000\052\002\026\000\052\002\
\\027\000\052\002\031\000\052\002\033\000\052\002\034\000\052\002\
\\036\000\052\002\037\000\052\002\040\000\052\002\041\000\052\002\
\\046\000\052\002\048\000\052\002\050\000\052\002\051\000\052\002\
\\052\000\141\001\058\000\072\002\059\000\072\002\066\000\052\002\000\000\
\\001\000\001\000\060\002\002\000\060\002\010\000\060\002\011\000\060\002\
\\016\000\060\002\019\000\060\002\022\000\060\002\026\000\060\002\
\\027\000\060\002\031\000\060\002\033\000\060\002\034\000\060\002\
\\036\000\060\002\037\000\060\002\040\000\060\002\041\000\060\002\
\\046\000\060\002\048\000\060\002\050\000\060\002\051\000\060\002\
\\058\000\057\002\059\000\057\002\066\000\060\002\000\000\
\\001\000\001\000\131\002\002\000\131\002\010\000\131\002\011\000\007\002\
\\016\000\131\002\022\000\131\002\026\000\131\002\027\000\131\002\
\\033\000\131\002\034\000\131\002\036\000\131\002\037\000\131\002\
\\040\000\131\002\041\000\131\002\046\000\131\002\048\000\131\002\
\\050\000\131\002\051\000\131\002\052\000\010\002\058\000\055\002\
\\059\000\055\002\000\000\
\\001\000\001\000\131\002\002\000\131\002\010\000\131\002\011\000\022\002\
\\016\000\131\002\022\000\131\002\026\000\131\002\027\000\131\002\
\\033\000\131\002\034\000\131\002\036\000\131\002\037\000\131\002\
\\040\000\131\002\041\000\131\002\046\000\131\002\048\000\131\002\
\\050\000\131\002\051\000\131\002\052\000\010\002\058\000\069\002\
\\059\000\069\002\000\000\
\\001\000\001\000\131\002\002\000\131\002\010\000\131\002\011\000\022\002\
\\016\000\131\002\022\000\131\002\026\000\131\002\027\000\131\002\
\\033\000\131\002\034\000\131\002\036\000\131\002\037\000\131\002\
\\040\000\131\002\041\000\131\002\046\000\131\002\048\000\131\002\
\\050\000\131\002\051\000\131\002\052\000\010\002\058\000\072\002\
\\059\000\072\002\000\000\
\\001\000\001\000\024\003\002\000\024\003\003\000\025\003\010\000\024\003\
\\011\000\024\003\012\000\024\003\013\000\024\003\014\000\024\003\
\\016\000\024\003\018\000\024\003\019\000\024\003\020\000\024\003\
\\021\000\024\003\022\000\024\003\023\000\024\003\024\000\024\003\
\\026\000\024\003\027\000\024\003\028\000\024\003\031\000\024\003\
\\032\000\024\003\033\000\024\003\034\000\024\003\036\000\024\003\
\\037\000\024\003\038\000\024\003\040\000\024\003\041\000\024\003\
\\044\000\024\003\046\000\024\003\048\000\024\003\049\000\024\003\
\\050\000\024\003\051\000\024\003\052\000\024\003\055\000\024\003\
\\056\000\024\003\057\000\025\003\058\000\024\003\059\000\024\003\
\\060\000\024\003\064\000\024\003\065\000\024\003\066\000\024\003\
\\067\000\024\003\068\000\024\003\000\000\
\\001\000\002\000\025\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\027\000\020\000\033\000\019\000\034\000\018\000\
\\036\000\017\000\037\000\016\000\040\000\015\000\041\000\014\000\
\\046\000\013\000\048\000\012\000\050\000\011\000\051\000\010\000\000\000\
\\001\000\002\000\231\001\012\000\255\000\000\000\
\\001\000\003\000\034\000\000\000\
\\001\000\003\000\034\000\004\000\058\000\020\000\033\000\039\000\130\000\
\\057\000\032\000\063\000\129\000\000\000\
\\001\000\003\000\034\000\005\000\081\000\017\000\125\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\005\000\081\000\017\000\125\000\057\000\032\000\
\\064\000\124\000\000\000\
\\001\000\003\000\034\000\005\000\081\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\005\000\081\000\057\000\032\000\064\000\008\001\000\000\
\\001\000\003\000\034\000\005\000\081\000\057\000\032\000\064\000\027\001\000\000\
\\001\000\003\000\034\000\016\000\095\001\057\000\032\000\000\000\
\\001\000\003\000\034\000\020\000\033\000\039\000\130\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\020\000\033\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\045\000\217\000\057\000\032\000\000\000\
\\001\000\003\000\034\000\050\000\149\001\057\000\032\000\000\000\
\\001\000\003\000\034\000\057\000\032\000\000\000\
\\001\000\003\000\059\000\004\000\058\000\005\000\057\000\006\000\056\000\
\\007\000\055\000\008\000\054\000\009\000\053\000\039\000\052\000\
\\043\000\051\000\054\000\050\000\057\000\049\000\061\000\048\000\
\\062\000\047\000\063\000\046\000\000\000\
\\001\000\003\000\059\000\004\000\058\000\005\000\057\000\006\000\056\000\
\\007\000\055\000\008\000\054\000\009\000\053\000\039\000\052\000\
\\054\000\050\000\057\000\049\000\061\000\048\000\062\000\047\000\
\\063\000\046\000\000\000\
\\001\000\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\015\000\194\000\020\000\193\000\
\\025\000\192\000\029\000\191\000\030\000\190\000\035\000\189\000\
\\039\000\188\000\042\000\187\000\053\000\186\000\057\000\049\000\
\\061\000\185\000\062\000\184\000\063\000\183\000\065\000\023\001\
\\070\000\182\000\000\000\
\\001\000\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\015\000\194\000\020\000\193\000\
\\025\000\192\000\029\000\191\000\030\000\190\000\035\000\189\000\
\\039\000\188\000\042\000\187\000\053\000\186\000\057\000\049\000\
\\061\000\185\000\062\000\184\000\063\000\183\000\066\000\020\001\
\\070\000\182\000\000\000\
\\001\000\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\015\000\194\000\020\000\193\000\
\\025\000\192\000\029\000\191\000\030\000\190\000\035\000\189\000\
\\039\000\188\000\042\000\187\000\053\000\186\000\057\000\049\000\
\\061\000\185\000\062\000\184\000\063\000\183\000\070\000\182\000\000\000\
\\001\000\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\039\000\052\000\054\000\050\000\
\\057\000\049\000\061\000\048\000\062\000\047\000\063\000\107\000\000\000\
\\001\000\003\000\059\000\020\000\193\000\057\000\049\000\000\000\
\\001\000\003\000\059\000\035\000\056\001\047\000\055\001\057\000\049\000\000\000\
\\001\000\003\000\059\000\050\000\078\001\057\000\049\000\000\000\
\\001\000\003\000\059\000\057\000\049\000\000\000\
\\001\000\003\000\173\000\000\000\
\\001\000\003\000\173\000\004\000\058\000\061\000\172\000\063\000\171\000\000\000\
\\001\000\003\000\173\000\057\000\001\001\000\000\
\\001\000\004\000\058\000\000\000\
\\001\000\008\000\016\001\000\000\
\\001\000\012\000\255\000\013\000\063\001\000\000\
\\001\000\012\000\255\000\060\000\106\001\066\000\105\001\000\000\
\\001\000\013\000\056\003\020\000\063\003\058\000\056\003\060\000\056\003\
\\064\000\056\003\000\000\
\\001\000\013\000\112\000\020\000\111\000\058\000\110\000\000\000\
\\001\000\013\000\112\000\024\000\127\001\058\000\110\000\000\000\
\\001\000\018\000\233\001\028\000\014\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\001\000\019\000\064\001\000\000\
\\001\000\019\000\096\001\000\000\
\\001\000\019\000\150\001\000\000\
\\001\000\019\000\205\001\000\000\
\\001\000\019\000\232\001\000\000\
\\001\000\019\000\254\001\000\000\
\\001\000\019\000\001\002\058\000\143\001\059\000\142\001\000\000\
\\001\000\020\000\139\000\000\000\
\\001\000\020\000\204\000\000\000\
\\001\000\020\000\210\000\000\000\
\\001\000\020\000\212\000\000\000\
\\001\000\020\000\239\000\000\000\
\\001\000\020\000\247\000\000\000\
\\001\000\020\000\254\000\000\000\
\\001\000\020\000\045\001\000\000\
\\001\000\020\000\123\001\000\000\
\\001\000\020\000\133\001\000\000\
\\001\000\020\000\163\001\000\000\
\\001\000\020\000\215\001\000\000\
\\001\000\020\000\224\001\000\000\
\\001\000\020\000\004\002\000\000\
\\001\000\023\000\124\001\028\000\014\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\001\000\025\000\132\001\000\000\
\\001\000\025\000\195\001\000\000\
\\001\000\028\000\014\001\038\000\129\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\001\000\028\000\014\001\049\000\126\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\001\000\031\000\143\000\000\000\
\\001\000\031\000\125\001\000\000\
\\001\000\031\000\174\001\000\000\
\\001\000\031\000\206\001\000\000\
\\001\000\050\000\149\001\000\000\
\\001\000\055\000\160\000\000\000\
\\001\000\058\000\140\000\000\000\
\\001\000\058\000\067\001\000\000\
\\001\000\058\000\109\001\000\000\
\\001\000\058\000\114\001\000\000\
\\001\000\058\000\143\001\059\000\142\001\000\000\
\\001\000\058\000\143\001\059\000\142\001\066\000\238\001\000\000\
\\001\000\058\000\159\001\000\000\
\\001\000\058\000\162\001\000\000\
\\001\000\063\000\145\000\000\000\
\\001\000\064\000\203\000\000\000\
\\001\000\064\000\107\001\000\000\
\\001\000\064\000\121\001\000\000\
\\001\000\065\000\201\000\000\000\
\\001\000\065\000\119\001\000\000\
\\001\000\066\000\196\000\000\000\
\\001\000\066\000\197\000\000\000\
\\001\000\066\000\211\000\000\000\
\\001\000\066\000\065\001\000\000\
\\001\000\066\000\104\001\000\000\
\\001\000\066\000\115\001\000\000\
\\001\000\066\000\118\001\000\000\
\\001\000\066\000\239\001\000\000\
\\028\002\000\000\
\\029\002\002\000\025\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\027\000\020\000\033\000\019\000\034\000\018\000\
\\036\000\017\000\037\000\016\000\040\000\015\000\041\000\014\000\
\\046\000\013\000\048\000\012\000\050\000\011\000\051\000\010\000\000\000\
\\030\002\002\000\025\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\027\000\020\000\033\000\019\000\034\000\018\000\
\\036\000\017\000\037\000\016\000\040\000\015\000\041\000\014\000\
\\046\000\013\000\048\000\012\000\050\000\011\000\051\000\010\000\000\000\
\\031\002\000\000\
\\032\002\000\000\
\\033\002\000\000\
\\034\002\000\000\
\\035\002\000\000\
\\036\002\000\000\
\\037\002\000\000\
\\038\002\002\000\078\000\003\000\059\000\010\000\024\000\016\000\023\000\
\\022\000\022\000\026\000\021\000\033\000\019\000\034\000\018\000\
\\035\000\056\001\036\000\017\000\037\000\016\000\040\000\015\000\
\\041\000\014\000\047\000\055\001\048\000\012\000\050\000\011\000\
\\051\000\010\000\057\000\049\000\000\000\
\\038\002\002\000\078\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\033\000\019\000\034\000\018\000\036\000\017\000\
\\037\000\016\000\040\000\015\000\041\000\014\000\048\000\012\000\
\\050\000\011\000\051\000\010\000\000\000\
\\039\002\000\000\
\\040\002\000\000\
\\041\002\000\000\
\\042\002\000\000\
\\043\002\000\000\
\\044\002\000\000\
\\045\002\000\000\
\\046\002\000\000\
\\047\002\000\000\
\\048\002\000\000\
\\049\002\000\000\
\\050\002\000\000\
\\051\002\000\000\
\\052\002\011\000\137\001\000\000\
\\053\002\000\000\
\\054\002\000\000\
\\055\002\052\000\058\001\000\000\
\\056\002\000\000\
\\057\002\000\000\
\\058\002\000\000\
\\059\002\000\000\
\\061\002\063\000\145\001\000\000\
\\062\002\000\000\
\\063\002\000\000\
\\064\002\000\000\
\\065\002\000\000\
\\066\002\000\000\
\\067\002\052\000\058\001\000\000\
\\068\002\000\000\
\\069\002\011\000\209\001\052\000\058\001\000\000\
\\070\002\000\000\
\\071\002\000\000\
\\072\002\011\000\209\001\052\000\058\001\000\000\
\\073\002\000\000\
\\074\002\000\000\
\\075\002\000\000\
\\076\002\000\000\
\\077\002\000\000\
\\077\002\003\000\034\000\057\000\032\000\000\000\
\\078\002\000\000\
\\079\002\011\000\062\001\052\000\061\001\000\000\
\\079\002\011\000\213\001\052\000\061\001\000\000\
\\080\002\000\000\
\\081\002\000\000\
\\082\002\000\000\
\\083\002\000\000\
\\084\002\012\000\255\000\000\000\
\\085\002\058\000\138\000\059\000\137\000\000\000\
\\086\002\000\000\
\\087\002\000\000\
\\088\002\002\000\236\000\003\000\034\000\016\000\235\000\021\000\234\000\
\\022\000\233\000\032\000\232\000\044\000\231\000\048\000\230\000\
\\050\000\229\000\051\000\228\000\057\000\032\000\000\000\
\\088\002\002\000\236\000\016\000\235\000\021\000\234\000\022\000\233\000\
\\032\000\232\000\044\000\231\000\048\000\230\000\050\000\229\000\
\\051\000\228\000\000\000\
\\089\002\000\000\
\\090\002\000\000\
\\091\002\000\000\
\\092\002\000\000\
\\093\002\000\000\
\\094\002\000\000\
\\095\002\000\000\
\\096\002\000\000\
\\097\002\000\000\
\\098\002\000\000\
\\099\002\000\000\
\\100\002\000\000\
\\101\002\000\000\
\\102\002\000\000\
\\103\002\020\000\163\001\000\000\
\\104\002\000\000\
\\105\002\020\000\224\001\000\000\
\\106\002\000\000\
\\107\002\000\000\
\\108\002\011\000\251\001\052\000\250\001\000\000\
\\108\002\011\000\019\002\052\000\250\001\000\000\
\\109\002\000\000\
\\110\002\000\000\
\\111\002\000\000\
\\112\002\000\000\
\\113\002\011\000\160\001\000\000\
\\114\002\000\000\
\\115\002\000\000\
\\116\002\020\000\220\001\000\000\
\\117\002\012\000\255\000\000\000\
\\118\002\011\000\158\001\000\000\
\\119\002\000\000\
\\120\002\012\000\255\000\000\000\
\\121\002\011\000\168\001\000\000\
\\122\002\000\000\
\\123\002\000\000\
\\124\002\038\000\170\001\000\000\
\\125\002\012\000\255\000\000\000\
\\126\002\000\000\
\\127\002\000\000\
\\128\002\000\000\
\\129\002\000\000\
\\130\002\000\000\
\\131\002\011\000\007\002\000\000\
\\132\002\000\000\
\\133\002\000\000\
\\134\002\000\000\
\\135\002\000\000\
\\136\002\000\000\
\\137\002\002\000\253\000\010\000\024\000\016\000\023\000\022\000\022\000\
\\026\000\021\000\033\000\019\000\034\000\018\000\036\000\252\000\
\\037\000\016\000\040\000\015\000\041\000\014\000\050\000\011\000\
\\051\000\010\000\000\000\
\\138\002\000\000\
\\139\002\000\000\
\\140\002\000\000\
\\141\002\000\000\
\\142\002\000\000\
\\143\002\011\000\108\000\000\000\
\\144\002\011\000\108\000\000\000\
\\145\002\011\000\207\000\000\000\
\\146\002\011\000\207\000\000\000\
\\147\002\000\000\
\\148\002\000\000\
\\149\002\000\000\
\\150\002\000\000\
\\151\002\000\000\
\\152\002\000\000\
\\153\002\000\000\
\\154\002\000\000\
\\155\002\000\000\
\\156\002\011\000\108\000\000\000\
\\157\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\158\002\058\000\150\000\000\000\
\\159\002\012\000\255\000\000\000\
\\160\002\000\000\
\\161\002\000\000\
\\162\002\011\000\207\000\000\000\
\\163\002\011\000\148\000\000\000\
\\164\002\000\000\
\\165\002\014\000\147\000\000\000\
\\166\002\000\000\
\\167\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\168\002\000\000\
\\169\002\011\000\135\001\012\000\255\000\000\000\
\\170\002\000\000\
\\171\002\000\000\
\\172\002\000\000\
\\173\002\004\000\058\000\063\000\063\000\000\000\
\\174\002\000\000\
\\175\002\000\000\
\\176\002\060\000\200\000\000\000\
\\177\002\000\000\
\\178\002\014\000\171\001\000\000\
\\179\002\000\000\
\\180\002\038\000\172\001\000\000\
\\181\002\012\000\255\000\000\000\
\\182\002\011\000\151\000\000\000\
\\183\002\000\000\
\\184\002\000\000\
\\185\002\000\000\
\\186\002\020\000\155\000\038\000\154\000\000\000\
\\187\002\012\000\255\000\000\000\
\\188\002\000\000\
\\189\002\005\000\081\000\000\000\
\\190\002\000\000\
\\191\002\005\000\081\000\000\000\
\\192\002\000\000\
\\193\002\000\000\
\\194\002\000\000\
\\195\002\000\000\
\\196\002\000\000\
\\197\002\000\000\
\\198\002\000\000\
\\199\002\000\000\
\\200\002\000\000\
\\201\002\000\000\
\\202\002\000\000\
\\203\002\000\000\
\\204\002\000\000\
\\205\002\011\000\158\000\000\000\
\\206\002\000\000\
\\207\002\000\000\
\\208\002\056\000\157\000\000\000\
\\209\002\000\000\
\\210\002\011\000\214\001\000\000\
\\211\002\000\000\
\\212\002\014\000\128\001\000\000\
\\213\002\000\000\
\\214\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\215\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\216\002\000\000\
\\217\002\060\000\122\001\000\000\
\\218\002\002\000\117\001\028\000\014\001\058\000\013\001\060\000\116\001\
\\067\000\012\001\068\000\011\001\000\000\
\\218\002\002\000\117\001\028\000\014\001\058\000\013\001\067\000\012\001\
\\068\000\011\001\000\000\
\\219\002\000\000\
\\220\002\000\000\
\\221\002\000\000\
\\222\002\058\000\013\001\068\000\011\001\000\000\
\\223\002\058\000\013\001\000\000\
\\224\002\012\000\255\000\000\000\
\\225\002\000\000\
\\226\002\000\000\
\\227\002\000\000\
\\228\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\229\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\230\002\028\000\014\001\058\000\013\001\067\000\012\001\068\000\011\001\000\000\
\\231\002\000\000\
\\232\002\000\000\
\\233\002\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\020\000\193\000\029\000\191\000\
\\035\000\189\000\039\000\188\000\057\000\049\000\061\000\185\000\
\\062\000\184\000\063\000\183\000\070\000\182\000\000\000\
\\234\002\000\000\
\\235\002\000\000\
\\236\002\000\000\
\\237\002\000\000\
\\238\002\000\000\
\\239\002\000\000\
\\240\002\000\000\
\\241\002\000\000\
\\242\002\000\000\
\\243\002\000\000\
\\244\002\000\000\
\\245\002\000\000\
\\246\002\000\000\
\\247\002\000\000\
\\248\002\028\000\014\001\058\000\013\001\060\000\116\001\067\000\012\001\
\\068\000\011\001\000\000\
\\249\002\028\000\014\001\058\000\013\001\060\000\120\001\067\000\012\001\
\\068\000\011\001\000\000\
\\250\002\000\000\
\\251\002\000\000\
\\252\002\013\000\112\000\058\000\110\000\000\000\
\\253\002\012\000\255\000\000\000\
\\254\002\000\000\
\\255\002\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\039\000\052\000\054\000\050\000\
\\057\000\049\000\061\000\048\000\062\000\047\000\063\000\107\000\000\000\
\\000\003\000\000\
\\001\003\000\000\
\\002\003\000\000\
\\003\003\000\000\
\\004\003\000\000\
\\005\003\000\000\
\\006\003\000\000\
\\007\003\000\000\
\\008\003\000\000\
\\009\003\000\000\
\\010\003\003\000\059\000\004\000\058\000\005\000\057\000\006\000\056\000\
\\007\000\055\000\008\000\054\000\009\000\053\000\039\000\052\000\
\\054\000\050\000\057\000\049\000\061\000\048\000\062\000\047\000\
\\063\000\107\000\000\000\
\\010\003\003\000\059\000\005\000\057\000\006\000\056\000\007\000\055\000\
\\008\000\054\000\009\000\053\000\039\000\052\000\054\000\050\000\
\\057\000\049\000\061\000\048\000\062\000\047\000\063\000\107\000\000\000\
\\011\003\000\000\
\\012\003\013\000\112\000\058\000\110\000\060\000\199\000\000\000\
\\013\003\000\000\
\\014\003\000\000\
\\015\003\060\000\202\000\000\000\
\\016\003\000\000\
\\017\003\013\000\112\000\058\000\110\000\000\000\
\\018\003\000\000\
\\019\003\013\000\043\001\000\000\
\\020\003\013\000\112\000\058\000\110\000\000\000\
\\021\003\000\000\
\\022\003\000\000\
\\023\003\012\000\255\000\000\000\
\\025\003\000\000\
\\026\003\000\000\
\\027\003\000\000\
\\028\003\000\000\
\\029\003\000\000\
\\030\003\000\000\
\\031\003\000\000\
\\032\003\000\000\
\\033\003\012\000\255\000\000\000\
\\034\003\000\000\
\\035\003\060\000\108\001\000\000\
\\036\003\000\000\
\\037\003\003\000\173\000\057\000\001\001\000\000\
\\038\003\012\000\255\000\060\000\106\001\000\000\
\\039\003\000\000\
\\040\003\000\000\
\\041\003\000\000\
\\042\003\000\000\
\\043\003\000\000\
\\044\003\000\000\
\\045\003\000\000\
\\046\003\000\000\
\\047\003\000\000\
\\048\003\000\000\
\\049\003\000\000\
\\050\003\000\000\
\\051\003\000\000\
\\052\003\000\000\
\\053\003\000\000\
\\054\003\000\000\
\\055\003\000\000\
\\056\003\000\000\
\\057\003\003\000\034\000\020\000\033\000\057\000\032\000\000\000\
\\058\003\000\000\
\\059\003\000\000\
\\060\003\000\000\
\\061\003\000\000\
\\062\003\000\000\
\\063\003\000\000\
\\064\003\000\000\
\\065\003\000\000\
\\066\003\000\000\
\\067\003\003\000\034\000\057\000\032\000\000\000\
\\068\003\000\000\
\\069\003\000\000\
\\070\003\000\000\
\\071\003\000\000\
\\072\003\000\000\
\\073\003\000\000\
\\074\003\000\000\
\\075\003\000\000\
\\076\003\003\000\059\000\057\000\049\000\000\000\
\\077\003\000\000\
\"
val actionRowNumbers =
"\009\000\099\000\104\000\100\000\
\\113\000\105\000\116\000\020\000\
\\024\000\248\000\023\000\023\000\
\\020\000\033\000\012\001\110\000\
\\010\001\008\001\023\000\025\000\
\\023\000\248\000\248\000\101\000\
\\103\000\229\000\133\001\122\001\
\\131\001\124\001\123\001\125\001\
\\121\001\029\000\218\000\077\001\
\\075\001\074\001\071\001\042\000\
\\078\001\249\000\080\001\149\001\
\\086\001\087\001\014\000\127\001\
\\081\001\012\000\023\000\120\001\
\\119\001\118\001\117\001\116\001\
\\138\001\126\001\011\000\224\000\
\\246\000\037\000\114\000\158\000\
\\141\001\106\000\052\000\142\001\
\\077\000\135\001\228\000\152\001\
\\151\001\110\000\108\000\071\000\
\\110\000\014\001\011\001\013\001\
\\009\001\107\000\085\000\145\001\
\\029\000\240\000\238\000\222\000\
\\233\000\001\001\227\000\005\001\
\\136\001\027\001\024\001\018\001\
\\017\001\015\001\225\000\011\000\
\\076\000\011\000\102\000\134\001\
\\219\000\087\001\029\000\076\001\
\\035\000\028\000\029\000\091\000\
\\092\000\089\001\251\000\089\000\
\\092\001\086\000\053\000\233\000\
\\041\000\140\001\084\001\093\001\
\\019\000\220\000\233\000\037\000\
\\020\000\079\001\132\001\054\000\
\\137\001\093\000\055\000\021\000\
\\021\000\021\000\035\000\153\001\
\\112\000\110\000\111\000\161\000\
\\223\000\029\000\029\000\056\000\
\\035\000\023\000\004\001\003\001\
\\035\000\033\000\022\001\248\000\
\\248\000\057\000\212\000\058\000\
\\231\000\098\001\073\001\099\001\
\\008\000\036\000\108\001\102\001\
\\146\001\035\000\016\000\128\001\
\\053\001\045\001\040\001\232\000\
\\053\001\056\001\148\001\129\001\
\\038\000\027\000\026\000\017\000\
\\028\000\028\000\020\000\212\000\
\\028\000\015\000\029\000\130\001\
\\028\000\072\001\250\000\082\001\
\\088\001\029\000\037\000\083\001\
\\013\000\085\001\029\000\096\001\
\\221\000\019\000\059\000\233\000\
\\035\000\247\000\031\000\147\000\
\\138\000\160\000\148\000\162\000\
\\159\000\151\000\039\000\045\000\
\\094\000\174\000\165\000\162\000\
\\211\000\078\000\020\000\248\000\
\\023\000\032\000\021\000\023\000\
\\248\000\248\000\162\000\241\000\
\\239\000\028\000\234\000\002\001\
\\006\001\007\001\150\001\028\001\
\\025\001\018\000\046\000\216\000\
\\215\000\212\000\212\000\212\000\
\\023\000\035\000\107\001\035\000\
\\095\000\040\000\087\000\111\001\
\\079\000\139\001\104\001\051\001\
\\054\001\028\000\028\000\035\000\
\\029\000\052\001\080\000\096\000\
\\037\001\097\000\060\001\090\000\
\\069\001\064\001\088\000\036\001\
\\060\000\059\001\066\000\050\001\
\\055\001\072\000\070\000\057\001\
\\043\000\031\001\046\001\069\000\
\\089\001\252\000\091\001\094\001\
\\095\001\029\000\237\000\067\000\
\\061\000\244\000\004\000\124\000\
\\001\000\126\000\081\000\117\000\
\\132\000\110\000\110\000\139\000\
\\075\000\047\000\146\000\075\000\
\\023\000\030\000\115\000\158\000\
\\164\000\021\000\193\000\166\000\
\\083\000\188\000\167\000\011\000\
\\171\000\084\000\176\000\062\000\
\\034\000\172\000\149\000\196\000\
\\170\000\199\000\168\000\023\001\
\\020\001\019\001\016\001\169\000\
\\163\000\242\000\253\000\026\001\
\\255\000\034\000\226\000\213\000\
\\073\000\214\000\100\001\112\001\
\\101\001\113\001\034\000\106\001\
\\035\000\103\001\015\000\035\000\
\\043\001\042\001\044\001\041\001\
\\035\000\062\001\028\000\028\000\
\\061\001\063\001\028\000\058\001\
\\015\000\028\000\028\000\028\000\
\\028\000\028\000\029\000\029\000\
\\090\001\097\001\029\000\068\000\
\\243\000\248\000\119\000\023\000\
\\128\000\120\000\118\000\075\000\
\\021\000\021\000\134\000\109\000\
\\048\000\074\000\140\000\248\000\
\\150\000\152\000\153\000\230\000\
\\029\001\147\001\063\000\210\000\
\\020\000\035\000\248\000\191\000\
\\021\000\033\000\175\000\064\000\
\\173\000\143\001\023\000\198\000\
\\035\000\023\000\035\000\021\001\
\\212\000\105\001\115\001\114\001\
\\110\001\109\001\010\000\067\001\
\\068\001\038\001\039\001\070\001\
\\035\001\034\001\048\001\049\000\
\\044\000\033\001\032\001\047\001\
\\235\000\029\000\245\000\125\000\
\\002\000\131\000\130\000\129\000\
\\127\000\082\000\098\000\133\000\
\\031\000\142\000\141\000\075\000\
\\034\000\154\000\155\000\022\000\
\\030\000\031\000\194\000\195\000\
\\189\000\190\000\035\000\182\000\
\\178\000\177\000\034\000\144\001\
\\197\000\200\000\254\000\000\001\
\\050\000\066\001\065\001\028\000\
\\236\000\121\000\122\000\022\000\
\\136\000\137\000\051\000\143\000\
\\065\000\152\000\030\001\206\000\
\\005\000\201\000\192\000\181\000\
\\075\000\023\000\180\000\179\000\
\\217\000\049\001\003\000\135\000\
\\144\000\145\000\035\000\156\000\
\\203\000\023\000\204\000\202\000\
\\075\000\183\000\184\000\123\000\
\\157\000\207\000\006\000\185\000\
\\186\000\022\000\205\000\208\000\
\\022\000\183\000\007\000\187\000\
\\209\000\000\000"
val gotoT =
"\
\\083\000\007\000\086\000\006\000\143\000\005\000\144\000\004\000\
\\145\000\003\000\146\000\002\000\147\000\001\000\148\000\025\002\000\000\
\\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\143\000\005\000\144\000\004\000\
\\145\000\003\000\146\000\002\000\147\000\024\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\026\000\
\\012\000\025\000\000\000\
\\007\000\043\000\009\000\042\000\016\000\041\000\028\000\040\000\
\\054\000\039\000\055\000\038\000\056\000\037\000\057\000\036\000\
\\058\000\035\000\066\000\034\000\075\000\033\000\000\000\
\\016\000\060\000\072\000\059\000\074\000\058\000\000\000\
\\003\000\064\000\005\000\027\000\019\000\063\000\115\000\062\000\000\000\
\\003\000\067\000\005\000\027\000\020\000\066\000\120\000\065\000\000\000\
\\003\000\029\000\004\000\069\000\005\000\027\000\013\000\068\000\000\000\
\\007\000\072\000\029\000\071\000\030\000\070\000\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\141\000\075\000\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\001\000\078\000\002\000\077\000\000\000\
\\001\000\080\000\002\000\077\000\000\000\
\\003\000\083\000\005\000\027\000\022\000\082\000\125\000\081\000\000\000\
\\007\000\043\000\009\000\042\000\016\000\041\000\028\000\040\000\
\\056\000\088\000\057\000\036\000\058\000\035\000\069\000\087\000\
\\070\000\086\000\071\000\085\000\075\000\084\000\000\000\
\\003\000\092\000\005\000\027\000\015\000\091\000\079\000\090\000\
\\080\000\089\000\000\000\
\\016\000\060\000\074\000\099\000\131\000\098\000\133\000\097\000\
\\135\000\096\000\136\000\095\000\138\000\094\000\139\000\093\000\000\000\
\\016\000\060\000\074\000\101\000\136\000\100\000\138\000\094\000\
\\139\000\093\000\000\000\
\\083\000\007\000\086\000\006\000\143\000\005\000\144\000\004\000\
\\145\000\003\000\146\000\002\000\147\000\102\000\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\026\000\
\\012\000\103\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\039\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\066\000\104\000\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\056\000\107\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\016\000\114\000\028\000\040\000\
\\054\000\113\000\055\000\038\000\056\000\037\000\057\000\036\000\
\\058\000\035\000\063\000\112\000\076\000\111\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\113\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\063\000\115\000\000\000\
\\002\000\121\000\003\000\120\000\005\000\027\000\011\000\119\000\
\\018\000\118\000\060\000\117\000\061\000\116\000\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\126\000\
\\016\000\041\000\068\000\125\000\075\000\124\000\000\000\
\\003\000\130\000\005\000\027\000\011\000\129\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\132\000\017\000\131\000\000\000\
\\000\000\
\\000\000\
\\016\000\114\000\076\000\133\000\000\000\
\\000\000\
\\110\000\134\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\072\000\029\000\071\000\030\000\139\000\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\141\000\140\000\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\141\000\142\000\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\056\000\088\000\
\\057\000\036\000\058\000\035\000\069\000\144\000\070\000\086\000\
\\071\000\085\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\067\000\147\000\000\000\
\\000\000\
\\000\000\
\\081\000\151\000\082\000\150\000\000\000\
\\000\000\
\\140\000\154\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\132\000\017\000\157\000\000\000\
\\000\000\
\\005\000\132\000\017\000\159\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\113\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\063\000\112\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\039\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\066\000\160\000\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\162\000\037\000\161\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\175\000\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\193\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\064\000\196\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\067\000\203\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\126\000\
\\068\000\204\000\000\000\
\\000\000\
\\067\000\206\000\000\000\
\\016\000\114\000\076\000\111\000\000\000\
\\003\000\029\000\004\000\069\000\005\000\027\000\013\000\207\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\106\000\213\000\
\\108\000\212\000\109\000\211\000\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\106\000\216\000\
\\108\000\212\000\109\000\211\000\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\108\000\217\000\
\\109\000\211\000\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\218\000\037\000\161\000\000\000\
\\000\000\
\\000\000\
\\083\000\007\000\086\000\006\000\141\000\219\000\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\000\000\
\\003\000\064\000\005\000\027\000\019\000\225\000\089\000\224\000\
\\090\000\223\000\091\000\222\000\092\000\221\000\130\000\220\000\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\056\000\088\000\
\\057\000\036\000\058\000\035\000\070\000\235\000\071\000\085\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\056\000\088\000\
\\057\000\036\000\058\000\035\000\069\000\236\000\070\000\086\000\
\\071\000\085\000\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\238\000\037\000\161\000\000\000\
\\003\000\092\000\005\000\027\000\015\000\091\000\079\000\239\000\
\\080\000\089\000\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\240\000\037\000\161\000\000\000\
\\007\000\242\000\026\000\241\000\000\000\
\\000\000\
\\016\000\060\000\072\000\243\000\074\000\058\000\000\000\
\\016\000\060\000\074\000\101\000\138\000\094\000\139\000\244\000\000\000\
\\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\246\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\254\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\001\001\037\000\161\000\
\\040\000\000\001\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\004\001\
\\031\000\003\001\032\000\002\001\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\048\000\008\001\049\000\007\001\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\048\000\008\001\049\000\013\001\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\045\000\017\001\046\000\016\001\047\000\174\000\048\000\173\000\
\\050\000\172\000\052\000\015\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\020\001\047\000\174\000\048\000\173\000\050\000\172\000\
\\053\000\019\001\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\024\001\
\\043\000\023\001\044\000\022\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\026\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\027\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\028\001\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\029\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\030\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\031\001\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\034\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\035\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\036\001\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\000\000\
\\016\000\114\000\076\000\037\001\000\000\
\\000\000\
\\002\000\121\000\003\000\120\000\005\000\027\000\011\000\119\000\
\\018\000\118\000\060\000\038\001\061\000\116\000\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\039\001\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\000\000\
\\065\000\040\001\000\000\
\\000\000\
\\003\000\029\000\004\000\028\000\005\000\027\000\010\000\126\000\
\\068\000\042\001\000\000\
\\000\000\
\\067\000\044\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\045\001\037\000\161\000\000\000\
\\000\000\
\\007\000\052\001\116\000\051\001\149\000\050\001\150\000\049\001\
\\151\000\048\001\152\000\047\001\153\000\046\001\000\000\
\\000\000\
\\123\000\055\001\000\000\
\\000\000\
\\000\000\
\\089\000\057\001\090\000\223\000\091\000\222\000\092\000\221\000\000\000\
\\000\000\
\\121\000\058\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\089\000\064\001\090\000\223\000\091\000\222\000\092\000\221\000\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\069\000\005\000\027\000\013\000\068\001\
\\099\000\067\001\100\000\066\001\000\000\
\\016\000\060\000\074\000\071\001\096\000\070\001\097\000\069\001\000\000\
\\003\000\064\000\005\000\027\000\019\000\073\001\093\000\072\001\000\000\
\\007\000\072\000\029\000\075\001\104\000\074\001\000\000\
\\003\000\067\000\005\000\027\000\020\000\078\001\106\000\077\001\
\\108\000\212\000\109\000\211\000\000\000\
\\003\000\092\000\005\000\027\000\015\000\081\001\101\000\080\001\
\\102\000\079\001\000\000\
\\016\000\060\000\074\000\071\001\096\000\082\001\097\000\069\001\000\000\
\\016\000\060\000\074\000\099\000\132\000\087\001\134\000\086\001\
\\135\000\085\001\137\000\084\001\138\000\094\000\139\000\083\001\000\000\
\\089\000\088\001\090\000\223\000\091\000\222\000\092\000\221\000\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\089\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\092\000\005\000\027\000\015\000\092\001\077\000\091\001\
\\078\000\090\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\095\001\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\096\001\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\097\001\000\000\
\\003\000\092\000\005\000\027\000\015\000\092\001\077\000\091\001\
\\078\000\090\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\098\001\037\000\161\000\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\101\001\
\\034\000\100\001\035\000\099\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\108\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\109\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\110\001\037\000\161\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\111\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\064\000\128\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\054\000\129\001\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\073\000\132\001\000\000\
\\000\000\
\\118\000\134\001\000\000\
\\117\000\138\001\118\000\137\001\123\000\136\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\111\000\142\001\000\000\
\\083\000\007\000\086\000\006\000\141\000\144\001\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\083\000\007\000\086\000\006\000\141\000\145\001\142\000\074\000\
\\143\000\073\000\144\000\004\000\000\000\
\\000\000\
\\105\000\146\001\000\000\
\\000\000\
\\000\000\
\\105\000\149\001\000\000\
\\003\000\067\000\005\000\027\000\020\000\066\000\120\000\150\001\000\000\
\\007\000\179\000\008\000\153\001\024\000\152\001\025\000\151\001\000\000\
\\000\000\
\\110\000\154\001\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\106\000\155\001\
\\108\000\212\000\109\000\211\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\132\000\017\000\159\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\163\001\103\000\162\001\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\165\001\021\000\164\001\000\000\
\\000\000\
\\000\000\
\\039\000\167\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\171\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\254\000\000\000\
\\006\000\168\000\023\000\173\001\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\175\001\037\000\161\000\
\\040\000\174\001\000\000\
\\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\004\001\
\\031\000\003\001\032\000\176\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\177\001\037\000\161\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\178\001\037\000\161\000\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\180\001\047\000\174\000\048\000\173\000\050\000\172\000\
\\052\000\179\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\045\000\182\001\046\000\181\001\047\000\174\000\048\000\173\000\
\\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\020\001\047\000\174\000\048\000\173\000\050\000\172\000\
\\053\000\183\001\000\000\
\\000\000\
\\002\000\121\000\003\000\005\001\005\000\027\000\018\000\024\001\
\\043\000\023\001\044\000\184\001\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\185\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\186\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\045\000\187\001\046\000\181\001\047\000\174\000\048\000\173\000\
\\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\188\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\189\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\190\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\191\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\192\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\016\000\060\000\072\000\194\001\074\000\058\000\000\000\
\\000\000\
\\003\000\064\000\005\000\027\000\019\000\063\000\115\000\195\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\105\000\196\001\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\108\000\197\001\
\\109\000\211\000\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\108\000\198\001\
\\109\000\211\000\000\000\
\\000\000\
\\007\000\052\001\083\000\007\000\086\000\006\000\141\000\202\001\
\\142\000\074\000\143\000\073\000\144\000\004\000\149\000\201\001\
\\150\000\049\001\151\000\200\001\153\000\199\001\000\000\
\\000\000\
\\000\000\
\\123\000\206\001\124\000\205\001\000\000\
\\016\000\060\000\074\000\208\001\000\000\
\\000\000\
\\121\000\210\001\122\000\209\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\029\000\004\000\069\000\005\000\027\000\013\000\068\001\
\\099\000\214\001\100\000\066\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\215\001\037\000\161\000\000\000\
\\016\000\060\000\074\000\071\001\096\000\216\001\097\000\069\001\000\000\
\\098\000\217\001\000\000\
\\003\000\067\000\005\000\027\000\020\000\214\000\108\000\219\001\
\\109\000\211\000\000\000\
\\007\000\072\000\029\000\221\001\104\000\220\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\165\001\021\000\223\001\000\000\
\\003\000\092\000\005\000\027\000\015\000\081\001\101\000\224\001\
\\102\000\079\001\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\225\001\037\000\161\000\000\000\
\\003\000\092\000\005\000\027\000\015\000\092\001\077\000\226\001\
\\078\000\090\001\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\227\001\037\000\161\000\000\000\
\\000\000\
\\083\000\007\000\084\000\249\000\085\000\248\000\086\000\247\000\
\\087\000\228\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\043\000\009\000\042\000\028\000\040\000\041\000\232\001\
\\042\000\033\001\054\000\032\001\055\000\038\000\056\000\037\000\
\\057\000\036\000\058\000\035\000\000\000\
\\000\000\
\\000\000\
\\117\000\234\001\118\000\137\001\119\000\233\001\123\000\206\001\
\\124\000\205\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\123\000\136\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\052\001\149\000\238\001\150\000\049\001\151\000\200\001\
\\153\000\199\001\000\000\
\\000\000\
\\000\000\
\\105\000\239\001\000\000\
\\006\000\168\000\023\000\240\001\000\000\
\\000\000\
\\000\000\
\\003\000\067\000\005\000\027\000\020\000\066\000\105\000\241\001\
\\120\000\150\001\000\000\
\\007\000\179\000\008\000\153\001\024\000\152\001\025\000\242\001\000\000\
\\007\000\052\001\126\000\245\001\149\000\050\001\150\000\049\001\
\\151\000\244\001\152\000\243\001\153\000\046\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\246\001\037\000\161\000\000\000\
\\094\000\247\001\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\023\000\251\001\103\000\250\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\007\000\179\000\008\000\178\000\009\000\177\000\027\000\176\000\
\\046\000\253\001\047\000\174\000\048\000\173\000\050\000\172\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\003\000\064\000\005\000\027\000\019\000\063\000\105\000\254\001\
\\115\000\195\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\123\000\001\002\124\000\000\002\000\000\
\\000\000\
\\121\000\210\001\122\000\003\002\000\000\
\\000\000\
\\129\000\004\002\000\000\
\\123\000\136\001\127\000\007\002\129\000\006\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\\105\000\009\002\000\000\
\\003\000\064\000\005\000\027\000\019\000\073\001\093\000\010\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\117\000\234\001\118\000\137\001\119\000\011\002\123\000\001\002\
\\124\000\000\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\\006\000\168\000\016\000\167\000\023\000\166\000\033\000\165\000\
\\034\000\164\000\035\000\163\000\036\000\012\002\037\000\161\000\000\000\
\\000\000\
\\000\000\
\\003\000\083\000\005\000\027\000\022\000\082\000\125\000\013\002\000\000\
\\000\000\
\\000\000\
\\105\000\014\002\000\000\
\\094\000\016\002\095\000\015\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\123\000\206\001\124\000\205\001\127\000\019\002\128\000\018\002\
\\129\000\006\002\000\000\
\\000\000\
\\000\000\
\\003\000\064\000\005\000\027\000\019\000\073\001\093\000\010\002\
\\105\000\021\002\000\000\
\\000\000\
\\000\000\
\\003\000\083\000\005\000\027\000\022\000\082\000\105\000\022\002\
\\125\000\013\002\000\000\
\\094\000\016\002\095\000\023\002\000\000\
\\123\000\001\002\124\000\000\002\127\000\019\002\128\000\024\002\
\\129\000\006\002\000\000\
\\000\000\
\\000\000\
\\000\000\
\"
val numstates = 538
val numrules = 306
val s = ref "" and index = ref 0
val string_to_int = fn () => 
let val i = !index
in index := i+2; Char.ord(String.sub(!s,i)) + Char.ord(String.sub(!s,i+1)) * 256
end
val string_to_list = fn s' =>
    let val len = String.size s'
        fun f () =
           if !index < len then string_to_int() :: f()
           else nil
   in index := 0; s := s'; f ()
   end
val string_to_pairlist = fn (conv_key,conv_entry) =>
     let fun f () =
         case string_to_int()
         of 0 => EMPTY
          | n => PAIR(conv_key (n-1),conv_entry (string_to_int()),f())
     in f
     end
val string_to_pairlist_default = fn (conv_key,conv_entry) =>
    let val conv_row = string_to_pairlist(conv_key,conv_entry)
    in fn () =>
       let val default = conv_entry(string_to_int())
           val row = conv_row()
       in (row,default)
       end
   end
val string_to_table = fn (convert_row,s') =>
    let val len = String.size s'
        fun f ()=
           if !index < len then convert_row() :: f()
           else nil
     in (s := s'; index := 0; f ())
     end
local
  val memo = Array.array(numstates+numrules,ERROR)
  val _ =let fun g i=(Array.update(memo,i,REDUCE(i-numstates)); g(i+1))
       fun f i =
            if i=numstates then g i
            else (Array.update(memo,i,SHIFT (STATE i)); f (i+1))
          in f 0 handle Subscript => ()
          end
in
val entry_to_action = fn 0 => ACCEPT | 1 => ERROR | j => Array.sub(memo,(j-2))
end
val gotoT=Array.fromList(string_to_table(string_to_pairlist(NT,STATE),gotoT))
val actionRows=string_to_table(string_to_pairlist_default(T,entry_to_action),actionRows)
val actionRowNumbers = string_to_list actionRowNumbers
val actionT = let val actionRowLookUp=
let val a=Array.fromList(actionRows) in fn i=>Array.sub(a,i) end
in Array.fromList(map actionRowLookUp actionRowNumbers)
end
in LrTable.mkLrTable {actions=actionT,gotos=gotoT,numRules=numrules,
numStates=numstates,initialState=STATE 0}
end
end
local open Header in
type pos = int
type arg = unit
structure MlyValue = 
struct
datatype svalue = VOID | ntVOID of unit ->  unit
 | CHAR of unit ->  (char) | STRING of unit ->  (string)
 | REAL of unit ->  (string) | WORD of unit ->  (Word32.word)
 | INT of unit ->  (string) | TYVAR of unit ->  (string)
 | LONGID of unit ->  (string) | strexp2node of unit ->  (Strexp.node)
 | strexp2 of unit ->  (Strexp.t)
 | strexp1 of unit ->  (Strexp.t* ( Sigexp.t -> SigConst.t ) *Sigexp.t)
 | strexpnode of unit ->  (Strexp.node)
 | strexp of unit ->  (Strexp.t) | program of unit ->  (Program.t)
 | topdecs of unit ->  (Topdec.t list)
 | topdecnode of unit ->  (Topdec.node)
 | topdec of unit ->  (Topdec.t)
 | strdecnode of unit ->  (Strdec.node)
 | strdec of unit ->  (Strdec.t)
 | strdecsnode of unit ->  (Strdec.node)
 | strdecs of unit ->  (Strdec.t) | withtypes of unit ->  (TypBind.t)
 | dbs of unit ->  (db list) | db of unit ->  (db)
 | datBindNoWithtype of unit ->  (DatBind.t)
 | datBind of unit ->  (DatBind.t)
 | repl of unit ->  (DatatypeRhs.node)
 | datatypeRhsnodeNoWithtype of unit ->  (DatatypeRhs.node)
 | datatypeRhsnode of unit ->  (DatatypeRhs.node)
 | datatypeRhsNoWithtype of unit ->  (DatatypeRhs.t)
 | datatypeRhs of unit ->  (DatatypeRhs.t)
 | fctarg of unit ->  (FctArg.node)
 | funbinds'2 of unit ->  (funbind list)
 | funbinds'1' of unit ->  (funbind whereAnd)
 | funbinds'1 of unit ->  (funbind whereAnd)
 | funbinds' of unit ->  (Strexp.t*funbind list)
 | funbinds of unit ->  (funbind list)
 | andspecs of unit ->  (wherespec list)
 | wherespecs of unit ->  (wherespec list)
 | sigbinds'' of unit ->  (sigbind whereAnd)
 | sigbinds' of unit ->  (sigbind whereAnd)
 | sigbinds of unit ->  (sigbind list)
 | strbinds'1' of unit ->  (strbind whereAnd)
 | strbinds'2 of unit ->  (strbind list)
 | strbinds'1 of unit ->  (strbind whereAnd)
 | strbinds' of unit ->  (Strexp.t*strbind list)
 | strbinds of unit ->  (strbind list) | sdecsPlus of unit ->  (Dec.t)
 | sdec of unit ->  (Dec.t) | sdecs of unit ->  (Dec.t)
 | arg_fct of unit ->  (Strexp.t) | sigconst of unit ->  (SigConst.t)
 | sigexp'node of unit ->  (Sigexp.node)
 | sigexp' of unit ->  (Sigexp.t)
 | sigexpnode of unit ->  (Sigexp.node)
 | sigexp of unit ->  (Sigexp.t) | wherespec of unit ->  (wherespec)
 | longstrideqns of unit ->  (Longstrid.t list)
 | longtyconeqns of unit ->  (Longtycon.t list)
 | exndesc of unit ->  (exndesc) | exndescs of unit ->  (exndesc list)
 | valdesc of unit ->  (valdesc) | valdescs of unit ->  (valdesc list)
 | typdescRhs of unit ->  (Type.t option)
 | typdesc of unit ->  (typdesc) | typdescs of unit ->  (typdesc list)
 | strdescs'' of unit ->  (strdesc whereAnd)
 | strdescs' of unit ->  (strdesc whereAnd)
 | strdescs of unit ->  (strdesc list)
 | sharespec of unit ->  (Equation.node)
 | specnode of unit ->  (Spec.node) | spec of unit ->  (Spec.t)
 | specs of unit ->  (Spec.t) | decsnode of unit ->  (Dec.node)
 | decs of unit ->  (Dec.t) | decnolocal of unit ->  (Dec.node)
 | decnode of unit ->  (Dec.node) | dec of unit ->  (Dec.t)
 | fixity of unit ->  (Fixity.t) | ebrhsnode of unit ->  (EbRhs.node)
 | ebrhs of unit ->  (EbRhs.t) | eb of unit ->  (eb)
 | ebs of unit ->  (eb list)
 | constr of unit ->  (Con.t*Type.t option)
 | constrs of unit ->  ( ( Con.t * Type.t option )  list)
 | tyvar_pc of unit ->  (Tyvar.t list)
 | tyvarseq of unit ->  (Tyvar.t list)
 | tyvars of unit ->  (Tyvar.t list)
 | typBind' of unit ->  (TypBind.node)
 | typBind of unit ->  (TypBind.t) | clause of unit ->  (clause)
 | clauses of unit ->  (clause list)
 | funs of unit ->  (clause list list)
 | rvbs of unit ->  ({ var:Var.t,ty:Type.t option,fixity:Vid.t option,rules:rules }  list)
 | constraint of unit ->  (Type.t option)
 | vbs of unit ->  ( ( Pat.t * Exp.t )  list)
 | opaspat of unit ->  (Pat.t option)
 | commapats of unit ->  (Pat.t list) | pats of unit ->  (Pat.t list)
 | pat_2c of unit ->  (Pat.t list) | patitem of unit ->  (Pat.Item.t)
 | patitems of unit ->  ( ( Pat.Item.t list * bool ) )
 | apat' of unit ->  (Pat.t) | apatnode of unit ->  (Pat.node)
 | apat of unit ->  (Pat.t) | apats of unit ->  (Pat.t list)
 | patnode of unit ->  (Pat.node) | pat of unit ->  (Pat.t)
 | exp_list of unit ->  (Exp.t list) | exp_2c of unit ->  (Exp.t list)
 | ot_list of unit ->  (Exp.t list) | aexp of unit ->  (Exp.node)
 | app_exp1 of unit ->  (Exp.t list)
 | app_exp of unit ->  (Exp.t list) | expnode of unit ->  (Exp.node)
 | exp of unit ->  (Exp.t) | exp_ps of unit ->  (Exp.t list)
 | elabels of unit ->  ( ( Field.t * Exp.t )  list)
 | elabel of unit ->  ( ( Field.t * Exp.t ) )
 | rule of unit ->  (rule) | rules of unit ->  (rules)
 | ty0_pc of unit ->  (Type.t list)
 | tyOpt of unit ->  (Type.t option) | ty1 of unit ->  (Type.t)
 | tynode of unit ->  (Type.node) | ty of unit ->  (Type.t)
 | tuple_ty of unit ->  (Type.t list)
 | ty'node of unit ->  (Type.node) | ty' of unit ->  (Type.t)
 | tlabels of unit ->  ( ( Field.t * Type.t )  list)
 | tlabel of unit ->  ( ( Field.t * Type.t ) )
 | longstrids of unit ->  (Longstrid.t list)
 | longstrid of unit ->  (Longstrid.t)
 | longvidNoEqual of unit ->  (Longvid.t)
 | longvid of unit ->  (Longvid.t) | longcon of unit ->  (Longcon.t)
 | longvarands of unit ->  (Longvar.t list)
 | longvar of unit ->  (Longvar.t)
 | longtycon of unit ->  (Longtycon.t) | fctid of unit ->  (Fctid.t)
 | sigids of unit ->  (Sigid.t list) | sigid of unit ->  (Sigid.t)
 | strid of unit ->  (Strid.t) | field of unit ->  (Field.t)
 | tycon of unit ->  (Tycon.t) | tyvar of unit ->  (Tyvar.t)
 | con of unit ->  (Con.t) | varNoEqual of unit ->  (Var.t)
 | var of unit ->  (Var.t) | vids of unit ->  (Vid.t list)
 | vidNoEqual of unit ->  (Vid.t) | vid of unit ->  (Vid.t)
 | const of unit ->  (Const.t) | longidEqual of unit ->  (string)
 | longid of unit ->  (string) | longidNoAsterisk of unit ->  (string)
 | idNoAsterisk of unit ->  (string) | idEqual of unit ->  (string)
 | id of unit ->  (string) | int of unit ->  (int)
 | digit of unit ->  (int)
end
type svalue = MlyValue.svalue
type result = Program.t
end
structure EC=
struct
open LrTable
val is_keyword =
fn (T 9) => true | (T 10) => true | (T 12) => true | (T 14) => true | 
(T 15) => true | (T 16) => true | (T 17) => true | (T 18) => true | 
(T 20) => true | (T 21) => true | (T 22) => true | (T 23) => true | 
(T 24) => true | (T 25) => true | (T 26) => true | (T 27) => true | 
(T 29) => true | (T 30) => true | (T 31) => true | (T 32) => true | 
(T 33) => true | (T 34) => true | (T 35) => true | (T 36) => true | 
(T 37) => true | (T 38) => true | (T 39) => true | (T 40) => true | 
(T 41) => true | (T 42) => true | (T 43) => true | (T 44) => true | 
(T 45) => true | (T 46) => true | (T 47) => true | (T 48) => true | 
(T 49) => true | (T 50) => true | (T 52) => true | (T 51) => true | 
(T 54) => true | (T 55) => true | (T 66) => true | (T 67) => true | _ => false
val preferred_change = 
(nil
,(T 50) :: nil
)::
(nil
,(T 48) :: nil
)::
(nil
,(T 17) :: nil
)::
(nil
,(T 62) :: nil
)::
(nil
,(T 1) :: nil
)::
((T 23) :: nil
,(T 19) :: nil
)::
((T 19) :: nil
,(T 23) :: nil
)::
((T 10) :: nil
,(T 67) :: nil
)::
((T 57) :: nil
,(T 37) :: nil
)::
((T 1) :: nil
,(T 59) :: nil
)::
((T 59) :: nil
,(T 1) :: nil
)::
(nil
,(T 30) :: (T 2) :: (T 18) :: nil
)::
(nil
,(T 17) :: (T 2) :: nil
)::
nil
val noShift = 
fn (T 0) => true | _ => false
val showTerminal =
fn (T 0) => "EOF"
  | (T 1) => "SEMICOLON"
  | (T 2) => "LONGID"
  | (T 3) => "TYVAR"
  | (T 4) => "INT"
  | (T 5) => "WORD"
  | (T 6) => "REAL"
  | (T 7) => "STRING"
  | (T 8) => "CHAR"
  | (T 9) => "ABSTYPE"
  | (T 10) => "AND"
  | (T 11) => "ARROW"
  | (T 12) => "AS"
  | (T 13) => "BAR"
  | (T 14) => "CASE"
  | (T 15) => "DATATYPE"
  | (T 16) => "DOTDOTDOT"
  | (T 17) => "ELSE"
  | (T 18) => "END"
  | (T 19) => "EQUALOP"
  | (T 20) => "EQTYPE"
  | (T 21) => "EXCEPTION"
  | (T 22) => "DO"
  | (T 23) => "DARROW"
  | (T 24) => "FN"
  | (T 25) => "FUN"
  | (T 26) => "FUNCTOR"
  | (T 27) => "HANDLE"
  | (T 28) => "HASH"
  | (T 29) => "IF"
  | (T 30) => "IN"
  | (T 31) => "INCLUDE"
  | (T 32) => "INFIX"
  | (T 33) => "INFIXR"
  | (T 34) => "LET"
  | (T 35) => "LOCAL"
  | (T 36) => "NONFIX"
  | (T 37) => "OF"
  | (T 38) => "OP"
  | (T 39) => "OPEN"
  | (T 40) => "OVERLOAD"
  | (T 41) => "RAISE"
  | (T 42) => "REC"
  | (T 43) => "SHARING"
  | (T 44) => "SIG"
  | (T 45) => "SIGNATURE"
  | (T 46) => "STRUCT"
  | (T 47) => "STRUCTURE"
  | (T 48) => "THEN"
  | (T 49) => "TYPE"
  | (T 50) => "VAL"
  | (T 51) => "WHERE"
  | (T 52) => "WHILE"
  | (T 53) => "WILD"
  | (T 54) => "WITH"
  | (T 55) => "WITHTYPE"
  | (T 56) => "ASTERISK"
  | (T 57) => "COLON"
  | (T 58) => "COLONGT"
  | (T 59) => "COMMA"
  | (T 60) => "LBRACE"
  | (T 61) => "LBRACKET"
  | (T 62) => "LPAREN"
  | (T 63) => "RBRACE"
  | (T 64) => "RBRACKET"
  | (T 65) => "RPAREN"
  | (T 66) => "ORELSE"
  | (T 67) => "ANDALSO"
  | (T 68) => "FUNSIG"
  | (T 69) => "PRIM"
  | _ => "bogus-term"
local open Header in
val errtermvalue=
fn _ => MlyValue.VOID
end
val terms = (T 0) :: (T 1) :: (T 9) :: (T 10) :: (T 11) :: (T 12) :: 
(T 13) :: (T 14) :: (T 15) :: (T 16) :: (T 17) :: (T 18) :: (T 19) :: 
(T 20) :: (T 21) :: (T 22) :: (T 23) :: (T 24) :: (T 25) :: (T 26) :: 
(T 27) :: (T 28) :: (T 29) :: (T 30) :: (T 31) :: (T 32) :: (T 33) :: 
(T 34) :: (T 35) :: (T 36) :: (T 37) :: (T 38) :: (T 39) :: (T 40) :: 
(T 41) :: (T 42) :: (T 43) :: (T 44) :: (T 45) :: (T 46) :: (T 47) :: 
(T 48) :: (T 49) :: (T 50) :: (T 51) :: (T 52) :: (T 53) :: (T 54) :: 
(T 55) :: (T 56) :: (T 57) :: (T 58) :: (T 59) :: (T 60) :: (T 61) :: 
(T 62) :: (T 63) :: (T 64) :: (T 65) :: (T 66) :: (T 67) :: (T 68) :: 
(T 69) :: nil
end
structure Actions =
struct 
exception mlyAction of int
local open Header in
val actions = 
fn (i392,defaultPos,stack,
    (()):arg) =>
case (i392,stack)
of (0,(_,(MlyValue.topdecs topdecs1,topdecs1left,topdecs1right))::
rest671) => let val result=MlyValue.program(fn _ => let val topdecs
 as topdecs1=topdecs1 ()
 in (Program.T topdecs) end
)
 in (LrTable.NT 147,(result,topdecs1left,topdecs1right),rest671) end
| (1,(_,(MlyValue.topdec topdec1,topdec1left,topdec1right))::rest671)
 => let val result=MlyValue.topdecs(fn _ => let val topdec as topdec1=
topdec1 ()
 in ([topdec]) end
)
 in (LrTable.NT 146,(result,topdec1left,topdec1right),rest671) end
| (2,(_,(_,SEMICOLON1left,SEMICOLON1right))::rest671) => let val 
result=MlyValue.topdecs(fn _ => ([]))
 in (LrTable.NT 146,(result,SEMICOLON1left,SEMICOLON1right),rest671)
 end
| (3,(_,(MlyValue.topdecs topdecs1,_,topdecs1right))::(_,(_,
SEMICOLON1left,_))::rest671) => let val result=MlyValue.topdecs(fn _
 => let val topdecs as topdecs1=topdecs1 ()
 in (topdecs) end
)
 in (LrTable.NT 146,(result,SEMICOLON1left,topdecs1right),rest671) end
| (4,(_,(MlyValue.topdecs topdecs1,_,topdecs1right))::(_,(
MlyValue.topdec topdec1,topdec1left,_))::rest671) => let val result=
MlyValue.topdecs(fn _ => let val topdec as topdec1=topdec1 ()
val topdecs as topdecs1=topdecs1 ()
 in (topdec::topdecs) end
)
 in (LrTable.NT 146,(result,topdec1left,topdecs1right),rest671) end
| (5,(_,(MlyValue.topdecnode topdecnode1,topdecnodeleft as 
topdecnode1left,topdecnoderight as topdecnode1right))::rest671) => 
let val result=MlyValue.topdec(fn _ => let val topdecnode as 
topdecnode1=topdecnode1 ()
 in (
Topdec.makeRegion'(topdecnode,
					topdecnodeleft,
					topdecnoderight)
) end
)
 in (LrTable.NT 144,(result,topdecnode1left,topdecnode1right),rest671)
 end
| (6,(_,(MlyValue.strdec strdec1,strdec1left,strdec1right))::rest671)
 => let val result=MlyValue.topdecnode(fn _ => let val strdec as 
strdec1=strdec1 ()
 in (Topdec.Strdec strdec) end
)
 in (LrTable.NT 145,(result,strdec1left,strdec1right),rest671) end
| (7,(_,(MlyValue.sigbinds sigbinds1,_,sigbinds1right))::(_,(_,
SIGNATURE1left,_))::rest671) => let val result=MlyValue.topdecnode(fn 
_ => let val sigbinds as sigbinds1=sigbinds1 ()
 in (Topdec.Signature sigbinds) end
)
 in (LrTable.NT 145,(result,SIGNATURE1left,sigbinds1right),rest671)
 end
| (8,(_,(MlyValue.funbinds funbinds1,_,funbinds1right))::(_,(_,
FUNCTOR1left,_))::rest671) => let val result=MlyValue.topdecnode(fn _
 => let val funbinds as funbinds1=funbinds1 ()
 in (Topdec.Functor funbinds) end
)
 in (LrTable.NT 145,(result,FUNCTOR1left,funbinds1right),rest671) end
| (9,(_,(MlyValue.strdecsnode strdecsnode1,strdecsnodeleft as 
strdecsnode1left,strdecsnoderight as strdecsnode1right))::rest671) => 
let val result=MlyValue.strdecs(fn _ => let val strdecsnode as 
strdecsnode1=strdecsnode1 ()
 in (
Strdec.makeRegion'(strdecsnode,
					  strdecsnodeleft, strdecsnoderight)
) end
)
 in (LrTable.NT 140,(result,strdecsnode1left,strdecsnode1right),
rest671) end
| (10,rest671) => let val result=MlyValue.strdecsnode(fn _ => (
Strdec.Seq[]))
 in (LrTable.NT 141,(result,defaultPos,defaultPos),rest671) end
| (11,(_,(MlyValue.strdecs strdecs1,_,strdecs1right))::(_,(_,
SEMICOLON1left,_))::rest671) => let val result=MlyValue.strdecsnode(
fn _ => let val strdecs as strdecs1=strdecs1 ()
 in (Strdec.Seq[strdecs]) end
)
 in (LrTable.NT 141,(result,SEMICOLON1left,strdecs1right),rest671) end
| (12,(_,(MlyValue.strdecs strdecs1,_,strdecs1right))::(_,(
MlyValue.strdec strdec1,strdec1left,_))::rest671) => let val result=
MlyValue.strdecsnode(fn _ => let val strdec as strdec1=strdec1 ()
val strdecs as strdecs1=strdecs1 ()
 in (Strdec.Seq[strdec,strdecs]) end
)
 in (LrTable.NT 141,(result,strdec1left,strdecs1right),rest671) end
| (13,(_,(MlyValue.strdecnode strdecnode1,strdecnodeleft as 
strdecnode1left,strdecnoderight as strdecnode1right))::rest671) => 
let val result=MlyValue.strdec(fn _ => let val strdecnode as 
strdecnode1=strdecnode1 ()
 in (
Strdec.makeRegion'(strdecnode,
					strdecnodeleft, strdecnoderight))
 end
)
 in (LrTable.NT 142,(result,strdecnode1left,strdecnode1right),rest671)
 end
| (14,(_,(MlyValue.strbinds strbinds1,_,strbinds1right))::(_,(_,
STRUCTURE1left,_))::rest671) => let val result=MlyValue.strdecnode(fn 
_ => let val strbinds as strbinds1=strbinds1 ()
 in (Strdec.Structure strbinds) end
)
 in (LrTable.NT 143,(result,STRUCTURE1left,strbinds1right),rest671)
 end
| (15,(_,(_,_,END1right))::(_,(MlyValue.strdecs strdecs2,_,_))::_::(_,
(MlyValue.strdecs strdecs1,_,_))::(_,(_,LOCAL1left,_))::rest671) => 
let val result=MlyValue.strdecnode(fn _ => let val strdecs1=strdecs1 
()
val strdecs2=strdecs2 ()
 in (Strdec.Local(strdecs1,strdecs2)) end
)
 in (LrTable.NT 143,(result,LOCAL1left,END1right),rest671) end
| (16,(_,(MlyValue.decnolocal decnolocal1,decnolocalleft as 
decnolocal1left,decnolocalright as decnolocal1right))::rest671) => 
let val result=MlyValue.strdecnode(fn _ => let val decnolocal as 
decnolocal1=decnolocal1 ()
 in (
Strdec.Core
		  (Dec.makeRegion'(decnolocal, decnolocalleft, decnolocalright))
) end
)
 in (LrTable.NT 143,(result,decnolocal1left,decnolocal1right),rest671)
 end
| (17,(_,(MlyValue.strbinds' strbinds'1,_,strbinds'1right))::_::(_,(
MlyValue.sigconst sigconst1,_,_))::(_,(MlyValue.strid strid1,
strid1left,_))::rest671) => let val result=MlyValue.strbinds(fn _ => 
let val strid as strid1=strid1 ()
val sigconst as sigconst1=sigconst1 ()
val strbinds' as strbinds'1=strbinds'1 ()
 in (
let val (def,strbinds) = strbinds'
	    in {name = strid, def = def, constraint = sigconst}
	       :: strbinds
	    end
) end
)
 in (LrTable.NT 114,(result,strid1left,strbinds'1right),rest671) end
| (18,(_,(MlyValue.strbinds'1 strbinds'11,_,strbinds'11right))::(_,(
MlyValue.strexp1 strexp11,strexp11left,_))::rest671) => let val result
=MlyValue.strbinds'(fn _ => let val strexp1 as strexp11=strexp11 ()
val strbinds'1 as strbinds'11=strbinds'11 ()
 in (augment1(strexp1, strbinds'1)) end
)
 in (LrTable.NT 115,(result,strexp11left,strbinds'11right),rest671)
 end
| (19,(_,(MlyValue.strbinds'2 strbinds'21,_,strbinds'21right))::(_,(
MlyValue.strexp2 strexp21,strexp21left,_))::rest671) => let val result
=MlyValue.strbinds'(fn _ => let val strexp2 as strexp21=strexp21 ()
val strbinds'2 as strbinds'21=strbinds'21 ()
 in ((strexp2,strbinds'2)) end
)
 in (LrTable.NT 115,(result,strexp21left,strbinds'21right),rest671)
 end
| (20,(_,(MlyValue.strbinds'2 strbinds'21,strbinds'2left as 
strbinds'21left,strbinds'21right))::rest671) => let val result=
MlyValue.strbinds'1(fn _ => let val strbinds'2 as strbinds'21=
strbinds'21 ()
 in (([], strbinds'2left, strbinds'2)) end
)
 in (LrTable.NT 116,(result,strbinds'21left,strbinds'21right),rest671)
 end
| (21,(_,(MlyValue.strbinds'1' strbinds'1'1,_,strbinds'1'1right))::(_,
(MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.strbinds'1(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val strbinds'1' as strbinds'1'1=strbinds'1'1 ()
 in (cons1(wherespec,strbinds'1')) end
)
 in (LrTable.NT 116,(result,WHERE1left,strbinds'1'1right),rest671) end
| (22,(_,(MlyValue.strbinds'1 strbinds'11,strbinds'11left,
strbinds'11right))::rest671) => let val result=MlyValue.strbinds'1'(
fn _ => let val strbinds'1 as strbinds'11=strbinds'11 ()
 in (strbinds'1) end
)
 in (LrTable.NT 118,(result,strbinds'11left,strbinds'11right),rest671)
 end
| (23,(_,(MlyValue.strbinds'1' strbinds'1'1,_,strbinds'1'1right))::(_,
(MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.strbinds'1'(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val strbinds'1' as strbinds'1'1=strbinds'1'1 ()
 in (cons1(wherespec,strbinds'1')) end
)
 in (LrTable.NT 118,(result,AND1left,strbinds'1'1right),rest671) end
| (24,rest671) => let val result=MlyValue.strbinds'2(fn _ => ([]))
 in (LrTable.NT 117,(result,defaultPos,defaultPos),rest671) end
| (25,(_,(MlyValue.strbinds strbinds1,_,strbinds1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.strbinds'2(fn _ => 
let val strbinds as strbinds1=strbinds1 ()
 in (strbinds) end
)
 in (LrTable.NT 117,(result,AND1left,strbinds1right),rest671) end
| (26,(_,(MlyValue.strexpnode strexpnode1,strexpnodeleft as 
strexpnode1left,strexpnoderight as strexpnode1right))::rest671) => 
let val result=MlyValue.strexp(fn _ => let val strexpnode as 
strexpnode1=strexpnode1 ()
 in (
Strexp.makeRegion'(strexpnode,
					strexpnodeleft, strexpnoderight))
 end
)
 in (LrTable.NT 148,(result,strexpnode1left,strexpnode1right),rest671)
 end
| (27,(_,(MlyValue.strexp1 strexp11,strexp11left,strexp11right))::
rest671) => let val result=MlyValue.strexpnode(fn _ => let val strexp1
 as strexp11=strexp11 ()
 in (
let val (strexp,sigconst,sigexp) = strexp1
			      in Strexp.Constrained(strexp, sigconst sigexp)
			      end
) end
)
 in (LrTable.NT 149,(result,strexp11left,strexp11right),rest671) end
| (28,(_,(MlyValue.wherespecs wherespecs1,_,wherespecsright as 
wherespecs1right))::(_,(MlyValue.strexp1 strexp11,strexp11left,_))::
rest671) => let val result=MlyValue.strexpnode(fn _ => let val strexp1
 as strexp11=strexp11 ()
val wherespecs as wherespecs1=wherespecs1 ()
 in (
let val (strexp,sigconst,sigexp) = strexp1
			      in Strexp.Constrained
				 (strexp,
				  sigconst(Sigexp.makeRegion'
					   (Sigexp.Where(sigexp,wherespecs),
					    Sigexp.left sigexp,
					    wherespecsright)))
			      end
) end
)
 in (LrTable.NT 149,(result,strexp11left,wherespecs1right),rest671)
 end
| (29,(_,(MlyValue.strexp2node strexp2node1,strexp2node1left,
strexp2node1right))::rest671) => let val result=MlyValue.strexpnode(
fn _ => let val strexp2node as strexp2node1=strexp2node1 ()
 in (strexp2node) end
)
 in (LrTable.NT 149,(result,strexp2node1left,strexp2node1right),
rest671) end
| (30,(_,(MlyValue.sigexp' sigexp'1,_,sigexp'1right))::_::(_,(
MlyValue.strexp strexp1,strexp1left,_))::rest671) => let val result=
MlyValue.strexp1(fn _ => let val strexp as strexp1=strexp1 ()
val sigexp' as sigexp'1=sigexp'1 ()
 in ((strexp,SigConst.Transparent,sigexp')) end
)
 in (LrTable.NT 150,(result,strexp1left,sigexp'1right),rest671) end
| (31,(_,(MlyValue.sigexp' sigexp'1,_,sigexp'1right))::_::(_,(
MlyValue.strexp strexp1,strexp1left,_))::rest671) => let val result=
MlyValue.strexp1(fn _ => let val strexp as strexp1=strexp1 ()
val sigexp' as sigexp'1=sigexp'1 ()
 in ((strexp,SigConst.Opaque,sigexp')) end
)
 in (LrTable.NT 150,(result,strexp1left,sigexp'1right),rest671) end
| (32,(_,(MlyValue.strexp2node strexp2node1,strexp2nodeleft as 
strexp2node1left,strexp2noderight as strexp2node1right))::rest671) => 
let val result=MlyValue.strexp2(fn _ => let val strexp2node as 
strexp2node1=strexp2node1 ()
 in (
Strexp.makeRegion'(strexp2node,
					  strexp2nodeleft, strexp2noderight)
) end
)
 in (LrTable.NT 151,(result,strexp2node1left,strexp2node1right),
rest671) end
| (33,(_,(MlyValue.longid longid1,longidleft as longid1left,
longidright as longid1right))::rest671) => let val result=
MlyValue.strexp2node(fn _ => let val longid as longid1=longid1 ()
 in (
Strexp.Var(Longstrid.fromString
						 (longid,
						  reg(longidleft,longidright)))
) end
)
 in (LrTable.NT 152,(result,longid1left,longid1right),rest671) end
| (34,(_,(_,_,END1right))::(_,(MlyValue.strdecs strdecs1,_,_))::(_,(_,
STRUCT1left,_))::rest671) => let val result=MlyValue.strexp2node(fn _
 => let val strdecs as strdecs1=strdecs1 ()
 in (Strexp.Struct strdecs) end
)
 in (LrTable.NT 152,(result,STRUCT1left,END1right),rest671) end
| (35,(_,(MlyValue.arg_fct arg_fct1,_,arg_fct1right))::(_,(
MlyValue.longid longid1,longidleft as longid1left,longidright))::
rest671) => let val result=MlyValue.strexp2node(fn _ => let val longid
 as longid1=longid1 ()
val arg_fct as arg_fct1=arg_fct1 ()
 in (
Strexp.App(Fctid.fromString
		      (ensureNonqualified(longid, reg(longidleft, longidright))),
		      arg_fct)
) end
)
 in (LrTable.NT 152,(result,longid1left,arg_fct1right),rest671) end
| (36,(_,(_,_,END1right))::(_,(MlyValue.strexp strexp1,_,_))::_::(_,(
MlyValue.strdecs strdecs1,_,_))::(_,(_,LET1left,_))::rest671) => let 
val result=MlyValue.strexp2node(fn _ => let val strdecs as strdecs1=
strdecs1 ()
val strexp as strexp1=strexp1 ()
 in (Strexp.Let(strdecs,strexp)) end
)
 in (LrTable.NT 152,(result,LET1left,END1right),rest671) end
| (37,(_,(_,_,RPAREN1right))::(_,(MlyValue.strexp strexp1,_,_))::(_,(_
,LPAREN1left,_))::rest671) => let val result=MlyValue.arg_fct(fn _ => 
let val strexp as strexp1=strexp1 ()
 in (strexp) end
)
 in (LrTable.NT 110,(result,LPAREN1left,RPAREN1right),rest671) end
| (38,(_,(_,_,RPAREN1right))::(_,(MlyValue.strdecs strdecs1,
strdecsleft,strdecsright))::(_,(_,LPAREN1left,_))::rest671) => let 
val result=MlyValue.arg_fct(fn _ => let val strdecs as strdecs1=
strdecs1 ()
 in (
Strexp.makeRegion'
				  (Strexp.Struct strdecs,
				   strdecsleft, strdecsright)
) end
)
 in (LrTable.NT 110,(result,LPAREN1left,RPAREN1right),rest671) end
| (39,(_,(MlyValue.sigexp' sigexp'1,sigexp'1left,sigexp'1right))::
rest671) => let val result=MlyValue.sigexp(fn _ => let val sigexp' as 
sigexp'1=sigexp'1 ()
 in (sigexp') end
)
 in (LrTable.NT 105,(result,sigexp'1left,sigexp'1right),rest671) end
| (40,(_,(MlyValue.wherespecs wherespecs1,_,wherespecsright as 
wherespecs1right))::(_,(MlyValue.sigexp' sigexp'1,sigexp'left as 
sigexp'1left,_))::rest671) => let val result=MlyValue.sigexp(fn _ => 
let val sigexp' as sigexp'1=sigexp'1 ()
val wherespecs as wherespecs1=wherespecs1 ()
 in (
Sigexp.makeRegion'
			      (Sigexp.Where(sigexp',wherespecs),
			       sigexp'left, wherespecsright)
) end
)
 in (LrTable.NT 105,(result,sigexp'1left,wherespecs1right),rest671)
 end
| (41,(_,(MlyValue.wherespec wherespec1,_,wherespec1right))::(_,(_,
WHERE1left,_))::rest671) => let val result=MlyValue.wherespecs(fn _
 => let val wherespec as wherespec1=wherespec1 ()
 in ([wherespec]) end
)
 in (LrTable.NT 122,(result,WHERE1left,wherespec1right),rest671) end
| (42,(_,(MlyValue.wherespecs wherespecs1,_,wherespecs1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.wherespecs(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val wherespecs as wherespecs1=wherespecs1 ()
 in (wherespec::wherespecs) end
)
 in (LrTable.NT 122,(result,WHERE1left,wherespecs1right),rest671) end
| (43,(_,(MlyValue.andspecs andspecs1,_,andspecs1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.wherespecs(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val andspecs as andspecs1=andspecs1 ()
 in (wherespec::andspecs) end
)
 in (LrTable.NT 122,(result,WHERE1left,andspecs1right),rest671) end
| (44,(_,(MlyValue.wherespec wherespec1,_,wherespec1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.andspecs(fn _ => 
let val wherespec as wherespec1=wherespec1 ()
 in ([wherespec]) end
)
 in (LrTable.NT 123,(result,AND1left,wherespec1right),rest671) end
| (45,(_,(MlyValue.andspecs andspecs1,_,andspecs1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.andspecs(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val andspecs as andspecs1=andspecs1 ()
 in (wherespec::andspecs) end
)
 in (LrTable.NT 123,(result,AND1left,andspecs1right),rest671) end
| (46,(_,(MlyValue.wherespecs wherespecs1,_,wherespecs1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.andspecs(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val wherespecs as wherespecs1=wherespecs1 ()
 in (wherespec::wherespecs) end
)
 in (LrTable.NT 123,(result,AND1left,wherespecs1right),rest671) end
| (47,(_,(MlyValue.sigbinds' sigbinds'1,_,sigbinds'1right))::(_,(
MlyValue.sigexp' sigexp'1,_,_))::_::(_,(MlyValue.sigid sigid1,
sigid1left,_))::rest671) => let val result=MlyValue.sigbinds(fn _ => 
let val sigid as sigid1=sigid1 ()
val sigexp' as sigexp'1=sigexp'1 ()
val sigbinds' as sigbinds'1=sigbinds'1 ()
 in (augment(sigid, sigexp', sigbinds')) end
)
 in (LrTable.NT 119,(result,sigid1left,sigbinds'1right),rest671) end
| (48,(_,(MlyValue.sigexp'node sigexp'node1,sigexp'nodeleft as 
sigexp'node1left,sigexp'noderight as sigexp'node1right))::rest671) => 
let val result=MlyValue.sigexp'(fn _ => let val sigexp'node as 
sigexp'node1=sigexp'node1 ()
 in (
Sigexp.makeRegion'(sigexp'node,
					  sigexp'nodeleft,
					  sigexp'noderight)
) end
)
 in (LrTable.NT 107,(result,sigexp'node1left,sigexp'node1right),
rest671) end
| (49,(_,(MlyValue.sigid sigid1,sigid1left,sigid1right))::rest671) => 
let val result=MlyValue.sigexp'node(fn _ => let val sigid as sigid1=
sigid1 ()
 in (Sigexp.Var sigid) end
)
 in (LrTable.NT 108,(result,sigid1left,sigid1right),rest671) end
| (50,(_,(_,_,END1right))::(_,(MlyValue.specs specs1,_,_))::(_,(_,
SIG1left,_))::rest671) => let val result=MlyValue.sigexp'node(fn _ => 
let val specs as specs1=specs1 ()
 in (Sigexp.Spec specs) end
)
 in (LrTable.NT 108,(result,SIG1left,END1right),rest671) end
| (51,rest671) => let val result=MlyValue.sigbinds'(fn _ => (
([], defaultPos, [])))
 in (LrTable.NT 120,(result,defaultPos,defaultPos),rest671) end
| (52,(_,(MlyValue.sigbinds sigbinds1,_,sigbinds1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.sigbinds'(fn _ => 
let val sigbinds as sigbinds1=sigbinds1 ()
 in (([], defaultPos, sigbinds)) end
)
 in (LrTable.NT 120,(result,AND1left,sigbinds1right),rest671) end
| (53,(_,(MlyValue.sigbinds'' sigbinds''1,_,sigbinds''1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.sigbinds'(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val sigbinds'' as sigbinds''1=sigbinds''1 ()
 in (cons1(wherespec,sigbinds'')) end
)
 in (LrTable.NT 120,(result,WHERE1left,sigbinds''1right),rest671) end
| (54,(_,(MlyValue.sigbinds' sigbinds'1,sigbinds'1left,sigbinds'1right
))::rest671) => let val result=MlyValue.sigbinds''(fn _ => let val 
sigbinds' as sigbinds'1=sigbinds'1 ()
 in (sigbinds') end
)
 in (LrTable.NT 121,(result,sigbinds'1left,sigbinds'1right),rest671)
 end
| (55,(_,(MlyValue.sigbinds'' sigbinds''1,_,sigbinds''1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.sigbinds''(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val sigbinds'' as sigbinds''1=sigbinds''1 ()
 in (cons1(wherespec,sigbinds'')) end
)
 in (LrTable.NT 121,(result,AND1left,sigbinds''1right),rest671) end
| (56,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.longtycon 
longtycon1,_,_))::(_,(MlyValue.tyvars tyvars1,_,_))::(_,(_,TYPE1left,_
))::rest671) => let val result=MlyValue.wherespec(fn _ => let val 
tyvars as tyvars1=tyvars1 ()
val longtycon as longtycon1=longtycon1 ()
val ty as ty1=ty1 ()
 in ({tyvars = tyvars,
						 longtycon = longtycon,
						 ty = ty})
 end
)
 in (LrTable.NT 104,(result,TYPE1left,ty1right),rest671) end
| (57,rest671) => let val result=MlyValue.sigconst(fn _ => (
SigConst.None))
 in (LrTable.NT 109,(result,defaultPos,defaultPos),rest671) end
| (58,(_,(MlyValue.sigexp sigexp1,_,sigexp1right))::(_,(_,COLON1left,_
))::rest671) => let val result=MlyValue.sigconst(fn _ => let val 
sigexp as sigexp1=sigexp1 ()
 in (SigConst.Transparent sigexp) end
)
 in (LrTable.NT 109,(result,COLON1left,sigexp1right),rest671) end
| (59,(_,(MlyValue.sigexp sigexp1,_,sigexp1right))::(_,(_,COLONGT1left
,_))::rest671) => let val result=MlyValue.sigconst(fn _ => let val 
sigexp as sigexp1=sigexp1 ()
 in (SigConst.Opaque sigexp) end
)
 in (LrTable.NT 109,(result,COLONGT1left,sigexp1right),rest671) end
| (60,rest671) => let val result=MlyValue.specs(fn _ => (
Spec.makeRegion(Spec.Empty, Region.bogus)))
 in (LrTable.NT 88,(result,defaultPos,defaultPos),rest671) end
| (61,(_,(MlyValue.specs specs1,_,specs1right))::(_,(_,SEMICOLON1left,
_))::rest671) => let val result=MlyValue.specs(fn _ => let val specs
 as specs1=specs1 ()
 in (specs) end
)
 in (LrTable.NT 88,(result,SEMICOLON1left,specs1right),rest671) end
| (62,(_,(MlyValue.specs specs1,_,specs1right))::(_,(MlyValue.spec 
spec1,spec1left,_))::rest671) => let val result=MlyValue.specs(fn _
 => let val spec as spec1=spec1 ()
val specs as specs1=specs1 ()
 in (Spec.seq(spec, specs)) end
)
 in (LrTable.NT 88,(result,spec1left,specs1right),rest671) end
| (63,(_,(MlyValue.specnode specnode1,specnodeleft as specnode1left,
specnoderight as specnode1right))::rest671) => let val result=
MlyValue.spec(fn _ => let val specnode as specnode1=specnode1 ()
 in (Spec.makeRegion'(specnode, specnodeleft, specnoderight)) end
)
 in (LrTable.NT 89,(result,specnode1left,specnode1right),rest671) end
| (64,(_,(MlyValue.valdescs valdescs1,_,valdescs1right))::(_,(_,
VAL1left,_))::rest671) => let val result=MlyValue.specnode(fn _ => 
let val valdescs as valdescs1=valdescs1 ()
 in (Spec.Val valdescs) end
)
 in (LrTable.NT 90,(result,VAL1left,valdescs1right),rest671) end
| (65,(_,(MlyValue.typdescs typdescs1,_,typdescs1right))::(_,(_,
TYPE1left,_))::rest671) => let val result=MlyValue.specnode(fn _ => 
let val typdescs as typdescs1=typdescs1 ()
 in (Spec.Type typdescs) end
)
 in (LrTable.NT 90,(result,TYPE1left,typdescs1right),rest671) end
| (66,(_,(MlyValue.typdescs typdescs1,_,typdescs1right))::(_,(_,
EQTYPE1left,_))::rest671) => let val result=MlyValue.specnode(fn _ => 
let val typdescs as typdescs1=typdescs1 ()
 in (Spec.Eqtype typdescs) end
)
 in (LrTable.NT 90,(result,EQTYPE1left,typdescs1right),rest671) end
| (67,(_,(MlyValue.datatypeRhsNoWithtype datatypeRhsNoWithtype1,_,
datatypeRhsNoWithtype1right))::(_,(_,DATATYPE1left,_))::rest671) => 
let val result=MlyValue.specnode(fn _ => let val datatypeRhsNoWithtype
 as datatypeRhsNoWithtype1=datatypeRhsNoWithtype1 ()
 in (Spec.Datatype datatypeRhsNoWithtype) end
)
 in (LrTable.NT 90,(result,DATATYPE1left,datatypeRhsNoWithtype1right),
rest671) end
| (68,(_,(MlyValue.exndescs exndescs1,_,exndescs1right))::(_,(_,
EXCEPTION1left,_))::rest671) => let val result=MlyValue.specnode(fn _
 => let val exndescs as exndescs1=exndescs1 ()
 in (Spec.Exception exndescs) end
)
 in (LrTable.NT 90,(result,EXCEPTION1left,exndescs1right),rest671) end
| (69,(_,(MlyValue.strdescs strdescs1,_,strdescs1right))::(_,(_,
STRUCTURE1left,_))::rest671) => let val result=MlyValue.specnode(fn _
 => let val strdescs as strdescs1=strdescs1 ()
 in (Spec.Structure strdescs) end
)
 in (LrTable.NT 90,(result,STRUCTURE1left,strdescs1right),rest671) end
| (70,(_,(MlyValue.sigexp sigexp1,_,sigexp1right))::(_,(_,INCLUDE1left
,_))::rest671) => let val result=MlyValue.specnode(fn _ => let val 
sigexp as sigexp1=sigexp1 ()
 in (Spec.IncludeSigexp sigexp) end
)
 in (LrTable.NT 90,(result,INCLUDE1left,sigexp1right),rest671) end
| (71,(_,(MlyValue.sigids sigids1,_,sigids1right))::(_,(MlyValue.sigid
 sigid1,_,_))::(_,(_,INCLUDE1left,_))::rest671) => let val result=
MlyValue.specnode(fn _ => let val sigid as sigid1=sigid1 ()
val sigids as sigids1=sigids1 ()
 in (Spec.IncludeSigids(sigid::sigids)) end
)
 in (LrTable.NT 90,(result,INCLUDE1left,sigids1right),rest671) end
| (72,(_,(MlyValue.sharespec sharespec1,sharespecleft as 
sharespec1left,sharespecright as sharespec1right))::rest671) => let 
val result=MlyValue.specnode(fn _ => let val sharespec as sharespec1=
sharespec1 ()
 in (
Spec.Sharing
				 {spec = Spec.makeRegion'(Spec.Empty,
							  sharespecleft,
							  sharespecright),
				  equations =
				  [Equation.makeRegion'(sharespec,
							sharespecleft,
							sharespecright)]}
) end
)
 in (LrTable.NT 90,(result,sharespec1left,sharespec1right),rest671)
 end
| (73,(_,(MlyValue.longtyconeqns longtyconeqns1,_,longtyconeqns1right)
)::_::(_,(_,SHARING1left,_))::rest671) => let val result=
MlyValue.sharespec(fn _ => let val longtyconeqns as longtyconeqns1=
longtyconeqns1 ()
 in (Equation.Type longtyconeqns) end
)
 in (LrTable.NT 91,(result,SHARING1left,longtyconeqns1right),rest671)
 end
| (74,(_,(MlyValue.longstrideqns longstrideqns1,_,longstrideqns1right)
)::(_,(_,SHARING1left,_))::rest671) => let val result=
MlyValue.sharespec(fn _ => let val longstrideqns as longstrideqns1=
longstrideqns1 ()
 in (Equation.Structure longstrideqns) end
)
 in (LrTable.NT 91,(result,SHARING1left,longstrideqns1right),rest671)
 end
| (75,(_,(MlyValue.longstrid longstrid2,_,longstrid2right))::_::(_,(
MlyValue.longstrid longstrid1,longstrid1left,_))::rest671) => let val 
result=MlyValue.longstrideqns(fn _ => let val longstrid1=longstrid1 ()
val longstrid2=longstrid2 ()
 in ([longstrid1,longstrid2]) end
)
 in (LrTable.NT 103,(result,longstrid1left,longstrid2right),rest671)
 end
| (76,(_,(MlyValue.longstrideqns longstrideqns1,_,longstrideqns1right)
)::_::(_,(MlyValue.longstrid longstrid1,longstrid1left,_))::rest671)
 => let val result=MlyValue.longstrideqns(fn _ => let val longstrid
 as longstrid1=longstrid1 ()
val longstrideqns as longstrideqns1=longstrideqns1 ()
 in (longstrid :: longstrideqns) end
)
 in (LrTable.NT 103,(result,longstrid1left,longstrideqns1right),
rest671) end
| (77,(_,(MlyValue.longtycon longtycon2,_,longtycon2right))::_::(_,(
MlyValue.longtycon longtycon1,longtycon1left,_))::rest671) => let val 
result=MlyValue.longtyconeqns(fn _ => let val longtycon1=longtycon1 ()
val longtycon2=longtycon2 ()
 in ([longtycon1,longtycon2]) end
)
 in (LrTable.NT 102,(result,longtycon1left,longtycon2right),rest671)
 end
| (78,(_,(MlyValue.longtyconeqns longtyconeqns1,_,longtyconeqns1right)
)::_::(_,(MlyValue.longtycon longtycon1,longtycon1left,_))::rest671)
 => let val result=MlyValue.longtyconeqns(fn _ => let val longtycon
 as longtycon1=longtycon1 ()
val longtyconeqns as longtyconeqns1=longtyconeqns1 ()
 in (longtycon :: longtyconeqns) end
)
 in (LrTable.NT 102,(result,longtycon1left,longtyconeqns1right),
rest671) end
| (79,(_,(MlyValue.strdescs' strdescs'1,_,strdescs'1right))::(_,(
MlyValue.sigexp' sigexp'1,_,_))::_::(_,(MlyValue.strid strid1,
strid1left,_))::rest671) => let val result=MlyValue.strdescs(fn _ => 
let val strid as strid1=strid1 ()
val sigexp' as sigexp'1=sigexp'1 ()
val strdescs' as strdescs'1=strdescs'1 ()
 in (augment(strid, sigexp', strdescs')) end
)
 in (LrTable.NT 92,(result,strid1left,strdescs'1right),rest671) end
| (80,rest671) => let val result=MlyValue.strdescs'(fn _ => (
([], defaultPos, [])))
 in (LrTable.NT 93,(result,defaultPos,defaultPos),rest671) end
| (81,(_,(MlyValue.strdescs strdescs1,_,strdescs1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.strdescs'(fn _ => 
let val strdescs as strdescs1=strdescs1 ()
 in (([], defaultPos, strdescs)) end
)
 in (LrTable.NT 93,(result,AND1left,strdescs1right),rest671) end
| (82,(_,(MlyValue.strdescs'' strdescs''1,_,strdescs''1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.strdescs'(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val strdescs'' as strdescs''1=strdescs''1 ()
 in (cons1(wherespec, strdescs'')) end
)
 in (LrTable.NT 93,(result,WHERE1left,strdescs''1right),rest671) end
| (83,(_,(MlyValue.strdescs' strdescs'1,strdescs'1left,strdescs'1right
))::rest671) => let val result=MlyValue.strdescs''(fn _ => let val 
strdescs' as strdescs'1=strdescs'1 ()
 in (strdescs') end
)
 in (LrTable.NT 94,(result,strdescs'1left,strdescs'1right),rest671)
 end
| (84,(_,(MlyValue.strdescs'' strdescs''1,_,strdescs''1right))::(_,(
MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671) => 
let val result=MlyValue.strdescs''(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val strdescs'' as strdescs''1=strdescs''1 ()
 in (cons1(wherespec, strdescs'')) end
)
 in (LrTable.NT 94,(result,AND1left,strdescs''1right),rest671) end
| (85,(_,(MlyValue.typdesc typdesc1,typdesc1left,typdesc1right))::
rest671) => let val result=MlyValue.typdescs(fn _ => let val typdesc
 as typdesc1=typdesc1 ()
 in ([typdesc]) end
)
 in (LrTable.NT 95,(result,typdesc1left,typdesc1right),rest671) end
| (86,(_,(MlyValue.typdescs typdescs1,_,typdescs1right))::_::(_,(
MlyValue.typdesc typdesc1,typdesc1left,_))::rest671) => let val result
=MlyValue.typdescs(fn _ => let val typdesc as typdesc1=typdesc1 ()
val typdescs as typdescs1=typdescs1 ()
 in (typdesc :: typdescs) end
)
 in (LrTable.NT 95,(result,typdesc1left,typdescs1right),rest671) end
| (87,(_,(MlyValue.typdescRhs typdescRhs1,_,typdescRhs1right))::(_,(
MlyValue.tycon tycon1,_,_))::(_,(MlyValue.tyvars tyvars1,tyvars1left,_
))::rest671) => let val result=MlyValue.typdesc(fn _ => let val tyvars
 as tyvars1=tyvars1 ()
val tycon as tycon1=tycon1 ()
val typdescRhs as typdescRhs1=typdescRhs1 ()
 in (
{tyvars = tyvars,
				     tycon = tycon,
				     ty = typdescRhs})
 end
)
 in (LrTable.NT 96,(result,tyvars1left,typdescRhs1right),rest671) end
| (88,rest671) => let val result=MlyValue.typdescRhs(fn _ => (NONE))
 in (LrTable.NT 97,(result,defaultPos,defaultPos),rest671) end
| (89,(_,(MlyValue.ty ty1,_,ty1right))::(_,(_,EQUALOP1left,_))::
rest671) => let val result=MlyValue.typdescRhs(fn _ => let val ty as 
ty1=ty1 ()
 in (SOME ty) end
)
 in (LrTable.NT 97,(result,EQUALOP1left,ty1right),rest671) end
| (90,(_,(MlyValue.valdesc valdesc1,valdesc1left,valdesc1right))::
rest671) => let val result=MlyValue.valdescs(fn _ => let val valdesc
 as valdesc1=valdesc1 ()
 in ([valdesc]) end
)
 in (LrTable.NT 98,(result,valdesc1left,valdesc1right),rest671) end
| (91,(_,(MlyValue.valdescs valdescs1,_,valdescs1right))::_::(_,(
MlyValue.valdesc valdesc1,valdesc1left,_))::rest671) => let val result
=MlyValue.valdescs(fn _ => let val valdesc as valdesc1=valdesc1 ()
val valdescs as valdescs1=valdescs1 ()
 in (valdesc :: valdescs) end
)
 in (LrTable.NT 98,(result,valdesc1left,valdescs1right),rest671) end
| (92,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.var var1,
var1left,_))::rest671) => let val result=MlyValue.valdesc(fn _ => let 
val var as var1=var1 ()
val ty as ty1=ty1 ()
 in ((var,ty)) end
)
 in (LrTable.NT 99,(result,var1left,ty1right),rest671) end
| (93,(_,(MlyValue.exndesc exndesc1,exndesc1left,exndesc1right))::
rest671) => let val result=MlyValue.exndescs(fn _ => let val exndesc
 as exndesc1=exndesc1 ()
 in ([exndesc]) end
)
 in (LrTable.NT 100,(result,exndesc1left,exndesc1right),rest671) end
| (94,(_,(MlyValue.exndescs exndescs1,_,exndescs1right))::_::(_,(
MlyValue.exndesc exndesc1,exndesc1left,_))::rest671) => let val result
=MlyValue.exndescs(fn _ => let val exndesc as exndesc1=exndesc1 ()
val exndescs as exndescs1=exndescs1 ()
 in (exndesc :: exndescs) end
)
 in (LrTable.NT 100,(result,exndesc1left,exndescs1right),rest671) end
| (95,(_,(MlyValue.tyOpt tyOpt1,_,tyOpt1right))::(_,(MlyValue.con con1
,con1left,_))::rest671) => let val result=MlyValue.exndesc(fn _ => 
let val con as con1=con1 ()
val tyOpt as tyOpt1=tyOpt1 ()
 in ((con, tyOpt)) end
)
 in (LrTable.NT 101,(result,con1left,tyOpt1right),rest671) end
| (96,rest671) => let val result=MlyValue.tyOpt(fn _ => (NONE))
 in (LrTable.NT 38,(result,defaultPos,defaultPos),rest671) end
| (97,(_,(MlyValue.ty ty1,_,ty1right))::(_,(_,OF1left,_))::rest671)
 => let val result=MlyValue.tyOpt(fn _ => let val ty as ty1=ty1 ()
 in (SOME ty) end
)
 in (LrTable.NT 38,(result,OF1left,ty1right),rest671) end
| (98,(_,(MlyValue.funbinds' funbinds'1,_,funbinds'1right))::_::(_,(
MlyValue.sigconst sigconst1,_,_))::_::(_,(MlyValue.fctarg fctarg1,
fctargleft,fctargright))::_::(_,(MlyValue.fctid fctid1,fctid1left,_))
::rest671) => let val result=MlyValue.funbinds(fn _ => let val fctid
 as fctid1=fctid1 ()
val fctarg as fctarg1=fctarg1 ()
val sigconst as sigconst1=sigconst1 ()
val funbinds' as funbinds'1=funbinds'1 ()
 in (
let val (strexp,funbinds) = funbinds'
	    in {name = fctid,
		arg = FctArg.makeRegion'(fctarg, fctargleft, fctargright),
		result = sigconst,
		body = strexp}
	       :: funbinds
	    end
) end
)
 in (LrTable.NT 124,(result,fctid1left,funbinds'1right),rest671) end
| (99,(_,(MlyValue.funbinds'1 funbinds'11,_,funbinds'11right))::(_,(
MlyValue.strexp1 strexp11,strexp11left,_))::rest671) => let val result
=MlyValue.funbinds'(fn _ => let val strexp1 as strexp11=strexp11 ()
val funbinds'1 as funbinds'11=funbinds'11 ()
 in (augment1(strexp1, funbinds'1)) end
)
 in (LrTable.NT 125,(result,strexp11left,funbinds'11right),rest671)
 end
| (100,(_,(MlyValue.funbinds'2 funbinds'21,_,funbinds'21right))::(_,(
MlyValue.strexp2 strexp21,strexp21left,_))::rest671) => let val result
=MlyValue.funbinds'(fn _ => let val strexp2 as strexp21=strexp21 ()
val funbinds'2 as funbinds'21=funbinds'21 ()
 in ((strexp2, funbinds'2)) end
)
 in (LrTable.NT 125,(result,strexp21left,funbinds'21right),rest671)
 end
| (101,(_,(MlyValue.funbinds'2 funbinds'21,funbinds'2left as 
funbinds'21left,funbinds'21right))::rest671) => let val result=
MlyValue.funbinds'1(fn _ => let val funbinds'2 as funbinds'21=
funbinds'21 ()
 in ([], funbinds'2left, funbinds'2) end
)
 in (LrTable.NT 126,(result,funbinds'21left,funbinds'21right),rest671)
 end
| (102,(_,(MlyValue.funbinds'1' funbinds'1'1,_,funbinds'1'1right))::(_
,(MlyValue.wherespec wherespec1,_,_))::(_,(_,WHERE1left,_))::rest671)
 => let val result=MlyValue.funbinds'1(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val funbinds'1' as funbinds'1'1=funbinds'1'1 ()
 in (cons1(wherespec,funbinds'1')) end
)
 in (LrTable.NT 126,(result,WHERE1left,funbinds'1'1right),rest671) end
| (103,rest671) => let val result=MlyValue.funbinds'2(fn _ => ([]))
 in (LrTable.NT 128,(result,defaultPos,defaultPos),rest671) end
| (104,(_,(MlyValue.funbinds funbinds1,_,funbinds1right))::(_,(_,
AND1left,_))::rest671) => let val result=MlyValue.funbinds'2(fn _ => 
let val funbinds as funbinds1=funbinds1 ()
 in (funbinds) end
)
 in (LrTable.NT 128,(result,AND1left,funbinds1right),rest671) end
| (105,(_,(MlyValue.funbinds'1 funbinds'11,funbinds'11left,
funbinds'11right))::rest671) => let val result=MlyValue.funbinds'1'(
fn _ => let val funbinds'1 as funbinds'11=funbinds'11 ()
 in (funbinds'1) end
)
 in (LrTable.NT 127,(result,funbinds'11left,funbinds'11right),rest671)
 end
| (106,(_,(MlyValue.funbinds'1' funbinds'1'1,_,funbinds'1'1right))::(_
,(MlyValue.wherespec wherespec1,_,_))::(_,(_,AND1left,_))::rest671)
 => let val result=MlyValue.funbinds'1'(fn _ => let val wherespec as 
wherespec1=wherespec1 ()
val funbinds'1' as funbinds'1'1=funbinds'1'1 ()
 in (cons1(wherespec,funbinds'1')) end
)
 in (LrTable.NT 127,(result,AND1left,funbinds'1'1right),rest671) end
| (107,(_,(MlyValue.sigexp sigexp1,_,sigexp1right))::_::(_,(
MlyValue.strid strid1,strid1left,_))::rest671) => let val result=
MlyValue.fctarg(fn _ => let val strid as strid1=strid1 ()
val sigexp as sigexp1=sigexp1 ()
 in (FctArg.Structure(strid, sigexp)) end
)
 in (LrTable.NT 129,(result,strid1left,sigexp1right),rest671) end
| (108,(_,(MlyValue.specs specs1,specs1left,specs1right))::rest671)
 => let val result=MlyValue.fctarg(fn _ => let val specs as specs1=
specs1 ()
 in (FctArg.Spec specs) end
)
 in (LrTable.NT 129,(result,specs1left,specs1right),rest671) end
| (109,rest671) => let val result=MlyValue.decs(fn _ => (
Dec.makeRegion'(Dec.SeqDec nil,
					 defaultPos, defaultPos)))
 in (LrTable.NT 86,(result,defaultPos,defaultPos),rest671) end
| (110,(_,(MlyValue.decs decs1,_,decs1right))::(_,(MlyValue.dec dec1,
dec1left,_))::rest671) => let val result=MlyValue.decs(fn _ => let 
val dec as dec1=dec1 ()
val decs as decs1=decs1 ()
 in (Dec.sequence(dec,decs)) end
)
 in (LrTable.NT 86,(result,dec1left,decs1right),rest671) end
| (111,(_,(MlyValue.decs decs1,_,decs1right))::(_,(_,SEMICOLON1left,_)
)::rest671) => let val result=MlyValue.decs(fn _ => let val decs as 
decs1=decs1 ()
 in (decs) end
)
 in (LrTable.NT 86,(result,SEMICOLON1left,decs1right),rest671) end
| (112,(_,(MlyValue.decnode decnode1,decnodeleft as decnode1left,
decnoderight as decnode1right))::rest671) => let val result=
MlyValue.dec(fn _ => let val decnode as decnode1=decnode1 ()
 in (Dec.makeRegion'(decnode, decnodeleft, decnoderight)) end
)
 in (LrTable.NT 83,(result,decnode1left,decnode1right),rest671) end
| (113,(_,(MlyValue.decnolocal decnolocal1,decnolocal1left,
decnolocal1right))::rest671) => let val result=MlyValue.decnode(fn _
 => let val decnolocal as decnolocal1=decnolocal1 ()
 in (decnolocal) end
)
 in (LrTable.NT 84,(result,decnolocal1left,decnolocal1right),rest671)
 end
| (114,(_,(_,_,END1right))::(_,(MlyValue.decs decs2,_,_))::_::(_,(
MlyValue.decs decs1,_,_))::(_,(_,LOCAL1left,_))::rest671) => let val 
result=MlyValue.decnode(fn _ => let val decs1=decs1 ()
val decs2=decs2 ()
 in (Dec.Local(decs1,decs2)) end
)
 in (LrTable.NT 84,(result,LOCAL1left,END1right),rest671) end
| (115,(_,(MlyValue.vbs vbs1,_,vbs1right))::(_,(_,VAL1left,_))::
rest671) => let val result=MlyValue.decnolocal(fn _ => let val vbs as 
vbs1=vbs1 ()
 in (Dec.Val(nil,vbs)) end
)
 in (LrTable.NT 85,(result,VAL1left,vbs1right),rest671) end
| (116,(_,(MlyValue.vbs vbs1,_,vbs1right))::(_,(MlyValue.tyvarseq 
tyvarseq1,_,_))::(_,(_,VAL1left,_))::rest671) => let val result=
MlyValue.decnolocal(fn _ => let val tyvarseq as tyvarseq1=tyvarseq1 ()
val vbs as vbs1=vbs1 ()
 in (Dec.Val(tyvarseq,vbs)) end
)
 in (LrTable.NT 85,(result,VAL1left,vbs1right),rest671) end
| (117,(_,(MlyValue.rvbs rvbs1,_,rvbs1right))::_::(_,(_,VAL1left,_))::
rest671) => let val result=MlyValue.decnolocal(fn _ => let val rvbs
 as rvbs1=rvbs1 ()
 in (Dec.Valrec(nil,rvbs)) end
)
 in (LrTable.NT 85,(result,VAL1left,rvbs1right),rest671) end
| (118,(_,(MlyValue.rvbs rvbs1,_,rvbs1right))::(_,(MlyValue.tyvarseq 
tyvarseq1,_,_))::_::(_,(_,VAL1left,_))::rest671) => let val result=
MlyValue.decnolocal(fn _ => let val tyvarseq as tyvarseq1=tyvarseq1 ()
val rvbs as rvbs1=rvbs1 ()
 in (Dec.Valrec(tyvarseq,rvbs)) end
)
 in (LrTable.NT 85,(result,VAL1left,rvbs1right),rest671) end
| (119,(_,(MlyValue.funs funs1,_,funs1right))::(_,(_,FUN1left,_))::
rest671) => let val result=MlyValue.decnolocal(fn _ => let val funs
 as funs1=funs1 ()
 in (Dec.Fun(nil,funs)) end
)
 in (LrTable.NT 85,(result,FUN1left,funs1right),rest671) end
| (120,(_,(MlyValue.funs funs1,_,funs1right))::(_,(MlyValue.tyvarseq 
tyvarseq1,_,_))::(_,(_,FUN1left,_))::rest671) => let val result=
MlyValue.decnolocal(fn _ => let val tyvarseq as tyvarseq1=tyvarseq1 ()
val funs as funs1=funs1 ()
 in (Dec.Fun(tyvarseq,funs)) end
)
 in (LrTable.NT 85,(result,FUN1left,funs1right),rest671) end
| (121,(_,(MlyValue.typBind typBind1,_,typBind1right))::(_,(_,
TYPE1left,_))::rest671) => let val result=MlyValue.decnolocal(fn _ => 
let val typBind as typBind1=typBind1 ()
 in (Dec.Type typBind) end
)
 in (LrTable.NT 85,(result,TYPE1left,typBind1right),rest671) end
| (122,(_,(MlyValue.datatypeRhs datatypeRhs1,_,datatypeRhs1right))::(_
,(_,DATATYPE1left,_))::rest671) => let val result=MlyValue.decnolocal(
fn _ => let val datatypeRhs as datatypeRhs1=datatypeRhs1 ()
 in (Dec.Datatype datatypeRhs) end
)
 in (LrTable.NT 85,(result,DATATYPE1left,datatypeRhs1right),rest671)
 end
| (123,(_,(_,_,END1right))::(_,(MlyValue.decs decs1,_,_))::_::(_,(
MlyValue.datBind datBind1,_,_))::(_,(_,ABSTYPE1left,_))::rest671) => 
let val result=MlyValue.decnolocal(fn _ => let val datBind as datBind1
=datBind1 ()
val decs as decs1=decs1 ()
 in (Dec.Abstype{datBind = datBind,
						       body = decs}) end
)
 in (LrTable.NT 85,(result,ABSTYPE1left,END1right),rest671) end
| (124,(_,(MlyValue.ebs ebs1,_,ebs1right))::(_,(_,EXCEPTION1left,_))::
rest671) => let val result=MlyValue.decnolocal(fn _ => let val ebs as 
ebs1=ebs1 ()
 in (Dec.Exception ebs) end
)
 in (LrTable.NT 85,(result,EXCEPTION1left,ebs1right),rest671) end
| (125,(_,(MlyValue.longstrids longstrids1,_,longstrids1right))::(_,(_
,OPEN1left,_))::rest671) => let val result=MlyValue.decnolocal(fn _
 => let val longstrids as longstrids1=longstrids1 ()
 in (Dec.Open longstrids) end
)
 in (LrTable.NT 85,(result,OPEN1left,longstrids1right),rest671) end
| (126,(_,(MlyValue.vids vids1,_,vids1right))::(_,(MlyValue.fixity 
fixity1,fixity1left,_))::rest671) => let val result=
MlyValue.decnolocal(fn _ => let val fixity as fixity1=fixity1 ()
val vids as vids1=vids1 ()
 in (Dec.Fix{fixity = fixity, ops = vids}) end
)
 in (LrTable.NT 85,(result,fixity1left,vids1right),rest671) end
| (127,(_,(MlyValue.longvarands longvarands1,_,longvarands1right))::_
::(_,(MlyValue.ty ty1,_,_))::_::(_,(MlyValue.var var1,_,_))::(_,(_,
OVERLOAD1left,_))::rest671) => let val result=MlyValue.decnolocal(fn _
 => let val var as var1=var1 ()
val ty as ty1=ty1 ()
val longvarands as longvarands1=longvarands1 ()
 in (Dec.Overload(var, ty, longvarands)) end
)
 in (LrTable.NT 85,(result,OVERLOAD1left,longvarands1right),rest671)
 end
| (128,(_,(MlyValue.vbs vbs2,_,vbs2right))::_::(_,(MlyValue.vbs vbs1,
vbs1left,_))::rest671) => let val result=MlyValue.vbs(fn _ => let val 
vbs1=vbs1 ()
val vbs2=vbs2 ()
 in (vbs1 @ vbs2) end
)
 in (LrTable.NT 65,(result,vbs1left,vbs2right),rest671) end
| (129,(_,(MlyValue.exp exp1,_,exp1right))::_::(_,(MlyValue.pat pat1,
pat1left,_))::rest671) => let val result=MlyValue.vbs(fn _ => let val 
pat as pat1=pat1 ()
val exp as exp1=exp1 ()
 in ([(pat,exp)]) end
)
 in (LrTable.NT 65,(result,pat1left,exp1right),rest671) end
| (130,rest671) => let val result=MlyValue.constraint(fn _ => (NONE))
 in (LrTable.NT 66,(result,defaultPos,defaultPos),rest671) end
| (131,(_,(MlyValue.ty ty1,_,ty1right))::(_,(_,COLON1left,_))::rest671
) => let val result=MlyValue.constraint(fn _ => let val ty as ty1=ty1 
()
 in (SOME ty) end
)
 in (LrTable.NT 66,(result,COLON1left,ty1right),rest671) end
| (132,(_,(MlyValue.rules rules1,_,rules1right))::_::_::(_,(
MlyValue.constraint constraint1,_,_))::(_,(MlyValue.vid vid1,vid1left,
_))::rest671) => let val result=MlyValue.rvbs(fn _ => let val vid as 
vid1=vid1 ()
val constraint as constraint1=constraint1 ()
val rules as rules1=rules1 ()
 in (
[{var = Vid.toVar vid,
	     fixity = SOME vid,
	     ty = constraint,
	     rules = rules}]
) end
)
 in (LrTable.NT 67,(result,vid1left,rules1right),rest671) end
| (133,(_,(MlyValue.rules rules1,_,rules1right))::_::_::(_,(
MlyValue.constraint constraint1,_,_))::(_,(MlyValue.var var1,_,_))::(_
,(_,OP1left,_))::rest671) => let val result=MlyValue.rvbs(fn _ => let 
val var as var1=var1 ()
val constraint as constraint1=constraint1 ()
val rules as rules1=rules1 ()
 in (
[{var = var,
						 fixity = NONE,
						 ty = constraint,
						 rules = rules}]
) end
)
 in (LrTable.NT 67,(result,OP1left,rules1right),rest671) end
| (134,(_,(MlyValue.rvbs rvbs2,_,rvbs2right))::_::(_,(MlyValue.rvbs 
rvbs1,rvbs1left,_))::rest671) => let val result=MlyValue.rvbs(fn _ => 
let val rvbs1=rvbs1 ()
val rvbs2=rvbs2 ()
 in (rvbs1 @ rvbs2) end
)
 in (LrTable.NT 67,(result,rvbs1left,rvbs2right),rest671) end
| (135,(_,(MlyValue.clauses clauses1,clauses1left,clauses1right))::
rest671) => let val result=MlyValue.funs(fn _ => let val clauses as 
clauses1=clauses1 ()
 in ([clauses]) end
)
 in (LrTable.NT 68,(result,clauses1left,clauses1right),rest671) end
| (136,(_,(MlyValue.funs funs1,_,funs1right))::_::(_,(MlyValue.clauses
 clauses1,clauses1left,_))::rest671) => let val result=MlyValue.funs(
fn _ => let val clauses as clauses1=clauses1 ()
val funs as funs1=funs1 ()
 in (clauses :: funs) end
)
 in (LrTable.NT 68,(result,clauses1left,funs1right),rest671) end
| (137,(_,(MlyValue.clause clause1,clause1left,clause1right))::rest671
) => let val result=MlyValue.clauses(fn _ => let val clause as clause1
=clause1 ()
 in ([clause]) end
)
 in (LrTable.NT 69,(result,clause1left,clause1right),rest671) end
| (138,(_,(MlyValue.clauses clauses1,_,clauses1right))::_::(_,(
MlyValue.clause clause1,clause1left,_))::rest671) => let val result=
MlyValue.clauses(fn _ => let val clause as clause1=clause1 ()
val clauses as clauses1=clauses1 ()
 in (clause :: clauses) end
)
 in (LrTable.NT 69,(result,clause1left,clauses1right),rest671) end
| (139,(_,(MlyValue.exp exp1,_,exp1right))::_::(_,(MlyValue.constraint
 constraint1,_,_))::(_,(MlyValue.apats apats1,apats1left,_))::rest671)
 => let val result=MlyValue.clause(fn _ => let val apats as apats1=
apats1 ()
val constraint as constraint1=constraint1 ()
val exp as exp1=exp1 ()
 in ({pats = apats,
					  resultType = constraint,
					  body = exp}
) end
)
 in (LrTable.NT 70,(result,apats1left,exp1right),rest671) end
| (140,(_,(MlyValue.typBind' typBind'1,_,typBind'right as 
typBind'1right))::(_,(MlyValue.ty ty1,_,_))::_::(_,(MlyValue.tycon 
tycon1,_,_))::(_,(MlyValue.tyvars tyvars1,tyvarsleft as tyvars1left,_)
)::rest671) => let val result=MlyValue.typBind(fn _ => let val tyvars
 as tyvars1=tyvars1 ()
val tycon as tycon1=tycon1 ()
val ty as ty1=ty1 ()
val typBind' as typBind'1=typBind'1 ()
 in (
let val TypBind.T rest = typBind'
			 in TypBind.makeRegion'
			    (TypBind.T({tyvars = tyvars,
					tycon = tycon,
					def = ty} :: rest),
			     tyvarsleft, typBind'right)
			 end
) end
)
 in (LrTable.NT 71,(result,tyvars1left,typBind'1right),rest671) end
| (141,rest671) => let val result=MlyValue.typBind'(fn _ => (
TypBind.T[]))
 in (LrTable.NT 72,(result,defaultPos,defaultPos),rest671) end
| (142,(_,(MlyValue.typBind typBind1,_,typBind1right))::(_,(_,AND1left
,_))::rest671) => let val result=MlyValue.typBind'(fn _ => let val 
typBind as typBind1=typBind1 ()
 in (TypBind.node typBind) end
)
 in (LrTable.NT 72,(result,AND1left,typBind1right),rest671) end
| (143,(_,(MlyValue.tyvar tyvar1,tyvar1left,tyvar1right))::rest671)
 => let val result=MlyValue.tyvars(fn _ => let val tyvar as tyvar1=
tyvar1 ()
 in ([tyvar]) end
)
 in (LrTable.NT 73,(result,tyvar1left,tyvar1right),rest671) end
| (144,(_,(_,_,RPAREN1right))::(_,(MlyValue.tyvar_pc tyvar_pc1,_,_))::
(_,(_,LPAREN1left,_))::rest671) => let val result=MlyValue.tyvars(fn _
 => let val tyvar_pc as tyvar_pc1=tyvar_pc1 ()
 in (tyvar_pc) end
)
 in (LrTable.NT 73,(result,LPAREN1left,RPAREN1right),rest671) end
| (145,rest671) => let val result=MlyValue.tyvars(fn _ => (nil))
 in (LrTable.NT 73,(result,defaultPos,defaultPos),rest671) end
| (146,(_,(MlyValue.tyvar tyvar1,tyvar1left,tyvar1right))::rest671)
 => let val result=MlyValue.tyvarseq(fn _ => let val tyvar as tyvar1=
tyvar1 ()
 in ([tyvar]) end
)
 in (LrTable.NT 74,(result,tyvar1left,tyvar1right),rest671) end
| (147,(_,(_,_,RPAREN1right))::(_,(MlyValue.tyvar_pc tyvar_pc1,_,_))::
(_,(_,LPAREN1left,_))::rest671) => let val result=MlyValue.tyvarseq(
fn _ => let val tyvar_pc as tyvar_pc1=tyvar_pc1 ()
 in (tyvar_pc) end
)
 in (LrTable.NT 74,(result,LPAREN1left,RPAREN1right),rest671) end
| (148,(_,(MlyValue.tyvar tyvar1,tyvar1left,tyvar1right))::rest671)
 => let val result=MlyValue.tyvar_pc(fn _ => let val tyvar as tyvar1=
tyvar1 ()
 in ([tyvar]) end
)
 in (LrTable.NT 75,(result,tyvar1left,tyvar1right),rest671) end
| (149,(_,(MlyValue.tyvar_pc tyvar_pc1,_,tyvar_pc1right))::_::(_,(
MlyValue.tyvar tyvar1,tyvar1left,_))::rest671) => let val result=
MlyValue.tyvar_pc(fn _ => let val tyvar as tyvar1=tyvar1 ()
val tyvar_pc as tyvar_pc1=tyvar_pc1 ()
 in (tyvar :: tyvar_pc) end
)
 in (LrTable.NT 75,(result,tyvar1left,tyvar_pc1right),rest671) end
| (150,(_,(MlyValue.constr constr1,constr1left,constr1right))::rest671
) => let val result=MlyValue.constrs(fn _ => let val constr as constr1
=constr1 ()
 in ([constr]) end
)
 in (LrTable.NT 76,(result,constr1left,constr1right),rest671) end
| (151,(_,(MlyValue.constrs constrs1,_,constrs1right))::_::(_,(
MlyValue.constr constr1,constr1left,_))::rest671) => let val result=
MlyValue.constrs(fn _ => let val constr as constr1=constr1 ()
val constrs as constrs1=constrs1 ()
 in (constr :: constrs) end
)
 in (LrTable.NT 76,(result,constr1left,constrs1right),rest671) end
| (152,(_,(MlyValue.con con1,con1left,con1right))::rest671) => let 
val result=MlyValue.constr(fn _ => let val con as con1=con1 ()
 in (con, NONE) end
)
 in (LrTable.NT 77,(result,con1left,con1right),rest671) end
| (153,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.con con1,
con1left,_))::rest671) => let val result=MlyValue.constr(fn _ => let 
val con as con1=con1 ()
val ty as ty1=ty1 ()
 in (con, SOME ty) end
)
 in (LrTable.NT 77,(result,con1left,ty1right),rest671) end
| (154,(_,(MlyValue.eb eb1,eb1left,eb1right))::rest671) => let val 
result=MlyValue.ebs(fn _ => let val eb as eb1=eb1 ()
 in ([eb]) end
)
 in (LrTable.NT 78,(result,eb1left,eb1right),rest671) end
| (155,(_,(MlyValue.ebs ebs1,_,ebs1right))::_::(_,(MlyValue.eb eb1,
eb1left,_))::rest671) => let val result=MlyValue.ebs(fn _ => let val 
eb as eb1=eb1 ()
val ebs as ebs1=ebs1 ()
 in (eb::ebs) end
)
 in (LrTable.NT 78,(result,eb1left,ebs1right),rest671) end
| (156,(_,(MlyValue.ebrhs ebrhs1,_,ebrhs1right))::(_,(MlyValue.con 
con1,con1left,_))::rest671) => let val result=MlyValue.eb(fn _ => let 
val con as con1=con1 ()
val ebrhs as ebrhs1=ebrhs1 ()
 in ((con,ebrhs)) end
)
 in (LrTable.NT 79,(result,con1left,ebrhs1right),rest671) end
| (157,(_,(MlyValue.ebrhsnode ebrhsnode1,ebrhsnodeleft as 
ebrhsnode1left,ebrhsnoderight as ebrhsnode1right))::rest671) => let 
val result=MlyValue.ebrhs(fn _ => let val ebrhsnode as ebrhsnode1=
ebrhsnode1 ()
 in (
EbRhs.makeRegion'(ebrhsnode,
				     ebrhsnodeleft, ebrhsnoderight))
 end
)
 in (LrTable.NT 80,(result,ebrhsnode1left,ebrhsnode1right),rest671)
 end
| (158,rest671) => let val result=MlyValue.ebrhsnode(fn _ => (
EbRhs.Gen NONE))
 in (LrTable.NT 81,(result,defaultPos,defaultPos),rest671) end
| (159,(_,(MlyValue.ty ty1,_,ty1right))::(_,(_,OF1left,_))::rest671)
 => let val result=MlyValue.ebrhsnode(fn _ => let val ty as ty1=ty1 ()
 in (EbRhs.Gen(SOME ty)) end
)
 in (LrTable.NT 81,(result,OF1left,ty1right),rest671) end
| (160,(_,(MlyValue.longcon longcon1,_,longcon1right))::(_,(_,
EQUALOP1left,_))::rest671) => let val result=MlyValue.ebrhsnode(fn _
 => let val longcon as longcon1=longcon1 ()
 in (EbRhs.Def longcon) end
)
 in (LrTable.NT 81,(result,EQUALOP1left,longcon1right),rest671) end
| (161,(_,(_,INFIX1left,INFIX1right))::rest671) => let val result=
MlyValue.fixity(fn _ => (Fixity.Infix NONE))
 in (LrTable.NT 82,(result,INFIX1left,INFIX1right),rest671) end
| (162,(_,(MlyValue.digit digit1,_,digit1right))::(_,(_,INFIX1left,_))
::rest671) => let val result=MlyValue.fixity(fn _ => let val digit as 
digit1=digit1 ()
 in (Fixity.Infix(SOME digit)) end
)
 in (LrTable.NT 82,(result,INFIX1left,digit1right),rest671) end
| (163,(_,(_,INFIXR1left,INFIXR1right))::rest671) => let val result=
MlyValue.fixity(fn _ => (Fixity.Infixr NONE))
 in (LrTable.NT 82,(result,INFIXR1left,INFIXR1right),rest671) end
| (164,(_,(MlyValue.digit digit1,_,digit1right))::(_,(_,INFIXR1left,_)
)::rest671) => let val result=MlyValue.fixity(fn _ => let val digit
 as digit1=digit1 ()
 in (Fixity.Infixr(SOME digit)) end
)
 in (LrTable.NT 82,(result,INFIXR1left,digit1right),rest671) end
| (165,(_,(_,NONFIX1left,NONFIX1right))::rest671) => let val result=
MlyValue.fixity(fn _ => (Fixity.Nonfix))
 in (LrTable.NT 82,(result,NONFIX1left,NONFIX1right),rest671) end
| (166,(_,(MlyValue.INT INT1,INTleft as INT1left,INTright as INT1right
))::rest671) => let val result=MlyValue.int(fn _ => let val INT as 
INT1=INT1 ()
 in (
case Int.fromString INT of
	      NONE => (error(reg(INTleft, INTright), "expected integer");
		       ~1)
	    | SOME n => n
) end
)
 in (LrTable.NT 1,(result,INT1left,INT1right),rest671) end
| (167,(_,(MlyValue.int int1,intleft as int1left,intright as int1right
))::rest671) => let val result=MlyValue.digit(fn _ => let val int as 
int1=int1 ()
 in (
if int < 0 orelse int >= 10
		then (error(reg(intleft, intright), "expected single digit") ;
		      0)
	     else int
) end
)
 in (LrTable.NT 0,(result,int1left,int1right),rest671) end
| (168,(_,(MlyValue.datatypeRhsnode datatypeRhsnode1,
datatypeRhsnodeleft as datatypeRhsnode1left,datatypeRhsnoderight as 
datatypeRhsnode1right))::rest671) => let val result=
MlyValue.datatypeRhs(fn _ => let val datatypeRhsnode as 
datatypeRhsnode1=datatypeRhsnode1 ()
 in (
DatatypeRhs.makeRegion'(datatypeRhsnode,
						       datatypeRhsnodeleft,
						       datatypeRhsnoderight)
) end
)
 in (LrTable.NT 130,(result,datatypeRhsnode1left,datatypeRhsnode1right
),rest671) end
| (169,(_,(MlyValue.datatypeRhsnodeNoWithtype 
datatypeRhsnodeNoWithtype1,datatypeRhsnodeNoWithtypeleft as 
datatypeRhsnodeNoWithtype1left,datatypeRhsnodeNoWithtyperight as 
datatypeRhsnodeNoWithtype1right))::rest671) => let val result=
MlyValue.datatypeRhsNoWithtype(fn _ => let val 
datatypeRhsnodeNoWithtype as datatypeRhsnodeNoWithtype1=
datatypeRhsnodeNoWithtype1 ()
 in (
DatatypeRhs.makeRegion'(datatypeRhsnodeNoWithtype,
						 datatypeRhsnodeNoWithtypeleft,
						 datatypeRhsnodeNoWithtyperight)
) end
)
 in (LrTable.NT 131,(result,datatypeRhsnodeNoWithtype1left,
datatypeRhsnodeNoWithtype1right),rest671) end
| (170,(_,(MlyValue.repl repl1,repl1left,repl1right))::rest671) => 
let val result=MlyValue.datatypeRhsnode(fn _ => let val repl as repl1=
repl1 ()
 in (repl) end
)
 in (LrTable.NT 132,(result,repl1left,repl1right),rest671) end
| (171,(_,(MlyValue.datBind datBind1,datBind1left,datBind1right))::
rest671) => let val result=MlyValue.datatypeRhsnode(fn _ => let val 
datBind as datBind1=datBind1 ()
 in (DatatypeRhs.DatBind datBind) end
)
 in (LrTable.NT 132,(result,datBind1left,datBind1right),rest671) end
| (172,(_,(MlyValue.repl repl1,repl1left,repl1right))::rest671) => 
let val result=MlyValue.datatypeRhsnodeNoWithtype(fn _ => let val repl
 as repl1=repl1 ()
 in (repl) end
)
 in (LrTable.NT 133,(result,repl1left,repl1right),rest671) end
| (173,(_,(MlyValue.datBindNoWithtype datBindNoWithtype1,
datBindNoWithtype1left,datBindNoWithtype1right))::rest671) => let val 
result=MlyValue.datatypeRhsnodeNoWithtype(fn _ => let val 
datBindNoWithtype as datBindNoWithtype1=datBindNoWithtype1 ()
 in (DatatypeRhs.DatBind datBindNoWithtype) end
)
 in (LrTable.NT 133,(result,datBindNoWithtype1left,
datBindNoWithtype1right),rest671) end
| (174,(_,(MlyValue.longtycon longtycon1,_,longtycon1right))::_::_::(_
,(MlyValue.tycon tycon1,_,_))::(_,(MlyValue.tyvars tyvars1,tyvarsleft
 as tyvars1left,tyvarsright))::rest671) => let val result=
MlyValue.repl(fn _ => let val tyvars as tyvars1=tyvars1 ()
val tycon as tycon1=tycon1 ()
val longtycon as longtycon1=longtycon1 ()
 in (
case tyvars of
	   _ :: _ => error(reg(tyvarsleft, tyvarsright),
			   "nonempty tyvars in datatype repl")
	 | _ => () ;
        DatatypeRhs.Repl{lhs = tycon, rhs = longtycon}
) end
)
 in (LrTable.NT 134,(result,tyvars1left,longtycon1right),rest671) end
| (175,(_,(MlyValue.withtypes withtypes1,_,withtypesright as 
withtypes1right))::(_,(MlyValue.dbs dbs1,dbsleft as dbs1left,_))::
rest671) => let val result=MlyValue.datBind(fn _ => let val dbs as 
dbs1=dbs1 ()
val withtypes as withtypes1=withtypes1 ()
 in (
DatBind.makeRegion'(DatBind.T{datatypes = dbs,
						       withtypes = withtypes},
					     dbsleft, withtypesright)
) end
)
 in (LrTable.NT 135,(result,dbs1left,withtypes1right),rest671) end
| (176,(_,(MlyValue.dbs dbs1,dbsleft as dbs1left,dbsright as dbs1right
))::rest671) => let val result=MlyValue.datBindNoWithtype(fn _ => let 
val dbs as dbs1=dbs1 ()
 in (
DatBind.makeRegion'(DatBind.T{datatypes = dbs,
					  withtypes = TypBind.empty},
				dbsleft, dbsright)
) end
)
 in (LrTable.NT 136,(result,dbs1left,dbs1right),rest671) end
| (177,(_,(MlyValue.db db1,db1left,db1right))::rest671) => let val 
result=MlyValue.dbs(fn _ => let val db as db1=db1 ()
 in ([db]) end
)
 in (LrTable.NT 138,(result,db1left,db1right),rest671) end
| (178,(_,(MlyValue.dbs dbs1,_,dbs1right))::_::(_,(MlyValue.db db1,
db1left,_))::rest671) => let val result=MlyValue.dbs(fn _ => let val 
db as db1=db1 ()
val dbs as dbs1=dbs1 ()
 in (db :: dbs) end
)
 in (LrTable.NT 138,(result,db1left,dbs1right),rest671) end
| (179,(_,(MlyValue.constrs constrs1,_,constrs1right))::_::(_,(
MlyValue.tycon tycon1,_,_))::(_,(MlyValue.tyvars tyvars1,tyvars1left,_
))::rest671) => let val result=MlyValue.db(fn _ => let val tyvars as 
tyvars1=tyvars1 ()
val tycon as tycon1=tycon1 ()
val constrs as constrs1=constrs1 ()
 in ({tycon = tycon,
					  tyvars = tyvars,
					  cons = constrs})
 end
)
 in (LrTable.NT 137,(result,tyvars1left,constrs1right),rest671) end
| (180,rest671) => let val result=MlyValue.withtypes(fn _ => (
TypBind.empty))
 in (LrTable.NT 139,(result,defaultPos,defaultPos),rest671) end
| (181,(_,(MlyValue.typBind typBind1,_,typBind1right))::(_,(_,
WITHTYPE1left,_))::rest671) => let val result=MlyValue.withtypes(fn _
 => let val typBind as typBind1=typBind1 ()
 in (typBind) end
)
 in (LrTable.NT 139,(result,WITHTYPE1left,typBind1right),rest671) end
| (182,(_,(MlyValue.longvar longvar1,longvar1left,longvar1right))::
rest671) => let val result=MlyValue.longvarands(fn _ => let val 
longvar as longvar1=longvar1 ()
 in ([longvar]) end
)
 in (LrTable.NT 24,(result,longvar1left,longvar1right),rest671) end
| (183,(_,(MlyValue.longvarands longvarands1,_,longvarands1right))::_
::(_,(MlyValue.longvar longvar1,longvar1left,_))::rest671) => let val 
result=MlyValue.longvarands(fn _ => let val longvar as longvar1=
longvar1 ()
val longvarands as longvarands1=longvarands1 ()
 in (longvar :: longvarands) end
)
 in (LrTable.NT 24,(result,longvar1left,longvarands1right),rest671)
 end
| (184,(_,(MlyValue.rule rule1,rule1left,rule1right))::rest671) => 
let val result=MlyValue.rules(fn _ => let val rule as rule1=rule1 ()
 in ([rule]) end
)
 in (LrTable.NT 40,(result,rule1left,rule1right),rest671) end
| (185,(_,(MlyValue.rules rules1,_,rules1right))::_::(_,(MlyValue.rule
 rule1,rule1left,_))::rest671) => let val result=MlyValue.rules(fn _
 => let val rule as rule1=rule1 ()
val rules as rules1=rules1 ()
 in (rule::rules) end
)
 in (LrTable.NT 40,(result,rule1left,rules1right),rest671) end
| (186,(_,(MlyValue.exp exp1,_,exp1right))::_::(_,(MlyValue.pat pat1,
pat1left,_))::rest671) => let val result=MlyValue.rule(fn _ => let 
val pat as pat1=pat1 ()
val exp as exp1=exp1 ()
 in ((pat,exp)) end
)
 in (LrTable.NT 41,(result,pat1left,exp1right),rest671) end
| (187,(_,(MlyValue.exp exp1,_,exp1right))::_::(_,(MlyValue.field 
field1,field1left,_))::rest671) => let val result=MlyValue.elabel(fn _
 => let val field as field1=field1 ()
val exp as exp1=exp1 ()
 in (field,exp) end
)
 in (LrTable.NT 42,(result,field1left,exp1right),rest671) end
| (188,(_,(MlyValue.elabels elabels1,_,elabels1right))::_::(_,(
MlyValue.elabel elabel1,elabel1left,_))::rest671) => let val result=
MlyValue.elabels(fn _ => let val elabel as elabel1=elabel1 ()
val elabels as elabels1=elabels1 ()
 in (elabel :: elabels) end
)
 in (LrTable.NT 43,(result,elabel1left,elabels1right),rest671) end
| (189,(_,(MlyValue.elabel elabel1,elabel1left,elabel1right))::rest671
) => let val result=MlyValue.elabels(fn _ => let val elabel as elabel1
=elabel1 ()
 in ([elabel]) end
)
 in (LrTable.NT 43,(result,elabel1left,elabel1right),rest671) end
| (190,(_,(MlyValue.exp exp1,exp1left,exp1right))::rest671) => let 
val result=MlyValue.exp_ps(fn _ => let val exp as exp1=exp1 ()
 in ([exp]) end
)
 in (LrTable.NT 44,(result,exp1left,exp1right),rest671) end
| (191,(_,(MlyValue.exp_ps exp_ps1,_,exp_ps1right))::_::(_,(
MlyValue.exp exp1,exp1left,_))::rest671) => let val result=
MlyValue.exp_ps(fn _ => let val exp as exp1=exp1 ()
val exp_ps as exp_ps1=exp_ps1 ()
 in (exp :: exp_ps) end
)
 in (LrTable.NT 44,(result,exp1left,exp_ps1right),rest671) end
| (192,(_,(MlyValue.expnode expnode1,expnodeleft as expnode1left,
expnoderight as expnode1right))::rest671) => let val result=
MlyValue.exp(fn _ => let val expnode as expnode1=expnode1 ()
 in (Exp.makeRegion'(expnode, expnodeleft, expnoderight)) end
)
 in (LrTable.NT 45,(result,expnode1left,expnode1right),rest671) end
| (193,(_,(MlyValue.rules rules1,_,rules1right))::_::(_,(MlyValue.exp 
exp1,exp1left,_))::rest671) => let val result=MlyValue.expnode(fn _
 => let val exp as exp1=exp1 ()
val rules as rules1=rules1 ()
 in (Exp.Handle(exp,rules)) end
)
 in (LrTable.NT 46,(result,exp1left,rules1right),rest671) end
| (194,(_,(MlyValue.exp exp2,_,exp2right))::_::(_,(MlyValue.exp exp1,
exp1left,_))::rest671) => let val result=MlyValue.expnode(fn _ => let 
val exp1=exp1 ()
val exp2=exp2 ()
 in (Exp.Orelse(exp1,exp2)) end
)
 in (LrTable.NT 46,(result,exp1left,exp2right),rest671) end
| (195,(_,(MlyValue.exp exp2,_,exp2right))::_::(_,(MlyValue.exp exp1,
exp1left,_))::rest671) => let val result=MlyValue.expnode(fn _ => let 
val exp1=exp1 ()
val exp2=exp2 ()
 in (Exp.Andalso(exp1,exp2)) end
)
 in (LrTable.NT 46,(result,exp1left,exp2right),rest671) end
| (196,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.exp exp1,
exp1left,_))::rest671) => let val result=MlyValue.expnode(fn _ => let 
val exp as exp1=exp1 ()
val ty as ty1=ty1 ()
 in (Exp.Constraint(exp,ty)) end
)
 in (LrTable.NT 46,(result,exp1left,ty1right),rest671) end
| (197,(_,(MlyValue.app_exp app_exp1,app_exp1left,app_exp1right))::
rest671) => let val result=MlyValue.expnode(fn _ => let val app_exp
 as app_exp1=app_exp1 ()
 in (Exp.FlatApp app_exp) end
)
 in (LrTable.NT 46,(result,app_exp1left,app_exp1right),rest671) end
| (198,(_,(MlyValue.rules rules1,_,rules1right))::(_,(_,FN1left,_))::
rest671) => let val result=MlyValue.expnode(fn _ => let val rules as 
rules1=rules1 ()
 in (Exp.Fn rules) end
)
 in (LrTable.NT 46,(result,FN1left,rules1right),rest671) end
| (199,(_,(MlyValue.rules rules1,_,rules1right))::_::(_,(MlyValue.exp 
exp1,_,_))::(_,(_,CASE1left,_))::rest671) => let val result=
MlyValue.expnode(fn _ => let val exp as exp1=exp1 ()
val rules as rules1=rules1 ()
 in (Exp.Case(exp,rules)) end
)
 in (LrTable.NT 46,(result,CASE1left,rules1right),rest671) end
| (200,(_,(MlyValue.exp exp2,_,exp2right))::_::(_,(MlyValue.exp exp1,_
,_))::(_,(_,WHILE1left,_))::rest671) => let val result=
MlyValue.expnode(fn _ => let val exp1=exp1 ()
val exp2=exp2 ()
 in (Exp.While{test = exp1,expr = exp2}) end
)
 in (LrTable.NT 46,(result,WHILE1left,exp2right),rest671) end
| (201,(_,(MlyValue.exp exp3,_,exp3right))::_::(_,(MlyValue.exp exp2,_
,_))::_::(_,(MlyValue.exp exp1,_,_))::(_,(_,IF1left,_))::rest671) => 
let val result=MlyValue.expnode(fn _ => let val exp1=exp1 ()
val exp2=exp2 ()
val exp3=exp3 ()
 in (Exp.If(exp1,exp2,exp3)) end
)
 in (LrTable.NT 46,(result,IF1left,exp3right),rest671) end
| (202,(_,(MlyValue.exp exp1,_,exp1right))::(_,(_,RAISE1left,_))::
rest671) => let val result=MlyValue.expnode(fn _ => let val exp as 
exp1=exp1 ()
 in (Exp.Raise exp) end
)
 in (LrTable.NT 46,(result,RAISE1left,exp1right),rest671) end
| (203,(_,(MlyValue.app_exp1 app_exp11,_,app_exp11right))::(_,(
MlyValue.aexp aexp1,aexpleft as aexp1left,aexpright))::rest671) => 
let val result=MlyValue.app_exp(fn _ => let val aexp as aexp1=aexp1 ()
val app_exp1 as app_exp11=app_exp11 ()
 in (Exp.makeRegion'(aexp,aexpleft,aexpright)
			     :: app_exp1) end
)
 in (LrTable.NT 47,(result,aexp1left,app_exp11right),rest671) end
| (204,(_,(MlyValue.app_exp1 app_exp11,_,app_exp11right))::(_,(
MlyValue.longvid longvid1,longvidleft as longvid1left,longvidright))::
rest671) => let val result=MlyValue.app_exp(fn _ => let val longvid
 as longvid1=longvid1 ()
val app_exp1 as app_exp11=app_exp11 ()
 in (
Exp.makeRegion'(Exp.Var{name = longvid,
						     fixop = Fixop.None},
					     longvidleft, longvidright)
			     :: app_exp1
) end
)
 in (LrTable.NT 47,(result,longvid1left,app_exp11right),rest671) end
| (205,rest671) => let val result=MlyValue.app_exp1(fn _ => ([]))
 in (LrTable.NT 48,(result,defaultPos,defaultPos),rest671) end
| (206,(_,(MlyValue.app_exp app_exp1,app_exp1left,app_exp1right))::
rest671) => let val result=MlyValue.app_exp1(fn _ => let val app_exp
 as app_exp1=app_exp1 ()
 in (app_exp) end
)
 in (LrTable.NT 48,(result,app_exp1left,app_exp1right),rest671) end
| (207,(_,(MlyValue.vid vid1,_,vid1right))::(_,(_,OP1left,_))::rest671
) => let val result=MlyValue.aexp(fn _ => let val vid as vid1=vid1 ()
 in (Exp.Var{name = Longvid.short vid,
					 fixop = Fixop.Op}) end
)
 in (LrTable.NT 49,(result,OP1left,vid1right),rest671) end
| (208,(_,(MlyValue.const const1,const1left,const1right))::rest671)
 => let val result=MlyValue.aexp(fn _ => let val const as const1=
const1 ()
 in (Exp.Const const) end
)
 in (LrTable.NT 49,(result,const1left,const1right),rest671) end
| (209,(_,(MlyValue.field field1,_,field1right))::(_,(_,HASH1left,_))
::rest671) => let val result=MlyValue.aexp(fn _ => let val field as 
field1=field1 ()
 in (Exp.Selector field) end
)
 in (LrTable.NT 49,(result,HASH1left,field1right),rest671) end
| (210,(_,(_,_,RBRACE1right))::(_,(MlyValue.elabels elabels1,_,_))::(_
,(_,LBRACE1left,_))::rest671) => let val result=MlyValue.aexp(fn _ => 
let val elabels as elabels1=elabels1 ()
 in (Exp.Record(Record.fromList elabels)) end
)
 in (LrTable.NT 49,(result,LBRACE1left,RBRACE1right),rest671) end
| (211,(_,(_,_,RBRACE1right))::(_,(_,LBRACE1left,_))::rest671) => let 
val result=MlyValue.aexp(fn _ => (Exp.unit))
 in (LrTable.NT 49,(result,LBRACE1left,RBRACE1right),rest671) end
| (212,(_,(_,_,RPAREN1right))::(_,(_,LPAREN1left,_))::rest671) => let 
val result=MlyValue.aexp(fn _ => (Exp.unit))
 in (LrTable.NT 49,(result,LPAREN1left,RPAREN1right),rest671) end
| (213,(_,(_,_,RPAREN1right))::(_,(MlyValue.exp_ps exp_ps1,_,_))::(_,(
_,LPAREN1left,_))::rest671) => let val result=MlyValue.aexp(fn _ => 
let val exp_ps as exp_ps1=exp_ps1 ()
 in (Exp.Seq exp_ps) end
)
 in (LrTable.NT 49,(result,LPAREN1left,RPAREN1right),rest671) end
| (214,(_,(_,_,RPAREN1right))::(_,(MlyValue.exp_2c exp_2c1,_,_))::(_,(
_,LPAREN1left,_))::rest671) => let val result=MlyValue.aexp(fn _ => 
let val exp_2c as exp_2c1=exp_2c1 ()
 in (Exp.tuple exp_2c) end
)
 in (LrTable.NT 49,(result,LPAREN1left,RPAREN1right),rest671) end
| (215,(_,(_,_,RBRACKET1right))::(_,(MlyValue.exp_list exp_list1,_,_))
::(_,(_,LBRACKET1left,_))::rest671) => let val result=MlyValue.aexp(
fn _ => let val exp_list as exp_list1=exp_list1 ()
 in (Exp.List exp_list) end
)
 in (LrTable.NT 49,(result,LBRACKET1left,RBRACKET1right),rest671) end
| (216,(_,(_,_,RBRACKET1right))::(_,(_,LBRACKET1left,_))::rest671) => 
let val result=MlyValue.aexp(fn _ => (Exp.List nil))
 in (LrTable.NT 49,(result,LBRACKET1left,RBRACKET1right),rest671) end
| (217,(_,(_,_,END1right))::(_,(MlyValue.exp_ps exp_ps1,exp_psleft,
exp_psright))::_::(_,(MlyValue.decs decs1,_,_))::(_,(_,LET1left,_))::
rest671) => let val result=MlyValue.aexp(fn _ => let val decs as decs1
=decs1 ()
val exp_ps as exp_ps1=exp_ps1 ()
 in (
Exp.Let(decs,
					       Exp.makeRegion'(Exp.Seq exp_ps,
							       exp_psleft,
							       exp_psright))
) end
)
 in (LrTable.NT 49,(result,LET1left,END1right),rest671) end
| (218,(_,(_,_,SEMICOLON1right))::(_,(MlyValue.ty ty1,_,_))::_::(_,(
MlyValue.STRING STRING1,_,_))::(_,(_,PRIM1left,_))::rest671) => let 
val result=MlyValue.aexp(fn _ => let val STRING as STRING1=STRING1 ()
val ty as ty1=ty1 ()
 in (Exp.Prim{name = STRING,
						   ty = ty}) end
)
 in (LrTable.NT 49,(result,PRIM1left,SEMICOLON1right),rest671) end
| (219,(_,(MlyValue.exp_2c exp_2c1,_,exp_2c1right))::_::(_,(
MlyValue.exp exp1,exp1left,_))::rest671) => let val result=
MlyValue.exp_2c(fn _ => let val exp as exp1=exp1 ()
val exp_2c as exp_2c1=exp_2c1 ()
 in (exp :: exp_2c) end
)
 in (LrTable.NT 51,(result,exp1left,exp_2c1right),rest671) end
| (220,(_,(MlyValue.exp exp2,_,exp2right))::_::(_,(MlyValue.exp exp1,
exp1left,_))::rest671) => let val result=MlyValue.exp_2c(fn _ => let 
val exp1=exp1 ()
val exp2=exp2 ()
 in ([exp1, exp2]) end
)
 in (LrTable.NT 51,(result,exp1left,exp2right),rest671) end
| (221,(_,(MlyValue.exp exp1,exp1left,exp1right))::rest671) => let 
val result=MlyValue.exp_list(fn _ => let val exp as exp1=exp1 ()
 in ([exp]) end
)
 in (LrTable.NT 52,(result,exp1left,exp1right),rest671) end
| (222,(_,(MlyValue.exp_list exp_list1,_,exp_list1right))::_::(_,(
MlyValue.exp exp1,exp1left,_))::rest671) => let val result=
MlyValue.exp_list(fn _ => let val exp as exp1=exp1 ()
val exp_list as exp_list1=exp_list1 ()
 in (exp :: exp_list) end
)
 in (LrTable.NT 52,(result,exp1left,exp_list1right),rest671) end
| (223,(_,(MlyValue.patnode patnode1,patnodeleft as patnode1left,
patnoderight as patnode1right))::rest671) => let val result=
MlyValue.pat(fn _ => let val patnode as patnode1=patnode1 ()
 in (Pat.makeRegion'(patnode, patnodeleft, patnoderight)) end
)
 in (LrTable.NT 53,(result,patnode1left,patnode1right),rest671) end
| (224,(_,(MlyValue.pat pat2,_,pat2right))::_::(_,(MlyValue.pat pat1,
pat1left,_))::rest671) => let val result=MlyValue.patnode(fn _ => let 
val pat1=pat1 ()
val pat2=pat2 ()
 in (Pat.makeAs(pat1,pat2)) end
)
 in (LrTable.NT 54,(result,pat1left,pat2right),rest671) end
| (225,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.pat pat1,
pat1left,_))::rest671) => let val result=MlyValue.patnode(fn _ => let 
val pat as pat1=pat1 ()
val ty as ty1=ty1 ()
 in (Pat.Constraint(pat,ty)) end
)
 in (LrTable.NT 54,(result,pat1left,ty1right),rest671) end
| (226,(_,(MlyValue.apats apats1,apats1left,apats1right))::rest671)
 => let val result=MlyValue.patnode(fn _ => let val apats as apats1=
apats1 ()
 in (Pat.FlatApp apats) end
)
 in (LrTable.NT 54,(result,apats1left,apats1right),rest671) end
| (227,(_,(MlyValue.apat apat1,apat1left,apat1right))::rest671) => 
let val result=MlyValue.apats(fn _ => let val apat as apat1=apat1 ()
 in ([apat]) end
)
 in (LrTable.NT 55,(result,apat1left,apat1right),rest671) end
| (228,(_,(MlyValue.apats apats1,_,apats1right))::(_,(MlyValue.apat 
apat1,apat1left,_))::rest671) => let val result=MlyValue.apats(fn _
 => let val apat as apat1=apat1 ()
val apats as apats1=apats1 ()
 in (apat :: apats) end
)
 in (LrTable.NT 55,(result,apat1left,apats1right),rest671) end
| (229,(_,(MlyValue.apatnode apatnode1,apatnodeleft as apatnode1left,
apatnoderight as apatnode1right))::rest671) => let val result=
MlyValue.apat(fn _ => let val apatnode as apatnode1=apatnode1 ()
 in (Pat.makeRegion'(apatnode,apatnodeleft,apatnoderight)) end
)
 in (LrTable.NT 56,(result,apatnode1left,apatnode1right),rest671) end
| (230,(_,(MlyValue.longvidNoEqual longvidNoEqual1,longvidNoEqual1left
,longvidNoEqual1right))::rest671) => let val result=MlyValue.apatnode(
fn _ => let val longvidNoEqual as longvidNoEqual1=longvidNoEqual1 ()
 in (Pat.Var{name = longvidNoEqual,
 					 fixop = Fixop.None}) end
)
 in (LrTable.NT 57,(result,longvidNoEqual1left,longvidNoEqual1right),
rest671) end
| (231,(_,(MlyValue.vidNoEqual vidNoEqual1,_,vidNoEqual1right))::(_,(_
,OP1left,_))::rest671) => let val result=MlyValue.apatnode(fn _ => 
let val vidNoEqual as vidNoEqual1=vidNoEqual1 ()
 in (Pat.Var{name = Longvid.short vidNoEqual,
					 fixop = Fixop.Op})
 end
)
 in (LrTable.NT 57,(result,OP1left,vidNoEqual1right),rest671) end
| (232,(_,(MlyValue.const const1,const1left,const1right))::rest671)
 => let val result=MlyValue.apatnode(fn _ => let val const as const1=
const1 ()
 in (Pat.Const const) end
)
 in (LrTable.NT 57,(result,const1left,const1right),rest671) end
| (233,(_,(_,WILD1left,WILD1right))::rest671) => let val result=
MlyValue.apatnode(fn _ => (Pat.Wild))
 in (LrTable.NT 57,(result,WILD1left,WILD1right),rest671) end
| (234,(_,(_,_,RPAREN1right))::(_,(MlyValue.pats pats1,_,_))::(_,(_,
LPAREN1left,_))::rest671) => let val result=MlyValue.apatnode(fn _ => 
let val pats as pats1=pats1 ()
 in (Pat.tuple pats) end
)
 in (LrTable.NT 57,(result,LPAREN1left,RPAREN1right),rest671) end
| (235,(_,(_,_,RBRACKET1right))::(_,(MlyValue.pats pats1,_,_))::(_,(_,
LBRACKET1left,_))::rest671) => let val result=MlyValue.apatnode(fn _
 => let val pats as pats1=pats1 ()
 in (Pat.List pats) end
)
 in (LrTable.NT 57,(result,LBRACKET1left,RBRACKET1right),rest671) end
| (236,(_,(_,_,RBRACE1right))::(_,(_,LBRACE1left,_))::rest671) => let 
val result=MlyValue.apatnode(fn _ => (Pat.unit))
 in (LrTable.NT 57,(result,LBRACE1left,RBRACE1right),rest671) end
| (237,(_,(_,_,RBRACE1right))::(_,(MlyValue.patitems patitems1,_,_))::
(_,(_,LBRACE1left,_))::rest671) => let val result=MlyValue.apatnode(
fn _ => let val patitems as patitems1=patitems1 ()
 in (
let val (items,f) = patitems
				  in Pat.Record{items = items, flexible = f}
				  end
) end
)
 in (LrTable.NT 57,(result,LBRACE1left,RBRACE1right),rest671) end
| (238,rest671) => let val result=MlyValue.pats(fn _ => ([]))
 in (LrTable.NT 62,(result,defaultPos,defaultPos),rest671) end
| (239,(_,(MlyValue.commapats commapats1,_,commapats1right))::(_,(
MlyValue.pat pat1,pat1left,_))::rest671) => let val result=
MlyValue.pats(fn _ => let val pat as pat1=pat1 ()
val commapats as commapats1=commapats1 ()
 in (pat :: commapats) end
)
 in (LrTable.NT 62,(result,pat1left,commapats1right),rest671) end
| (240,rest671) => let val result=MlyValue.commapats(fn _ => ([]))
 in (LrTable.NT 63,(result,defaultPos,defaultPos),rest671) end
| (241,(_,(MlyValue.commapats commapats1,_,commapats1right))::(_,(
MlyValue.pat pat1,_,_))::(_,(_,COMMA1left,_))::rest671) => let val 
result=MlyValue.commapats(fn _ => let val pat as pat1=pat1 ()
val commapats as commapats1=commapats1 ()
 in (pat :: commapats) end
)
 in (LrTable.NT 63,(result,COMMA1left,commapats1right),rest671) end
| (242,(_,(MlyValue.patitems patitems1,_,patitems1right))::_::(_,(
MlyValue.patitem patitem1,patitem1left,_))::rest671) => let val result
=MlyValue.patitems(fn _ => let val patitem as patitem1=patitem1 ()
val patitems as patitems1=patitems1 ()
 in (
let val (items,f) = patitems
				    in (patitem::items, f)
				    end
) end
)
 in (LrTable.NT 59,(result,patitem1left,patitems1right),rest671) end
| (243,(_,(MlyValue.patitem patitem1,patitem1left,patitem1right))::
rest671) => let val result=MlyValue.patitems(fn _ => let val patitem
 as patitem1=patitem1 ()
 in ([patitem],false) end
)
 in (LrTable.NT 59,(result,patitem1left,patitem1right),rest671) end
| (244,(_,(_,DOTDOTDOT1left,DOTDOTDOT1right))::rest671) => let val 
result=MlyValue.patitems(fn _ => ([], true))
 in (LrTable.NT 59,(result,DOTDOTDOT1left,DOTDOTDOT1right),rest671)
 end
| (245,(_,(MlyValue.pat pat1,_,pat1right))::_::(_,(MlyValue.field 
field1,field1left,_))::rest671) => let val result=MlyValue.patitem(fn 
_ => let val field as field1=field1 ()
val pat as pat1=pat1 ()
 in (Pat.Item.Field(field,pat)) end
)
 in (LrTable.NT 60,(result,field1left,pat1right),rest671) end
| (246,(_,(MlyValue.opaspat opaspat1,_,opaspat1right))::(_,(
MlyValue.constraint constraint1,_,_))::(_,(MlyValue.vidNoEqual 
vidNoEqual1,vidNoEqual1left,_))::rest671) => let val result=
MlyValue.patitem(fn _ => let val vidNoEqual as vidNoEqual1=vidNoEqual1
 ()
val constraint as constraint1=constraint1 ()
val opaspat as opaspat1=opaspat1 ()
 in (Pat.Item.Vid(vidNoEqual,
						       constraint, opaspat)) end
)
 in (LrTable.NT 60,(result,vidNoEqual1left,opaspat1right),rest671) end
| (247,rest671) => let val result=MlyValue.opaspat(fn _ => (NONE))
 in (LrTable.NT 64,(result,defaultPos,defaultPos),rest671) end
| (248,(_,(MlyValue.pat pat1,_,pat1right))::(_,(_,AS1left,_))::rest671
) => let val result=MlyValue.opaspat(fn _ => let val pat as pat1=pat1 
()
 in (SOME pat) end
)
 in (LrTable.NT 64,(result,AS1left,pat1right),rest671) end
| (249,(_,(MlyValue.tynode tynode1,tynodeleft as tynode1left,
tynoderight as tynode1right))::rest671) => let val result=MlyValue.ty(
fn _ => let val tynode as tynode1=tynode1 ()
 in (Type.makeRegion'(tynode, tynodeleft, tynoderight)) end
)
 in (LrTable.NT 35,(result,tynode1left,tynode1right),rest671) end
| (250,(_,(MlyValue.tuple_ty tuple_ty1,tuple_ty1left,tuple_ty1right))
::rest671) => let val result=MlyValue.tynode(fn _ => let val tuple_ty
 as tuple_ty1=tuple_ty1 ()
 in (Type.tuple tuple_ty) end
)
 in (LrTable.NT 36,(result,tuple_ty1left,tuple_ty1right),rest671) end
| (251,(_,(MlyValue.ty ty2,_,ty2right))::_::(_,(MlyValue.ty ty1,
ty1left,_))::rest671) => let val result=MlyValue.tynode(fn _ => let 
val ty1=ty1 ()
val ty2=ty2 ()
 in (Type.arrow(ty1, ty2)) end
)
 in (LrTable.NT 36,(result,ty1left,ty2right),rest671) end
| (252,(_,(MlyValue.ty'node ty'node1,ty'node1left,ty'node1right))::
rest671) => let val result=MlyValue.tynode(fn _ => let val ty'node as 
ty'node1=ty'node1 ()
 in (ty'node) end
)
 in (LrTable.NT 36,(result,ty'node1left,ty'node1right),rest671) end
| (253,(_,(MlyValue.ty'node ty'node1,ty'nodeleft as ty'node1left,
ty'noderight as ty'node1right))::rest671) => let val result=
MlyValue.ty'(fn _ => let val ty'node as ty'node1=ty'node1 ()
 in (Type.makeRegion'(ty'node, ty'nodeleft, ty'noderight)) end
)
 in (LrTable.NT 32,(result,ty'node1left,ty'node1right),rest671) end
| (254,(_,(MlyValue.tyvar tyvar1,tyvar1left,tyvar1right))::rest671)
 => let val result=MlyValue.ty'node(fn _ => let val tyvar as tyvar1=
tyvar1 ()
 in (Type.Var tyvar) end
)
 in (LrTable.NT 33,(result,tyvar1left,tyvar1right),rest671) end
| (255,(_,(_,_,RBRACE1right))::(_,(MlyValue.tlabels tlabels1,_,_))::(_
,(_,LBRACE1left,_))::rest671) => let val result=MlyValue.ty'node(fn _
 => let val tlabels as tlabels1=tlabels1 ()
 in (Type.Record(Srecord.fromList tlabels)) end
)
 in (LrTable.NT 33,(result,LBRACE1left,RBRACE1right),rest671) end
| (256,(_,(_,_,RBRACE1right))::(_,(_,LBRACE1left,_))::rest671) => let 
val result=MlyValue.ty'node(fn _ => (Type.unit))
 in (LrTable.NT 33,(result,LBRACE1left,RBRACE1right),rest671) end
| (257,(_,(MlyValue.longtycon longtycon1,_,longtycon1right))::_::(_,(
MlyValue.ty0_pc ty0_pc1,_,_))::(_,(_,LPAREN1left,_))::rest671) => let 
val result=MlyValue.ty'node(fn _ => let val ty0_pc as ty0_pc1=ty0_pc1 
()
val longtycon as longtycon1=longtycon1 ()
 in (Type.Con(longtycon, ty0_pc)) end
)
 in (LrTable.NT 33,(result,LPAREN1left,longtycon1right),rest671) end
| (258,(_,(_,_,RPAREN1right))::(_,(MlyValue.ty ty1,_,_))::(_,(_,
LPAREN1left,_))::rest671) => let val result=MlyValue.ty'node(fn _ => 
let val ty as ty1=ty1 ()
 in (Type.node ty) end
)
 in (LrTable.NT 33,(result,LPAREN1left,RPAREN1right),rest671) end
| (259,(_,(MlyValue.longtycon longtycon1,_,longtycon1right))::(_,(
MlyValue.ty' ty'1,ty'1left,_))::rest671) => let val result=
MlyValue.ty'node(fn _ => let val ty' as ty'1=ty'1 ()
val longtycon as longtycon1=longtycon1 ()
 in (Type.Con(longtycon, [ty'])) end
)
 in (LrTable.NT 33,(result,ty'1left,longtycon1right),rest671) end
| (260,(_,(MlyValue.longtycon longtycon1,longtycon1left,
longtycon1right))::rest671) => let val result=MlyValue.ty'node(fn _
 => let val longtycon as longtycon1=longtycon1 ()
 in (Type.Con(longtycon, [])) end
)
 in (LrTable.NT 33,(result,longtycon1left,longtycon1right),rest671)
 end
| (261,(_,(MlyValue.ty ty1,_,ty1right))::_::(_,(MlyValue.field field1,
field1left,_))::rest671) => let val result=MlyValue.tlabel(fn _ => 
let val field as field1=field1 ()
val ty as ty1=ty1 ()
 in (field, ty) end
)
 in (LrTable.NT 30,(result,field1left,ty1right),rest671) end
| (262,(_,(MlyValue.tlabels tlabels1,_,tlabels1right))::_::(_,(
MlyValue.tlabel tlabel1,tlabel1left,_))::rest671) => let val result=
MlyValue.tlabels(fn _ => let val tlabel as tlabel1=tlabel1 ()
val tlabels as tlabels1=tlabels1 ()
 in (tlabel :: tlabels) end
)
 in (LrTable.NT 31,(result,tlabel1left,tlabels1right),rest671) end
| (263,(_,(MlyValue.tlabel tlabel1,tlabel1left,tlabel1right))::rest671
) => let val result=MlyValue.tlabels(fn _ => let val tlabel as tlabel1
=tlabel1 ()
 in ([tlabel]) end
)
 in (LrTable.NT 31,(result,tlabel1left,tlabel1right),rest671) end
| (264,(_,(MlyValue.tuple_ty tuple_ty1,_,tuple_ty1right))::_::(_,(
MlyValue.ty' ty'1,ty'1left,_))::rest671) => let val result=
MlyValue.tuple_ty(fn _ => let val ty' as ty'1=ty'1 ()
val tuple_ty as tuple_ty1=tuple_ty1 ()
 in (ty' :: tuple_ty) end
)
 in (LrTable.NT 34,(result,ty'1left,tuple_ty1right),rest671) end
| (265,(_,(MlyValue.ty' ty'2,_,ty'2right))::_::(_,(MlyValue.ty' ty'1,
ty'1left,_))::rest671) => let val result=MlyValue.tuple_ty(fn _ => 
let val ty'1=ty'1 ()
val ty'2=ty'2 ()
 in ([ty'1, ty'2]) end
)
 in (LrTable.NT 34,(result,ty'1left,ty'2right),rest671) end
| (266,(_,(MlyValue.ty ty2,_,ty2right))::_::(_,(MlyValue.ty ty1,
ty1left,_))::rest671) => let val result=MlyValue.ty0_pc(fn _ => let 
val ty1=ty1 ()
val ty2=ty2 ()
 in ([ty1, ty2]) end
)
 in (LrTable.NT 39,(result,ty1left,ty2right),rest671) end
| (267,(_,(MlyValue.ty0_pc ty0_pc1,_,ty0_pc1right))::_::(_,(
MlyValue.ty ty1,ty1left,_))::rest671) => let val result=
MlyValue.ty0_pc(fn _ => let val ty as ty1=ty1 ()
val ty0_pc as ty0_pc1=ty0_pc1 ()
 in (ty :: ty0_pc) end
)
 in (LrTable.NT 39,(result,ty1left,ty0_pc1right),rest671) end
| (268,(_,(MlyValue.INT INT1,INT1left,INT1right))::rest671) => let 
val result=MlyValue.const(fn _ => let val INT as INT1=INT1 ()
 in (Const.Int INT) end
)
 in (LrTable.NT 8,(result,INT1left,INT1right),rest671) end
| (269,(_,(MlyValue.WORD WORD1,WORD1left,WORD1right))::rest671) => 
let val result=MlyValue.const(fn _ => let val WORD as WORD1=WORD1 ()
 in (Const.Word WORD) end
)
 in (LrTable.NT 8,(result,WORD1left,WORD1right),rest671) end
| (270,(_,(MlyValue.REAL REAL1,REAL1left,REAL1right))::rest671) => 
let val result=MlyValue.const(fn _ => let val REAL as REAL1=REAL1 ()
 in (Const.Real REAL) end
)
 in (LrTable.NT 8,(result,REAL1left,REAL1right),rest671) end
| (271,(_,(MlyValue.STRING STRING1,STRING1left,STRING1right))::rest671
) => let val result=MlyValue.const(fn _ => let val STRING as STRING1=
STRING1 ()
 in (Const.String STRING) end
)
 in (LrTable.NT 8,(result,STRING1left,STRING1right),rest671) end
| (272,(_,(MlyValue.CHAR CHAR1,CHAR1left,CHAR1right))::rest671) => 
let val result=MlyValue.const(fn _ => let val CHAR as CHAR1=CHAR1 ()
 in (Const.Char CHAR) end
)
 in (LrTable.NT 8,(result,CHAR1left,CHAR1right),rest671) end
| (273,(_,(MlyValue.LONGID LONGID1,LONGIDleft as LONGID1left,
LONGIDright as LONGID1right))::rest671) => let val result=
MlyValue.idNoAsterisk(fn _ => let val LONGID as LONGID1=LONGID1 ()
 in (ensureNonqualified(LONGID, reg(LONGIDleft, LONGIDright))) end
)
 in (LrTable.NT 4,(result,LONGID1left,LONGID1right),rest671) end
| (274,(_,(MlyValue.idNoAsterisk idNoAsterisk1,idNoAsterisk1left,
idNoAsterisk1right))::rest671) => let val result=MlyValue.id(fn _ => 
let val idNoAsterisk as idNoAsterisk1=idNoAsterisk1 ()
 in (idNoAsterisk) end
)
 in (LrTable.NT 2,(result,idNoAsterisk1left,idNoAsterisk1right),
rest671) end
| (275,(_,(_,ASTERISK1left,ASTERISK1right))::rest671) => let val 
result=MlyValue.id(fn _ => ("*"))
 in (LrTable.NT 2,(result,ASTERISK1left,ASTERISK1right),rest671) end
| (276,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.idEqual(fn _ => let val id as id1=id1 ()
 in (id) end
)
 in (LrTable.NT 3,(result,id1left,id1right),rest671) end
| (277,(_,(_,EQUALOP1left,EQUALOP1right))::rest671) => let val result=
MlyValue.idEqual(fn _ => ("="))
 in (LrTable.NT 3,(result,EQUALOP1left,EQUALOP1right),rest671) end
| (278,(_,(MlyValue.LONGID LONGID1,LONGID1left,LONGID1right))::rest671
) => let val result=MlyValue.longid(fn _ => let val LONGID as LONGID1=
LONGID1 ()
 in (LONGID) end
)
 in (LrTable.NT 6,(result,LONGID1left,LONGID1right),rest671) end
| (279,(_,(_,ASTERISK1left,ASTERISK1right))::rest671) => let val 
result=MlyValue.longid(fn _ => ("*"))
 in (LrTable.NT 6,(result,ASTERISK1left,ASTERISK1right),rest671) end
| (280,(_,(MlyValue.LONGID LONGID1,LONGID1left,LONGID1right))::rest671
) => let val result=MlyValue.longidNoAsterisk(fn _ => let val LONGID
 as LONGID1=LONGID1 ()
 in (LONGID) end
)
 in (LrTable.NT 5,(result,LONGID1left,LONGID1right),rest671) end
| (281,(_,(MlyValue.longid longid1,longid1left,longid1right))::rest671
) => let val result=MlyValue.longidEqual(fn _ => let val longid as 
longid1=longid1 ()
 in (longid) end
)
 in (LrTable.NT 7,(result,longid1left,longid1right),rest671) end
| (282,(_,(_,EQUALOP1left,EQUALOP1right))::rest671) => let val result=
MlyValue.longidEqual(fn _ => ("="))
 in (LrTable.NT 7,(result,EQUALOP1left,EQUALOP1right),rest671) end
| (283,(_,(MlyValue.idEqual idEqual1,idEqual1left,idEqual1right))::
rest671) => let val result=MlyValue.vid(fn _ => let val idEqual as 
idEqual1=idEqual1 ()
 in (Vid.fromString idEqual) end
)
 in (LrTable.NT 9,(result,idEqual1left,idEqual1right),rest671) end
| (284,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.vidNoEqual(fn _ => let val id as id1=id1 ()
 in (Vid.fromString id) end
)
 in (LrTable.NT 10,(result,id1left,id1right),rest671) end
| (285,(_,(MlyValue.vid vid1,vid1left,vid1right))::rest671) => let 
val result=MlyValue.vids(fn _ => let val vid as vid1=vid1 ()
 in ([vid]) end
)
 in (LrTable.NT 11,(result,vid1left,vid1right),rest671) end
| (286,(_,(MlyValue.vids vids1,_,vids1right))::(_,(MlyValue.vid vid1,
vid1left,_))::rest671) => let val result=MlyValue.vids(fn _ => let 
val vid as vid1=vid1 ()
val vids as vids1=vids1 ()
 in (vid::vids) end
)
 in (LrTable.NT 11,(result,vid1left,vids1right),rest671) end
| (287,(_,(MlyValue.idEqual idEqual1,idEqual1left,idEqual1right))::
rest671) => let val result=MlyValue.var(fn _ => let val idEqual as 
idEqual1=idEqual1 ()
 in (Var.fromString idEqual) end
)
 in (LrTable.NT 12,(result,idEqual1left,idEqual1right),rest671) end
| (288,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.con(fn _ => let val id as id1=id1 ()
 in (Con.fromString id) end
)
 in (LrTable.NT 14,(result,id1left,id1right),rest671) end
| (289,(_,(MlyValue.idNoAsterisk idNoAsterisk1,idNoAsterisk1left,
idNoAsterisk1right))::rest671) => let val result=MlyValue.tycon(fn _
 => let val idNoAsterisk as idNoAsterisk1=idNoAsterisk1 ()
 in (Tycon.fromString idNoAsterisk) end
)
 in (LrTable.NT 16,(result,idNoAsterisk1left,idNoAsterisk1right),
rest671) end
| (290,(_,(MlyValue.TYVAR TYVAR1,TYVARleft as TYVAR1left,TYVARright
 as TYVAR1right))::rest671) => let val result=MlyValue.tyvar(fn _ => 
let val TYVAR as TYVAR1=TYVAR1 ()
 in (Tyvar.newString(TYVAR, TYVARleft, TYVARright)) end
)
 in (LrTable.NT 15,(result,TYVAR1left,TYVAR1right),rest671) end
| (291,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.field(fn _ => let val id as id1=id1 ()
 in (Field.String id) end
)
 in (LrTable.NT 17,(result,id1left,id1right),rest671) end
| (292,(_,(MlyValue.int int1,intleft as int1left,intright as int1right
))::rest671) => let val result=MlyValue.field(fn _ => let val int as 
int1=int1 ()
 in (
Field.Int
				(if int <= 0
				    then (error(reg(intleft, intright),
						"expected integer") ;
					  ~1)
				 else int)
) end
)
 in (LrTable.NT 17,(result,int1left,int1right),rest671) end
| (293,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.strid(fn _ => let val id as id1=id1 ()
 in (Strid.fromString id) end
)
 in (LrTable.NT 18,(result,id1left,id1right),rest671) end
| (294,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.sigid(fn _ => let val id as id1=id1 ()
 in (Sigid.fromString id) end
)
 in (LrTable.NT 19,(result,id1left,id1right),rest671) end
| (295,(_,(MlyValue.sigid sigid1,sigid1left,sigid1right))::rest671)
 => let val result=MlyValue.sigids(fn _ => let val sigid as sigid1=
sigid1 ()
 in ([sigid]) end
)
 in (LrTable.NT 20,(result,sigid1left,sigid1right),rest671) end
| (296,(_,(MlyValue.sigids sigids1,_,sigids1right))::(_,(
MlyValue.sigid sigid1,sigid1left,_))::rest671) => let val result=
MlyValue.sigids(fn _ => let val sigid as sigid1=sigid1 ()
val sigids as sigids1=sigids1 ()
 in (sigid::sigids) end
)
 in (LrTable.NT 20,(result,sigid1left,sigids1right),rest671) end
| (297,(_,(MlyValue.id id1,id1left,id1right))::rest671) => let val 
result=MlyValue.fctid(fn _ => let val id as id1=id1 ()
 in (Fctid.fromString id) end
)
 in (LrTable.NT 21,(result,id1left,id1right),rest671) end
| (298,(_,(MlyValue.longidNoAsterisk longidNoAsterisk1,
longidNoAsteriskleft as longidNoAsterisk1left,longidNoAsteriskright
 as longidNoAsterisk1right))::rest671) => let val result=
MlyValue.longtycon(fn _ => let val longidNoAsterisk as 
longidNoAsterisk1=longidNoAsterisk1 ()
 in (
Longtycon.fromString(longidNoAsterisk,
						   reg(longidNoAsteriskleft,
						       longidNoAsteriskright))
) end
)
 in (LrTable.NT 22,(result,longidNoAsterisk1left,
longidNoAsterisk1right),rest671) end
| (299,(_,(MlyValue.longidEqual longidEqual1,longidEqualleft as 
longidEqual1left,longidEqualright as longidEqual1right))::rest671) => 
let val result=MlyValue.longvar(fn _ => let val longidEqual as 
longidEqual1=longidEqual1 ()
 in (
Longvar.fromString(longidEqual,
					       reg(longidEqualleft,
						   longidEqualright))
) end
)
 in (LrTable.NT 23,(result,longidEqual1left,longidEqual1right),rest671
) end
| (300,(_,(MlyValue.longidEqual longidEqual1,longidEqualleft as 
longidEqual1left,longidEqualright as longidEqual1right))::rest671) => 
let val result=MlyValue.longvid(fn _ => let val longidEqual as 
longidEqual1=longidEqual1 ()
 in (
Longvid.fromString(longidEqual,
					       reg(longidEqualleft,
						   longidEqualright))
) end
)
 in (LrTable.NT 26,(result,longidEqual1left,longidEqual1right),rest671
) end
| (301,(_,(MlyValue.longid longid1,longidleft as longid1left,
longidright as longid1right))::rest671) => let val result=
MlyValue.longvidNoEqual(fn _ => let val longid as longid1=longid1 ()
 in (
Longvid.fromString(longid,
					       reg(longidleft,longidright)))
 end
)
 in (LrTable.NT 27,(result,longid1left,longid1right),rest671) end
| (302,(_,(MlyValue.longid longid1,longidleft as longid1left,
longidright as longid1right))::rest671) => let val result=
MlyValue.longcon(fn _ => let val longid as longid1=longid1 ()
 in (
Longcon.fromString(longid,
					       reg(longidleft,longidright)))
 end
)
 in (LrTable.NT 25,(result,longid1left,longid1right),rest671) end
| (303,(_,(MlyValue.longid longid1,longidleft as longid1left,
longidright as longid1right))::rest671) => let val result=
MlyValue.longstrid(fn _ => let val longid as longid1=longid1 ()
 in (Longstrid.fromString(longid,
						 reg(longidleft,longidright)))
 end
)
 in (LrTable.NT 28,(result,longid1left,longid1right),rest671) end
| (304,(_,(MlyValue.longstrid longstrid1,longstrid1left,
longstrid1right))::rest671) => let val result=MlyValue.longstrids(fn _
 => let val longstrid as longstrid1=longstrid1 ()
 in ([longstrid]) end
)
 in (LrTable.NT 29,(result,longstrid1left,longstrid1right),rest671)
 end
| (305,(_,(MlyValue.longstrids longstrids1,_,longstrids1right))::(_,(
MlyValue.longstrid longstrid1,longstrid1left,_))::rest671) => let val 
result=MlyValue.longstrids(fn _ => let val longstrid as longstrid1=
longstrid1 ()
val longstrids as longstrids1=longstrids1 ()
 in (longstrid :: longstrids) end
)
 in (LrTable.NT 29,(result,longstrid1left,longstrids1right),rest671)
 end
| _ => raise (mlyAction i392)
end
val void = MlyValue.VOID
val extract = fn a => (fn MlyValue.program x => x
| _ => let exception ParseInternal
	in raise ParseInternal end) a ()
end
end
structure Tokens : ML_TOKENS =
struct
type svalue = ParserData.svalue
type ('a,'b) token = ('a,'b) Token.token
fun EOF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 0,(
ParserData.MlyValue.VOID,p1,p2))
fun SEMICOLON (p1,p2) = Token.TOKEN (ParserData.LrTable.T 1,(
ParserData.MlyValue.VOID,p1,p2))
fun LONGID (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 2,(
ParserData.MlyValue.LONGID (fn () => i),p1,p2))
fun TYVAR (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 3,(
ParserData.MlyValue.TYVAR (fn () => i),p1,p2))
fun INT (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 4,(
ParserData.MlyValue.INT (fn () => i),p1,p2))
fun WORD (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 5,(
ParserData.MlyValue.WORD (fn () => i),p1,p2))
fun REAL (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 6,(
ParserData.MlyValue.REAL (fn () => i),p1,p2))
fun STRING (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 7,(
ParserData.MlyValue.STRING (fn () => i),p1,p2))
fun CHAR (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 8,(
ParserData.MlyValue.CHAR (fn () => i),p1,p2))
fun ABSTYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 9,(
ParserData.MlyValue.VOID,p1,p2))
fun AND (p1,p2) = Token.TOKEN (ParserData.LrTable.T 10,(
ParserData.MlyValue.VOID,p1,p2))
fun ARROW (p1,p2) = Token.TOKEN (ParserData.LrTable.T 11,(
ParserData.MlyValue.VOID,p1,p2))
fun AS (p1,p2) = Token.TOKEN (ParserData.LrTable.T 12,(
ParserData.MlyValue.VOID,p1,p2))
fun BAR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 13,(
ParserData.MlyValue.VOID,p1,p2))
fun CASE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 14,(
ParserData.MlyValue.VOID,p1,p2))
fun DATATYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 15,(
ParserData.MlyValue.VOID,p1,p2))
fun DOTDOTDOT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 16,(
ParserData.MlyValue.VOID,p1,p2))
fun ELSE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 17,(
ParserData.MlyValue.VOID,p1,p2))
fun END (p1,p2) = Token.TOKEN (ParserData.LrTable.T 18,(
ParserData.MlyValue.VOID,p1,p2))
fun EQUALOP (p1,p2) = Token.TOKEN (ParserData.LrTable.T 19,(
ParserData.MlyValue.VOID,p1,p2))
fun EQTYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 20,(
ParserData.MlyValue.VOID,p1,p2))
fun EXCEPTION (p1,p2) = Token.TOKEN (ParserData.LrTable.T 21,(
ParserData.MlyValue.VOID,p1,p2))
fun DO (p1,p2) = Token.TOKEN (ParserData.LrTable.T 22,(
ParserData.MlyValue.VOID,p1,p2))
fun DARROW (p1,p2) = Token.TOKEN (ParserData.LrTable.T 23,(
ParserData.MlyValue.VOID,p1,p2))
fun FN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 24,(
ParserData.MlyValue.VOID,p1,p2))
fun FUN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 25,(
ParserData.MlyValue.VOID,p1,p2))
fun FUNCTOR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 26,(
ParserData.MlyValue.VOID,p1,p2))
fun HANDLE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 27,(
ParserData.MlyValue.VOID,p1,p2))
fun HASH (p1,p2) = Token.TOKEN (ParserData.LrTable.T 28,(
ParserData.MlyValue.VOID,p1,p2))
fun IF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 29,(
ParserData.MlyValue.VOID,p1,p2))
fun IN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 30,(
ParserData.MlyValue.VOID,p1,p2))
fun INCLUDE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 31,(
ParserData.MlyValue.VOID,p1,p2))
fun INFIX (p1,p2) = Token.TOKEN (ParserData.LrTable.T 32,(
ParserData.MlyValue.VOID,p1,p2))
fun INFIXR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 33,(
ParserData.MlyValue.VOID,p1,p2))
fun LET (p1,p2) = Token.TOKEN (ParserData.LrTable.T 34,(
ParserData.MlyValue.VOID,p1,p2))
fun LOCAL (p1,p2) = Token.TOKEN (ParserData.LrTable.T 35,(
ParserData.MlyValue.VOID,p1,p2))
fun NONFIX (p1,p2) = Token.TOKEN (ParserData.LrTable.T 36,(
ParserData.MlyValue.VOID,p1,p2))
fun OF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 37,(
ParserData.MlyValue.VOID,p1,p2))
fun OP (p1,p2) = Token.TOKEN (ParserData.LrTable.T 38,(
ParserData.MlyValue.VOID,p1,p2))
fun OPEN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 39,(
ParserData.MlyValue.VOID,p1,p2))
fun OVERLOAD (p1,p2) = Token.TOKEN (ParserData.LrTable.T 40,(
ParserData.MlyValue.VOID,p1,p2))
fun RAISE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 41,(
ParserData.MlyValue.VOID,p1,p2))
fun REC (p1,p2) = Token.TOKEN (ParserData.LrTable.T 42,(
ParserData.MlyValue.VOID,p1,p2))
fun SHARING (p1,p2) = Token.TOKEN (ParserData.LrTable.T 43,(
ParserData.MlyValue.VOID,p1,p2))
fun SIG (p1,p2) = Token.TOKEN (ParserData.LrTable.T 44,(
ParserData.MlyValue.VOID,p1,p2))
fun SIGNATURE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 45,(
ParserData.MlyValue.VOID,p1,p2))
fun STRUCT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 46,(
ParserData.MlyValue.VOID,p1,p2))
fun STRUCTURE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 47,(
ParserData.MlyValue.VOID,p1,p2))
fun THEN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 48,(
ParserData.MlyValue.VOID,p1,p2))
fun TYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 49,(
ParserData.MlyValue.VOID,p1,p2))
fun VAL (p1,p2) = Token.TOKEN (ParserData.LrTable.T 50,(
ParserData.MlyValue.VOID,p1,p2))
fun WHERE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 51,(
ParserData.MlyValue.VOID,p1,p2))
fun WHILE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 52,(
ParserData.MlyValue.VOID,p1,p2))
fun WILD (p1,p2) = Token.TOKEN (ParserData.LrTable.T 53,(
ParserData.MlyValue.VOID,p1,p2))
fun WITH (p1,p2) = Token.TOKEN (ParserData.LrTable.T 54,(
ParserData.MlyValue.VOID,p1,p2))
fun WITHTYPE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 55,(
ParserData.MlyValue.VOID,p1,p2))
fun ASTERISK (p1,p2) = Token.TOKEN (ParserData.LrTable.T 56,(
ParserData.MlyValue.VOID,p1,p2))
fun COLON (p1,p2) = Token.TOKEN (ParserData.LrTable.T 57,(
ParserData.MlyValue.VOID,p1,p2))
fun COLONGT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 58,(
ParserData.MlyValue.VOID,p1,p2))
fun COMMA (p1,p2) = Token.TOKEN (ParserData.LrTable.T 59,(
ParserData.MlyValue.VOID,p1,p2))
fun LBRACE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 60,(
ParserData.MlyValue.VOID,p1,p2))
fun LBRACKET (p1,p2) = Token.TOKEN (ParserData.LrTable.T 61,(
ParserData.MlyValue.VOID,p1,p2))
fun LPAREN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 62,(
ParserData.MlyValue.VOID,p1,p2))
fun RBRACE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 63,(
ParserData.MlyValue.VOID,p1,p2))
fun RBRACKET (p1,p2) = Token.TOKEN (ParserData.LrTable.T 64,(
ParserData.MlyValue.VOID,p1,p2))
fun RPAREN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 65,(
ParserData.MlyValue.VOID,p1,p2))
fun ORELSE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 66,(
ParserData.MlyValue.VOID,p1,p2))
fun ANDALSO (p1,p2) = Token.TOKEN (ParserData.LrTable.T 67,(
ParserData.MlyValue.VOID,p1,p2))
fun FUNSIG (p1,p2) = Token.TOKEN (ParserData.LrTable.T 68,(
ParserData.MlyValue.VOID,p1,p2))
fun PRIM (p1,p2) = Token.TOKEN (ParserData.LrTable.T 69,(
ParserData.MlyValue.VOID,p1,p2))
end
end
